/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <assert.h>

#include "VideoFile.h"
#include "A_Codec.h"
#include "debug.h"
#include "translate.h"



/***** Konstruktor / Destruktor *****/
A_Codec::A_Codec( A_Track *Track )
:Codec( Track ), A_Format( *Track )
{
    // Variablen initialisieren
    track = Track;

    sample_puffer_int16 = NULL;
    sample_puffer_float = NULL;
    sample_puffer_channels = 0;
    sample_puffer_len = 0;

    // schonmal die Puffer anlegen
    alloc_buffers( 0 );
}




A_Codec::~A_Codec()
{
    for( int i=0; i < Channels(); i++ ) {
        delete [] sample_puffer_float[i];
        delete [] sample_puffer_int16[i];
    }
    delete [] sample_puffer_float;
    delete [] sample_puffer_int16;
}




// Bild-Position setzen
/* Rckgabewert: 1   -  OK
                 0   -  Fehler */
int A_Codec::SeekTime( double seconds )
{
    if( iomode != VideoFile::read )
        return 0;
        
    long long sample = (long long)(seconds * SampleRate() );
    return SeekSample( sample );
}




int A_Codec::SeekSample( int64_t sample )
{
    // Block, in dem Sample steht
    double samp_pro_block = (double)SampleRate() / track->Rate();
    long block = int(sample / samp_pro_block);
    int64_t rest = sample - int(block * samp_pro_block);
    assert( rest >= 0 );

    // Auf Block positionieren
    track->Seek( block );

    // restliche Samples im Block berspringen
    if( rest > 0 ) {
        int16_t **buf=NULL;
        return decodeAudio( buf, rest );
    }

    return 1;
}




int A_Codec::encodeAudio( int16_t **daten, int samples )
{
    if( iomode != VideoFile::write )
        return 0;

    // Puffer anlegen
    alloc_buffers( samples );
    
    // konvertieren
    for( int i=0; i < samples; i++ ) {
        for( int j=0; j < Channels(); j++ ) {
            sample_puffer_float[j][i] = (float)daten[j][i] / (float)(1<<15);
        }
    }

    return encode( sample_puffer_float, samples );;
}





int A_Codec::encodeAudio( float **daten, int samples )
{
    if( iomode != VideoFile::write )
        return 0;

    // Puffer anlegen
    alloc_buffers( samples );

    return encode( daten, samples );;
}





int A_Codec::decodeAudio( int16_t **&daten, int samples )
{
    int stat = 0;

    if( iomode != VideoFile::read ) {
        return 0;
    }

    // Puffer anlegen
    alloc_buffers( samples );

    daten = sample_puffer_int16;

    // dekodieren
    stat = decode( sample_puffer_float, samples );
    if( !stat )
        return stat;

    // konvertieren
    for( int i=0; i < samples; i++ ) {
        for( int j=0; j < Channels(); j++ ) {
            daten[j][i] = (int16_t)( sample_puffer_float[j][i] * (1<<15) );
        }
    }

    return stat;
}





int A_Codec::decodeAudio( float **&daten, int samples )
{
    int stat = 0;

    if( iomode != VideoFile::read ) {
        return 0;
    }

    // Puffer anlegen
    alloc_buffers( samples );

    daten = sample_puffer_float;

    // dekodieren
    stat = decode( sample_puffer_float, samples );
    return stat;
}







/***** Puffer-Verwaltung *****/
void A_Codec::alloc_buffers( long len )
{
    //reichen vorhandene Puffer aus?
    if( Channels() > sample_puffer_channels ) {
        // Alte Puffer lschen
        for( int i=0; i < sample_puffer_channels; i++ ) {
            delete [] sample_puffer_float[i];
            delete [] sample_puffer_int16[i];
        }
        delete [] sample_puffer_float;
        delete [] sample_puffer_int16;
    
        sample_puffer_int16 = new int16_t*[ Channels() ];
        sample_puffer_float = new float*[ Channels() ];
        sample_puffer_channels = Channels();
        sample_puffer_len = 0;

        //Sample Puffer initialisierne
        for( int i=0; i < Channels(); i++ ) {
            sample_puffer_int16[i] = NULL;
            sample_puffer_float[i] = NULL;
        }
    }
    
    //Sample-Puffer anlegen
    if( len > sample_puffer_len ) {
        sample_puffer_len = len;
        for( int i=0; i < sample_puffer_channels; i++ ) {
            delete [] sample_puffer_int16[i];
            delete [] sample_puffer_float[i];
            sample_puffer_int16[i] = new int16_t[sample_puffer_len];
            sample_puffer_float[i] = new float[sample_puffer_len];
        }
    }
}






/***** Diese Funktionen werden vom Codec implementiert *****/

// Audiodaten schreiben
int A_Codec::encode( float **daten, int sampels )
{
    ERROR_TXT( _("encode audio not implemented!") );
    return 0;
}


// Audiodaten schreiben
int A_Codec::decode( float **daten, int sampels )
{
    ERROR_TXT( _("decode audio not implemented!") );
    return 0;
}

