/*****************************************************************************\
                          FrameData.C  -  description
                             -------------------
    begin                : Son Dez 15 2002
    copyright            : (C) 2002 by Christian Tusche
    email                : ctusche@gmx.net
\*****************************************************************************/

/*****************************************************************************\
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <assert.h>
#include <string.h>
#include "FrameData.h"
#include "cmodels.h"

FrameData::FrameData( int x, int y, uint8_t *memseg, size_t size )
{
    data = NULL;
    data_size = 0;
    y_plane = u_plane = v_plane = NULL;
    frame_size = 0;
    y_size = u_size = v_size = 0;

    c_model = RGB24;

    breite = x;
    hoehe = y;

    alloc_buf = false;
 
    // Puffer anlegen
    if( memseg == NULL ) {
        alloc_data();
    } else {
        assert( size > 0 );
        data_size = size;
        data = memseg;
        yuv_planes();
    }
    
    ColorModel( RGB24 );    
}



FrameData::~FrameData()
{
    free_data();
}




/**Daten aus *bild werden in den internen Framebuffer kopiert!
  */
void FrameData::SetFrame( FrameData *bild )
{
    if( bild == NULL ) {
        return;
    }

    if( bild->breite != breite || bild->hoehe != hoehe || bild->c_model != c_model ) {
        return;
    }

    // Datenfeld auf *bild setzen
    memcpy( data, bild->data, frame_size );
}




/**Daten aus *bild werden in den internen Framebuffer kopiert!
  */
void FrameData::SetFrame( uint8_t *bild )
{
    if( bild == NULL ) {
        return;
    }

    // Datenfeld auf *bild setzen
    memcpy( data, bild, frame_size );
}





    
/**Daten aus *bild werden in den internen Framebuffer kopiert!
  */
void FrameData::SetFrame( uint8_t *y, uint8_t *u, uint8_t *v )
{
    // nur bei planaren Farbrumen
    if( !(c_model & PLANAR) )
        return;
        
    if( y == NULL || u == NULL || v == NULL ) {
        return;
    }

    // yuv-Planes kopieren
    memcpy( y_plane, y, y_size );
    memcpy( u_plane, u, u_size );
    memcpy( v_plane, v, v_size );
}






    
/**Zeiger auf Datenfeld - linearer Framebuffer<br>
  *Rckgabewert: Gre der Bilddaten
  */
int FrameData::FramePointer( uint8_t *&bild )
{
    bild = data;
    return frame_size;
}




    
/**Zeiger auf Datenfeld - planar<br>
  *Rckgabewert: Gre der Bilddaten von Y+U+V-Komponente
  */
int FrameData::FramePointer( uint8_t *&y, uint8_t *&u, uint8_t *&v )
{
    if( y_plane != NULL ) {
        y = y_plane;
        u = u_plane;
        v = v_plane;
        return frame_size;
    }

    // kein planares YUV-Format
    y = u = v = NULL;
    return 0;
}






/**Farbraum der Bilddaten festlegen
  */
void FrameData::ColorModel( int cmodel )
{
    int bpp = 0;
    
    // Farbraum unbekannt
    if( cmodel & CMODEL_MASK > NUM_CMODELS )
        return;

    c_model = cmodel;

    // Gre der yuv-Planes berechnen
    if( c_model & YUV422 && c_model & PLANAR ) {
        y_size = breite * hoehe;
        u_size = (breite * hoehe) >> 1;
        v_size = u_size;
        frame_size = y_size + u_size + v_size;
    } else if( c_model & YUV420 && c_model & PLANAR ) {
        y_size = breite * hoehe;
        u_size = (breite >> 1) * ((hoehe + 1) >> 1);
        v_size = u_size;
        frame_size = y_size + u_size + v_size;
    } else {
        // Datengre berechnen
        bpp = CMODEL_BPP[c_model & CMODEL_MASK];
        frame_size = (breite * hoehe * bpp + 7) >> 3;
        y_size = u_size = v_size = 0;
    }

    // yuv-Planes anlegen
    yuv_planes();
}


    

    


    
/**Neue Puffer anlegen
  */
void FrameData::alloc_data(void)
{
    if( !alloc_buf ) {
        data_size = breite * (hoehe+1) * MAX_BPP / 8;
        data = new uint8_t[ data_size ];

        // Puffer lschen
        memset( data, 0, data_size );
        
        // yuv-Planes erzeugen
        yuv_planes();

        alloc_buf = true;
    }
}





    
/**Puffer freigeben
  */
void FrameData::free_data( void )
{
    if( alloc_buf ) {
        delete [] data;
        data = NULL;
        data_size = 0;
        y_plane = u_plane = v_plane = NULL;
        alloc_buf = false;
    }
}








/**yuv-Planes erzeugen
  */
void FrameData::yuv_planes( void )
{
    if( !(c_model & PLANAR) ) {
        y_plane = u_plane = v_plane = NULL;
        return;
    }
        
    y_plane = data;
    u_plane = y_plane + y_size;
    v_plane = u_plane + u_size;
}
