/*
 * High quality image resampling with polyphase filters 
 * Copyright (c) 2001 Gerard Lantau.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

//#define HAVE_MMX

#include "FrameRescaler.h"
#include "translate.h"
#include "debug.h"

#define POS_FRAC_BITS 16
#define POS_FRAC      (1 << POS_FRAC_BITS)
/* 6 bits precision is needed for MMX */
#define FILTER_BITS   8

#define LINE_BUF_HEIGHT (NB_TAPS * 8)



FrameRescaler::FrameRescaler( int cmodel, int iwidth, int iheight, int owidth, int oheight )
{
    // Variablen initialisieren
    line_buf = new int16_t[ owidth * (LINE_BUF_HEIGHT + NB_TAPS) ];
    
    o_width = owidth;
    o_height = oheight;
    i_width = iwidth;
    i_height = iheight;
    
    h_incr = (iwidth * POS_FRAC) / owidth;
    v_incr = (iheight * POS_FRAC) / oheight;

    //mm_flags = MM_MMX;
    //mm_flags = 0x0001;
    
    // Farbraum fr scale auswhlen
    c_model = cmodel;
    cmodel &= CMODEL_MASK;
    cmodel_converter = NULL;
    temp_buf = NULL;
    if( c_model != YUV420P && cmodel != YUV422 && cmodel != RGB24 && cmodel != GREY8 ) {
        cmodel_converter = new Convert( c_model, YUV420P, i_width, i_height );
        temp_buf = new uint8_t[ i_width * i_height * 2 ];
        c_model = YUV420P;
    }

    build_filter( h_filters, (float)o_width / (float)i_width);
    build_filter( v_filters, (float)o_height / (float)i_height);
}




FrameRescaler::~FrameRescaler()
{
    delete [] line_buf;
}




void FrameRescaler::Scale( uint8_t *input, uint8_t *output)
{
    int cmodel = c_model & CMODEL_MASK;

    TIMER_START( 1 );

    if( cmodel_converter != NULL ) {
        // Farbraum umwandeln
        cmodel_converter->do_convert( input, temp_buf );
        scale_yuv420( temp_buf, output );
    }
            
    if( c_model == YUV420P ) {
        scale_yuv420( input, output );
    } else if( cmodel == YUV422 ) {
        scale_yuv422( input, output );
    } else if( cmodel == RGB24 ) {
        scale_rgb24( input, output );
    } else if( cmodel == GREY8 ) {
        scale_grey8( input, output );
    } else {
        ERROR_TXT( _("Unknown colormodel - cannot scale") );
    }

    TIMER_STOP( 1, "Scale Colormodel %i", c_model );
}








void FrameRescaler::scale_yuv420( uint8_t *input, uint8_t *output)
{
    int i, shift;
    uint8_t *in[3], *out[3];

    if( c_model != YUV420P ) {
        ERROR_TXT( _("Unknown colormodel - cannot scale") );
        return;
    }

    in[0] = input;
    in[1] = in[0] + i_width * i_height;
    in[2] = in[1] + (i_width >> 1) * ((i_height + 1) >> 1);

    out[0] = output;
    out[1] = out[0] + o_width * o_height;
    out[2] = out[1] + (o_width >> 1 ) * ((o_height + 1) >> 1);

    for( i=0; i<3; i++ ) {
        shift = (i==0 ? 0:1);
        component_resample( out[i], o_width >> shift,
                            o_width >> shift, o_height >> shift,
                            in[i], i_width >> shift,
                            i_width >> shift, i_height >> shift,
                            1 );
    }
}






void FrameRescaler::scale_yuv422( uint8_t *input, uint8_t *output)
{
    int i, shift, inc[3];
    uint8_t *in[3], *out[3];

    if( (c_model & CMODEL_MASK) != YUV422 ) {
        ERROR_TXT( _("Unknown colormodel - cannot scale") );
        return;
    }

    if( c_model & PLANAR ) {
        inc[0] = inc[1] = inc[2] = 0;

        in[0] = input;
        in[1] = in[0] + i_width * i_height;
        in[2] = in[1] + (i_width >> 1) * ((i_height + 1) >> 1);

        out[0] = output;
        out[1] = out[0] + o_width * o_height;
        out[2] = out[1] + (o_width >> 1 ) * ((o_height + 1) >> 1);

    } else if( c_model == YUV422 ) {
        inc[0] = 2;
        inc[1] = inc[2] = 4;

        in[0] = input;
        in[1] = in[0] + 1;
        in[2] = in[1] + 2;

        out[0] = output;
        out[1] = out[0] + 1;
        out[2] = out[1] + 2;
    }

    for( i=0; i<3; i++ ) {
        shift = (i==0 ? 0:1);
        component_resample( out[i], o_width >> shift,
                            o_width >> shift, o_height,
                            in[i], i_width >> shift,
                            i_width >> shift, i_height,
                            inc[i] );
    }
}






void FrameRescaler::scale_rgb24( uint8_t *input, uint8_t *output)
{
    int i;
    uint8_t *in[3], *out[3];

    if( (c_model & CMODEL_MASK) != RGB24 ) {
        ERROR_TXT( _("Unknown colormodel - cannot scale") );
        return;
    }

    in[0] = input;
    in[1] = in[0] + 1;
    in[2] = in[1] + 1;

    out[0] = output;
    out[1] = out[0] + 1;
    out[2] = out[1] + 1;

    for( i=0; i<3; i++ ) {
        component_resample( out[i], o_width, o_width, o_height,
                            in[i], i_width, i_width, i_height, 3 );
    }
}





void FrameRescaler::scale_grey8( uint8_t *input, uint8_t *output)
{
    if( (c_model & CMODEL_MASK) != GREY8 ) {
        ERROR_TXT( _("Unknown colormodel - cannot scale") );
        return;
    }

    component_resample( output, o_width, o_width, o_height,
                        input, i_width, i_width, i_height, 1 );
}






/* XXX: the following filter is quite naive, but it seems to suffice
   for 4 taps */
void FrameRescaler::build_filter(int16_t *filter, float factor)
{
    int ph, i, v;
    float x, y, tab[NB_TAPS], norm, mult;

    /* if upsampling, only need to interpolate, no filter */
    if (factor > 1.0)
        factor = 1.0;

    for(ph=0;ph<NB_PHASES;ph++) {
        norm = 0;
        for(i=0;i<NB_TAPS;i++) {

            x = M_PI * ((float)(i - FCENTER) - (float)ph / NB_PHASES) * factor;
            if (x == 0)
                y = 1.0;
            else
                y = sin(x) / x;
            tab[i] = y;
            norm += y;
        }

        /* normalize so that an uniform color remains the same */
        mult = (float)(1 << FILTER_BITS) / norm;
        for(i=0;i<NB_TAPS;i++) {
            v = (int)(tab[i] * mult);
            filter[ph * NB_TAPS + i] = v;
        }
    }
}









inline int16_t FrameRescaler::get_phase(int pos)
{
    return ((pos) >> (POS_FRAC_BITS - PHASE_BITS)) & ((1 << PHASE_BITS) - 1);
}








/* This function must be optimized */
void FrameRescaler::h_resample_fast(int16_t *dst, int dst_width, uint8_t *src, int src_width,
                            int src_start, int src_incr, int16_t *filters, int inc )
{
    int32_t src_pos;
    int16_t phase, i;
    register int32_t sum;
    uint8_t *s;
    int16_t *filter;

    src_pos = src_start;
    for(i=0;i<dst_width;i++) {
#if 0
        /* test */
        if ((src_pos >> POS_FRAC_BITS) < 0 ||
            (src_pos >> POS_FRAC_BITS) > (src_width - NB_TAPS))
            abort();
#endif
        s = src + (src_pos >> POS_FRAC_BITS) * inc;
        phase = get_phase(src_pos);
        filter = &filters[phase * NB_TAPS];
#if NB_TAPS == 4
        
        sum = s[0]     * filter[0] +
              s[1*inc] * filter[1] +
              s[2*inc] * filter[2] +
              s[3*inc] * filter[3];
#else
        {
            int j;
            sum = 0;
            for(j=0;j<NB_TAPS;j++)
                sum += s[j*inc] * filter[j];
        }
#endif
        sum = sum >> FILTER_BITS;

        // Daten werden hier als int16_t an v_resample bergeben
        // Grenzen werden in v_resample gerundet
        /*
        if( sum & 0x7FFFFF00 ) {
            if (sum < 0) {
                sum = 0;
            } else if (sum > 255){
                sum = 255;
            }
        }
        */
        
        dst[0] = sum;
        src_pos += src_incr;
        dst++;
    }
}



#if NB_TAPS == 4
/* This function must be optimized */
void FrameRescaler::h_resample_fast4_inc1(int16_t *dst, int dst_width, uint8_t *src, int src_width,
                            int src_start, int src_incr, int16_t *filters )
{
    int32_t src_pos;
    int16_t phase, i;
    register int32_t sum;
    uint8_t *s;
    int16_t *filter;

    src_pos = src_start;
    for(i=0;i<dst_width;i++) {

        s = src + (src_pos >> POS_FRAC_BITS);
        phase = get_phase(src_pos);
        filter = &filters[phase * NB_TAPS];

        sum = s[0] * filter[0] +
              s[1] * filter[1] +
              s[2] * filter[2] +
              s[3] * filter[3];
              
        sum = sum >> FILTER_BITS;

        // Daten werden hier als int16_t an v_resample bergeben
        // Grenzen werden in v_resample gerundet
        /*
        if( sum & 0x7FFFFF00 ) {
            if (sum < 0) {
                sum = 0;
            } else if (sum > 255){
                sum = 255;
            }
        }
        */

        dst[0] = sum;
        src_pos += src_incr;
        dst++;
    }
}
#endif





/* This function must be optimized */
void FrameRescaler::v_resample(uint8_t *dst, int dst_width, int16_t *src, int wrap, 
                       int16_t *filter, int inc )
{
    int sum, i;
    int16_t *s;

    s = src;
    for(i=0;i<dst_width;i++) {
#if NB_TAPS == 4
        sum = s[0]        * filter[0] +
              s[1 * wrap] * filter[1] +
              s[2 * wrap] * filter[2] +
              s[3 * wrap] * filter[3];
#else
        {
            int j;
            uint8_t *s1 = s;

            sum = 0;
            for(j=0;j<NB_TAPS;j++) {
                sum += s1[0] * filter[j];
                s1 += wrap;
            }
        }
#endif
        sum = sum >> FILTER_BITS;

        // Zuerst prfen, ob sum im Bereich 0-255 liegt - & mit der Bitbaske
        // mu dann 0 sein! sum ist int32_t
        if( sum & 0x7FFFFF00 ) {
            if (sum < 0)
                sum = 0;
            else if (sum > 255)
                sum = 255;
        }
        
        dst[0] = sum;
        dst += inc;
        s++;
    }
}






/* slow version to handle limit cases. Does not need optimisation */
void FrameRescaler::h_resample_slow(int16_t *dst, int dst_width, uint8_t *src, int src_width,
                            int src_start, int src_incr, int16_t *filters, int inc)
{
    int src_pos, phase, sum, j, v, i;
    uint8_t *s, *src_end;
    int16_t *filter;

    src_end = src + src_width * inc;
    src_pos = src_start;
    for(i=0;i<dst_width;i++) {
        s = src + (src_pos >> POS_FRAC_BITS) * inc;
        phase = get_phase(src_pos);
        filter = &filters[phase * NB_TAPS];
        sum = 0;
        for(j=0;j<NB_TAPS;j++) {
            if (s < src)
                v = src[0];
            else if (s >= src_end)
                v = src_end[-1];
            else
                v = s[0];
            sum += v * filter[j];
            s += inc;
        }
        sum = sum >> FILTER_BITS;

        // Daten werden hier als int16_t an v_resample bergeben
        // Grenzen werden in v_resample gerundet
        /*
        if (sum < 0)
            sum = 0;
        else if (sum > 255)
            sum = 255;
        */
        dst[0] = sum;
        src_pos += src_incr;
        dst++;
    }
}






void FrameRescaler::h_resample(int16_t *dst, int dst_width, uint8_t *src, int src_width,
                       int src_start, int src_incr, int16_t *filters, int inc )
{
    int n, src_end;

    if (src_start < 0) {
        n = (0 - src_start + src_incr - 1) / src_incr;
        h_resample_slow(dst, n, src, src_width, src_start, src_incr, filters, inc);
        dst += n;
        dst_width -= n;
        src_start += n * src_incr;
    }
    src_end = src_start + dst_width * src_incr;
    if (src_end > ((src_width - NB_TAPS) << POS_FRAC_BITS) ) {
        n = (((src_width - NB_TAPS + 1) << POS_FRAC_BITS) - 1 - src_start) / src_incr;
    } else {
        n = dst_width;
    }
#if 0
    // MMX ist nicht wirklich schneller als der C-Code!!!
    if ((mm_flags & MM_MMX) && NB_TAPS == 4 && inc == 1)
        h_resample_fast4_mmx(dst, n, src, src_width, src_start, src_incr, filters);
    else
#endif
#if NB_TAPS == 4
    if( NB_TAPS == 4 && inc == 1 )
        h_resample_fast4_inc1(dst, n, src, src_width, src_start, src_incr, filters );
    else
#endif
        h_resample_fast(dst, n, src, src_width, src_start, src_incr, filters, inc);


    if (n < dst_width) {
        dst += n;
        dst_width -= n;
        src_start += n * src_incr;
        h_resample_slow(dst, dst_width, src, src_width, src_start, src_incr, filters, inc);
    }
}




void FrameRescaler::component_resample( uint8_t *output, int owrap, int owidth, int oheight,
                                        uint8_t *input, int iwrap, int iwidth, int iheight, int inc )
{
    int src_y, src_y1, last_src_y, ring_y, phase_y, y1, y;
    int16_t *new_line;
    uint8_t *src_line;

    last_src_y = - FCENTER - 1;
    
    /* position of the bottom of the filter in the source image */
    src_y = (last_src_y + NB_TAPS) * POS_FRAC; 
    ring_y = NB_TAPS; /* position in ring buffer */
    
    for(y=0;y<oheight;y++) {
        /* apply horizontal filter on new lines from input if needed */
        src_y1 = src_y >> POS_FRAC_BITS;
        while (last_src_y < src_y1) {
            if (++ring_y >= LINE_BUF_HEIGHT + NB_TAPS)
                ring_y = NB_TAPS;
                
            last_src_y++;
            /* handle limit conditions : replicate line (slighly
               inefficient because we filter multiple times) */
            y1 = last_src_y;
            if (y1 < 0) {
                y1 = 0;
            } else if (y1 >= iheight) {
                y1 = iheight - 1;
            }
            src_line = input + y1 * iwrap * inc;
            new_line = line_buf + ring_y * owidth;
            
            /* apply filter and handle limit cases correctly */
            h_resample(new_line, owidth, 
                       src_line, iwidth, - FCENTER * POS_FRAC, h_incr, 
                       h_filters, inc );
                       
            /* handle ring buffer wraping */
            if (ring_y >= LINE_BUF_HEIGHT) {
                memcpy( line_buf + (ring_y - LINE_BUF_HEIGHT) * owidth,
                       new_line, owidth * 2);
            }
            
        }
        
        /* apply vertical filter */
        phase_y = get_phase(src_y);
#if 0
        /* desactivated MMX because loss of precision */
        if ((mm_flags & MM_MMX) && NB_TAPS == 4 && inc == 1 && 0)
            v_resample4_mmx(output, owidth, 
                            line_buf + (ring_y - NB_TAPS + 1) * owidth, owidth, 
                            v_filters );
        else
#endif
            v_resample(output, owidth, 
                       line_buf + (ring_y - NB_TAPS + 1) * owidth, owidth, 
                       v_filters, inc );
            
        src_y += v_incr;
        output += owrap * inc;
    }
}
