/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <stdarg.h>
#include <stdio.h>
#include <assert.h>

#include "config.h"
#include "VideoFile.h"
#include "codecplugins.h"
#include "iomodulplugins.h"
#include "debug.h"
#include "translate.h"


bool VideoFile::init = false;
CodecPlugins *VideoFile::AV_Codecs = NULL;
IOModulPlugins *VideoFile::IO_Moduls = NULL;

// VideoFile initialisieren
void VideoFile::Init( void )
{
    if( init )
        return;
        
#ifdef ENABLE_NLS
    setlocale( LC_ALL, "" );
    bindtextdomain( GETTEXT_DOMAIN, LOCALEDIR );
#endif

    AV_Codecs = new CodecPlugins;
    IO_Moduls = new IOModulPlugins;

    init = true;
}



// ein IO-Modul erstellen
VideoFile *VideoFile::Create_VideoFile( const String file, const String typ, IO_Mode mode )
{
    VideoFile *io_handler = NULL;
            
    // Initialisiere Videofile, wenn ntig
    Init();
    
    // Wenn ein Dateityp / IO-Modul Name angegeben ist, benutze diesen
    if( typ != "" ) {
        try {
            io_handler = IO_Moduls->CreateModul( file, typ, mode );
        }
        catch( Errors &error ) {
            // Dateityp wurde nicht erkannt
            if( error == Errors::FileType || error == Errors::IO_Mode ) {
                DEBUG_TXT( error.ErrorText() );
                delete io_handler;
                io_handler = NULL;
            } else if( error == Errors::Open ) {
                ERROR_TXT( error.ErrorText() );
                delete io_handler;
                return NULL;             
            } else if( error != Errors::NoError ) {
                // andere Fehler weiterreichen
                throw;
            }
        }
        return io_handler;
        
    // Automatische Dateityp-Erkennung, nur bei echter Datei
    // und nur beim Lesen       
    } else if( file != "" && mode == read) {

        // Probiere alle Module bis eins passt
        PluginList<IOModul_Info>* moduls = IO_Moduls->IO_Moduls();
        int n = moduls->NumElements();

        for( int i=0; i<n; i++ ) {
            String file_typ = moduls->Element(i).FileTyp();
            DEBUG_TXT( _("Try modul %s"), file_typ.chars() );

            try {
                io_handler = IO_Moduls->CreateModul( file, file_typ, mode );
            }

            catch( Errors &error ) {
                // probiere nchstes Modul, wenn Dateityp nicht erkannt wurden
                if( error == Errors::FileType || error == Errors::IO_Mode ) {
                    DEBUG_TXT( error.ErrorText() );
                    delete io_handler;
                    io_handler = NULL;
                    continue;
                } else if( error == Errors::Open ) {
                    ERROR_TXT( error.ErrorText() );
                    delete io_handler;
                    return NULL;
                } else if( error != Errors::NoError ) {
                    // andere Fehler weiterreichen
                    throw;
                }
            }

            break;
        }

        return io_handler;
    }
    

    return NULL;
}





/**Liste mit IO-Modulen
  */
PluginList<IOModul_Info> *VideoFile::IOModulList( void )
{
    Init();
    return IO_Moduls->IO_Moduls();
}




/**Liste mit Codecs
  */
PluginList<Codec_Info> *VideoFile::AVCodecList( void )
{
    Init();
    return AV_Codecs->AV_Codecs();
}














/*****************************************************************************/
/***** Konstruktor / Destruktor *****/
VideoFile::VideoFile( const String file, const String typ, IO_Mode mode )
{
    if( !init )
        Init();

    // Variablen initialisieren
    iomode = mode;
    iofile = file;
    iomodul = typ;

    v_tracks = 0;
    a_tracks = 0;
}




VideoFile::~VideoFile()
{
    deleteTracks();
}







long long VideoFile::BytesWritten( void )
{
    long long bytes = 0;
    int i, size;

    list<V_Track*>::iterator vtr_iter;
    list<A_Track*>::iterator atr_iter;

    // summiere ber alle Tracks
    // Video
    vtr_iter =  VTrackListe.begin();
    size = VTrackListe.size();
    for( i=0; i < size; i++ ) {
        bytes += (*vtr_iter)->BytesWritten();
        vtr_iter++;
    }

    // Audio
    atr_iter =  ATrackListe.begin();
    size = ATrackListe.size();
    for( i=0; i < size; i++ ) {
        bytes += (*atr_iter)->BytesWritten();
        atr_iter++;
    }

    return bytes;
}







/***** Video-Funktionen *********************************************/
// Video codec initialisieren
// Rckgabewert: false - Fehler
//               true  - Erfolg
V_Codec *VideoFile::initVideoCodec( double Rate, int x, int y, const char *Codec )
{
   
    V_Track *Track = openVideoTrack( Rate, x, y, Codec );

    if( Track != NULL ) {
        v_tracks++;
        return Track->Codec();
    }
        
    return NULL;
}


// Track ffnen
V_Track *VideoFile::openVideoTrack( int track )
{
    //return new V_Track( this, track );
    return NULL;
}



// Track ffnen
V_Track *VideoFile::openVideoTrack( double Rate, int x, int y, const char *Codec )
{
    //return new V_Track( this, track );
    return NULL;
}









/***** Audiofunktionen **********************************************/

// Audio codec initialisieren
A_Codec *VideoFile::initAudioCodec( int Sample_rate, int Channels, const char *Codec )
{
    A_Track *Track = openAudioTrack( Sample_rate, Channels, Codec );
    
    if( Track != NULL ) {
        a_tracks++;
        return Track->Codec();
    }
    
    return NULL;
}


// Track ffnen
A_Track *VideoFile::openAudioTrack( int track )
{
    //return new A_Track( this, track );
    return NULL;
}


// Track ffnen
A_Track *VideoFile::openAudioTrack( int Sample_rate, int Channels, const char *Codec )
{
    //return new A_Track( this, track );
    return NULL;
}








/***** Trackverwaltung ********************************************************/
// Zeiger auf Video Track / Video Codec
V_Codec *VideoFile::VideoCodec( int track )
{
    V_Track *Track;

    Track = getVideoTrack( track );

    // noch kein Track offen
    if( Track == NULL && iomode != write ) {
        Track = openVideoTrack( track );
    }

    if( Track == NULL )
        return NULL;

    return Track->Codec();
}



// Zeiger auf Audio Track / Codec
A_Codec *VideoFile::AudioCodec( int track )
{
    A_Track *Track;

    Track = getAudioTrack( track );

    // noch kein Track offen
    if( Track == NULL && iomode != write ) {
        Track = openAudioTrack( track );
    }

    if( Track == NULL )
        return NULL;

    return Track->Codec();
}




// Zeiger auf Vidoe Track
V_Track *VideoFile::getVideoTrack( int track )
{
    list<V_Track*>::iterator Track;

   if( VTrackListe.empty() )
        return NULL;

    Track = VTrackListe.begin();

    while( (*Track)->Number() != track ) {
        if( Track == VTrackListe.end() )
            return NULL;

        Track++;
    }

    return *Track;
}





// Zeiger auf Audio Track
A_Track *VideoFile::getAudioTrack( int track )
{
    list<A_Track*>::iterator Track;

    if( ATrackListe.empty() )
        return NULL;

    Track = ATrackListe.begin();

    while( (*Track)->Number() != track ) {
        if( Track == ATrackListe.end() )
            return NULL;

        Track++;
    }

    return *Track;
}





// Track in Liste einreihen
void VideoFile::addTrack( V_Track *track )
{
    if( track != NULL )
        VTrackListe.push_back( track );
}



// Track in Liste einreihen
void VideoFile::addTrack( A_Track *track )
{
    if( track != NULL )
        ATrackListe.push_back( track );

}





// Track aus Liste entfernen
void VideoFile::removeTrack( V_Track *track )
{
    list<V_Track*>::iterator Track;

    if( track == NULL )
        return;

    if( VTrackListe.empty() )
        return;

    Track = VTrackListe.begin();

    while( (*Track) != track ) {
        if( Track == VTrackListe.end() )
            return;

        Track++;
    }

    VTrackListe.erase( Track );
}




// Track aus Liste entfernen
void VideoFile::removeTrack( A_Track *track )
{
    list<A_Track*>::iterator Track;

    if( track == NULL )
        return;

    if( ATrackListe.empty() )
        return;

    Track = ATrackListe.begin();

    while( (*Track) != track ) {
        if( Track == ATrackListe.end() )
            return;

        Track++;
    }

    ATrackListe.erase( Track );
}





// Tracks schlieen
void VideoFile::deleteTracks( void )
{
    // Videotracks vernichten
    while( !VTrackListe.empty() ) {
        delete *VTrackListe.begin();
    }

    // Audiotracks vernichten
    while( !ATrackListe.empty() ) {
        A_Track *tr;
        tr = *ATrackListe.begin();
        delete tr;
    }
}




/***** Hilfsfunktionen **********************************************/
// Fehler ausgeben
/*
void VideoFile::Fehler( const char *fmt, ... )
{
    char error_Text[ 4000 ];
    va_list var;
    int i;

    va_start( var, fmt );
    i = vsprintf( error_Text, fmt, var );

    fprintf( stderr, error_Text );
    // assert( false );
    // exit( 1 );

    return;
}
*/
