/***************************************************************************
                          VideoFrame.h  -  description
                             -------------------
    begin                : Son Dez 15 2002
    copyright            : (C) 2002 by Christian Tusche
    email                : ctusche@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef VIDEOFRAME_H
#define VIDEOFRAME_H

#include <inttypes.h>

#include "FrameData.h"
#include "Convert.h"
#include "FrameRescaler.h"


/**Die Klasse VideoFrame speichert die Daten eines Bildes unabhngig von dessen<br>
  *Farbraum und Gre und konvertiert diese bei Bedarf fr die Ausgabe
  *@author Christian Tusche
  */

class VideoFrame {
public:
    /**Neuen Videoframe mit der Gre<br>
      *  x * y<br>
      *erstellen<br>
      *Es wird ein Framebuffer fr den Eingabeframe angelegt, oder der
      *Speicherbereich *memseg mit der Gre size als Eingangs-Framebuffer verwendet.
      */
    VideoFrame( int x, int y, uint8_t *memseg = NULL, size_t size = 0 );

    ~VideoFrame();

    
    /**Bilddaten in Frame speichern - lineare Framebuffer<br>
      *Ausgabe-Daten aus *bild werden als Quellbild bernommen. Dabei wird
      *die Ausgabe von *bild verndert, und mu bei erneutem Zugrif auf
      **bild neu berechnet werden.
      */
    void SetFrame( VideoFrame *bild );

    /**Bilddaten in Frame speichern - lineare Framebuffer<br>
      *Daten aus *bild werden in den internen Framebuffer kopiert!
      */
    void SetFrame( uint8_t *bild );

    /**Bilddaten in Frame speichern - planar<br>
      *Daten aus den yuv-Zeigern werden in den internen Framebuffer kopiert!
      */
    void SetFrame( uint8_t *y, uint8_t *u, uint8_t *v );

    /**Zeiger auf Datenfeld - linearer Framebuffer<br>
      *Rckgabewert: Gre der Bilddaten
      */
    int FramePointer( uint8_t *&bild );

    /**Zeiger auf Datenfeld - planar<br>
      *Rckgabewert: Gre der Bilddaten der Y-Komponente
      */
    int FramePointer( uint8_t *&y, uint8_t *&u, uint8_t *&v );


    
    /**Zeiger auf Ausgabebild - linearer Framebuffer<br>
      *Rckgabewert: Gre der Bilddaten<br>
      *eigentlich sollten hier "const uint8_t *" stehen, aber einige Codecs
      *(die meisten) verwenden keine const-Datenfelder fr ihre Eingabe. :(<br>
      *Die Daten drfen aber trotzden nicht verndert werden!!!
      */
    int GetFrame( uint8_t *&bild );

    /**Zeiger auf Ausgabebild - planar<br>
      *Rckgabewert: Gre der Bilddaten Y+U+V
      */
    int GetFrame( uint8_t *&y, uint8_t *&u, uint8_t *&v );

    
    
    /**Legt den Bildausschnitt fest, der ausgegeben werden soll<br>
      *Erste Ecke des Rechtecks:  x1, y1<br>
      *Zweite Ecke des Rechtecks: x2, y2<br>
      *Die Koordinaten werden von der oberen linken Ecke aus gezhlt.<br>
      *Nach einem Aufruf von CropBox() ist die Ausgabegre auf die Dimension
      *der Box zurckgesetzt.
      */
    void CropBox( int x1, int y1, int x2, int y2 );

    /**Skalierte Gre des Ausgabebildes festlegen
      */
    void SetOutSize( int x, int y );

    /**Liefert die Gre des auszugebenden Bildes
      */
    void OutSize( int &x, int &y );

    /**Liefert die Gre des ursprnglichen Bildes
      */
    void Size( int &x, int &y );

    /**Setzt gewnschten Farbraum fr die Bildausgabe
      */
    void OutColorModel( int cmodel );

    /**Farbraum der Bilddaten abfragen
      */
    int ColorModel( void );

    /**Farbraum der Bilddaten festlegen
      */
    void ColorModel( int cmodel );


private:
    /**erzeuge das Ausgabebild - Farbraumkonvertierung/ausschneiden
      */
    void ausgabe_erzeugen( void );

    /**Rechteck ausschneiden
      */
    void crop_frame( FrameData *in, FrameData *out );


    /**Umwandlungen einrichten
      */
    void setup_crop( void );
    void setup_scale( void );
    void setup_cmodel( void );
    
    // Daten des Bildes
    FrameData *bilddaten;
    
    // Daten nach Ausschneiden
    FrameData *crop_out;
    // Daten nach skalieren
    FrameData *scale_out;
    // Daten nach Farbraumkonvertierung
    FrameData *cmodel_out;
    
    // Zeiger auf Ausgabebild
    FrameData *ausgabebild;

    
    Convert *cmodel_converter;  // konvertiert Farbraum
    FrameRescaler *re_scaler;   // Skalierung des Bildes auf Ausgabegre
  
    bool ausgabe_ok;                // wurde die Ausgabe schon verarbeitet?

    // Koordinaten des Bildausschnittes
    int box_x1, box_x2, box_y1, box_y2;

    // Gre der Bildausgabe
    int out_breite, out_hoehe;

    // Ausgabe-Farbraum
    int out_cmodel;
};

#endif
