/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

/* Video-Digitalisierer */



#include <stdio.h>
#include <unistd.h>
#include <termios.h>
#include <sys/time.h>
#include <sys/types.h>
#include <assert.h>

#include "debug.h"
#include "grab_threads/v4l.h"
#include "grab_threads/OSS.h"
#include "grab_threads/Sync.h"
#include "VideoFile/VideoFile.h"
#include "CmdlineInterface/ausgabedatei.h"
#include "CmdlineInterface/codec_optionen.h"
#include "CmdlineInterface/pluginliste.h"

#define GETTEXT_DOMAIN "av_rec"
#include "VideoFile/translate.h"

unsigned char *buf;

AusgabeDatei *Ausgabe = NULL;
V_Codec *Video_Ausgabe = NULL;
A_Codec *Audio_Ausgabe = NULL;

static Codec_Optionen A_Optionen, V_Optionen;

Video_th *v_thread = NULL;
Audio_th *a_thread = NULL;
Sync_th *sync_thread = NULL;


// Optionen
long laenge = 0;

// Video-Optionen
int interpolate = 0;
int deinterlace = 0;
unsigned char *previous_Y = NULL;
int w=640, h=480;
double rate=25.0;
int quelle = 0;

// Audio-Einstellungen
int sound_channels = 2;
int sound_bits = 16;
int sound_rate = 44100;
int monitor_audio = 0;




struct termios orig_stdin;

static inline int get_taste(void);
static void init_video_grab( void );
static inline int video_grab( void );
void exit_func( void );
void options(int argc, char *argv[], char *envp[]);

void YUV420P_h_interpolate_frame( VideoFrame *frame );
void YUV420P_h_interpolate_time( VideoFrame *frame );
void YUV420P_deinterlace(VideoFrame *frame,
        int SpatialTolerance, int TemporalTolerance, int mode);





int main( int argc, char *argv[], char *envp[] )
{
    struct termios neu_stdin;

    // bersetzungen aktivieren
    setlocale( LC_ALL, "" );
    textdomain( GETTEXT_DOMAIN );
    bindtextdomain( GETTEXT_DOMAIN, LOCALEDIR );
    bindtextdomain( "av_convert", LOCALEDIR );

    // Terminal initialisieren
    tcgetattr( STDIN_FILENO, &orig_stdin);
    neu_stdin = orig_stdin;
    neu_stdin.c_lflag &= ~(ICANON|ECHO);
    tcsetattr( STDIN_FILENO, TCSANOW, &neu_stdin );

    atexit( &exit_func );

    // Optionen verarbeiten
    options( argc, argv, envp );

    init_video_grab();

    while( video_grab() < laenge || laenge == 0 ) {
        if( get_taste() == 32 )
            break;
    }

    return 0;
}







int get_taste(void)
{
    fd_set eingang;
    struct timeval zeit;

    FD_ZERO (&eingang);
    FD_SET (STDIN_FILENO, &eingang);
    zeit.tv_sec=0;
    zeit.tv_usec=0;

    fflush(stdout);
    if ( select (FD_SETSIZE, &eingang, NULL, NULL, &zeit) )
    {
        return getchar();
    }else{
        return EOF;
    }
}









/* Aufzeichnung vorbereiten und Threads starten */
void init_video_grab( void )
{

    // neuen Video-Thread erzeugen
    v_thread = new v4l_th;
    v_thread->Size( w, h) ;
    v_thread->Rate( rate );
    v_thread->Source( quelle );             // Signalquelle
    v_thread->cModel( YUV420P );            // Farbformat
//    v_thread->cModel( RGB24 );            // Farbformat

    // Audio Thread initialisieren
    a_thread = new OSS_th;
    a_thread->Channels( sound_channels );
    a_thread->Bits( sound_bits );
    a_thread->Rate( sound_rate );
    a_thread->MonitorAudio( monitor_audio );



    // Videodatei ffnen
    if( !Ausgabe->open( rate, w, h, sound_rate, sound_channels ) ) {
        fprintf( stderr, _("Error: Unable to open output file!\n"));
        exit( 1 );
    }


    Video_Ausgabe = Ausgabe->Video();
    Audio_Ausgabe = Ausgabe->Audio();

    if( Video_Ausgabe == NULL ) {
        fprintf( stderr, _("Error: Unable to open video codec!\n"));
        exit( 1 );
    }
    
    Video_Ausgabe->ColorModel( YUV420P );

    previous_Y = new unsigned char[ w * h ];
    
    /* Threads starten */
    fprintf( stderr, _("starting threads...") );
    sync_thread = new Sync_th;
    v_thread->starte_Thread();
    a_thread->starte_Thread();
    fprintf( stderr, _(" done\n") );    

}









/* Arbeitsschleife - eingelesene Daten in Datei speichern */
static inline int video_grab( void )
{
    int picnr, dat_len;
    VideoFrame *inbuf;
    static int n = 0;
    int16_t **a_buf;

    /***** Videodaten verarbeiten *****/
    inbuf = v_thread->get();
    picnr = v_thread->get_picnr();

    if( picnr == 0 )
        n = 0;

    if( deinterlace ){
        YUV420P_deinterlace( inbuf, 10, 10, 1 );
    }
    
    if( interpolate == 1 ) {
        YUV420P_h_interpolate_time( inbuf );
    } else if( interpolate == 2 ) {
        YUV420P_h_interpolate_frame( inbuf );
    }
    


     // verlorene Bilder auffllen
    while( n < picnr ) {
        if( !Video_Ausgabe->CopyFrame() )
            Video_Ausgabe->encodeVideo( inbuf );
            
        fprintf( stderr, "\n--> [%i]\n", n );
        n++;
    }

    Video_Ausgabe->encodeVideo( inbuf );

    if( !(n % 10) )
        fprintf( stderr, _("encode frame [%i]\r"), n );

    n++;

    // Bild-Puffer freigeben
    v_thread->release();



    /***** Audiodaten vrearbeiten ******/
    while( (dat_len = a_thread->get( a_buf )) > 0 ) {
        if( Audio_Ausgabe != NULL )
            Audio_Ausgabe->encodeAudio( a_buf, dat_len );

        a_thread->release();
    }


    return picnr;

}







void exit_func( void )
{
    // Aufrumen
    if( Ausgabe != NULL ) {
        fprintf( stderr, _("\nclose files...\n") );
        delete Ausgabe;
    }

    if( v_thread != NULL ) {
        fprintf( stderr, _("stop video thread...\n") );
        delete v_thread;
    }

    if( a_thread != NULL ) {
        fprintf( stderr, _("stop sound thread...\n") );
        delete a_thread;
    }

    if( sync_thread != NULL ) {
        fprintf( stderr, _("stop sync thread...\n") );
        delete sync_thread;
    }
    
    printf( _("done!\n") );
    
    delete [] previous_Y;

    tcsetattr( STDIN_FILENO, TCSANOW, &orig_stdin );
}






/****************************************************************************************/
/***** Optionen einlesen                                                            *****/
/****************************************************************************************/
void options(int argc, char *argv[], char *envp[])
{
    int optchar;
    bool ausgabe=false;

    while(( optchar = getopt(argc, argv, "v:a:o:V:A:r:x:y:q:R:C:B:mc:i:dh:")) != -1 ) {
        switch (optchar) {
        case 'V':
            V_Optionen.SetArg( optarg );
            break;
        case 'A':
            A_Optionen.SetArg( optarg );
            break;
        case 'o':
            if( Ausgabe == NULL ) {
                Ausgabe = new AusgabeDatei;
                Ausgabe->parseArg( optarg );
                ausgabe = true;
            }
            break;

        case 'r':
            rate = strtod( optarg, NULL );
            break;

        case 'x':
            w = atoi( optarg );
            break;
        case 'y':
            h = atoi( optarg );
            break;
        case 'q':
            quelle = atoi( optarg );
            break;

        case 'R':
            sound_rate = atoi( optarg );
            break;
        case 'C':
            sound_channels = atoi( optarg );
            assert( sound_channels == 1 || sound_channels == 2 );
            break;
        case 'B':
            sound_bits = atoi( optarg );
            assert( sound_bits == 8 || sound_bits == 16 );
            break;

        case 'm':
            monitor_audio = 1;
            break;
        case 'c':
            laenge = strtol( optarg, NULL, 0 );
            break;
        case 'i':
            interpolate = atoi( optarg );
            break;
        case 'd':
            deinterlace = 1;
            break;    

        case 'h':
            if( strcmp( optarg, "IO-Modules") == 0 ) {
                Pluginliste::Print_IO_Moduls();
                break;
            }
            if( strcmp( optarg, "A-Codecs") == 0 ) {
                Pluginliste::Print_A_Codecs();
                break;
            }
            if( strcmp( optarg, "V-Codecs") == 0 ) {
                Pluginliste::Print_V_Codecs();
                break;
            }
        default:
            printf( _("Syntax: %s [options] -o <fmt::file> \n"), argv[0] );
            printf( "\n" );
            printf( _("   -v <device>               : video input device\n") );
            printf( _("   -a <Device>               : audio input device\n") );
            printf( _("   -o <fmt::file>            : output file\n") );
            printf( _("                               fmt: IO-modul\n") );
            printf( _("   -h IO-Modules             : list IO-modules\n") );
            printf( "\n" );
            printf( _("   -V <codec>                : set video codec\n") );
            printf( _("   -V <option=value>         : set codec options\n") );
            printf( _("      -V DIVX -V Bitrate=800 -V MinQuantisation=3 -V ...\n") );
            printf( _("   -h V-Codecs               : list video codecs\n") );
            printf( "\n" );
            printf( _("   -A <codec>                : set audio codec\n") );
            printf( _("   -A <option=value>         : set codec options\n") );
            printf( _("   -h A-Codecs               : list audio codecs\n") );
            printf( "\n" );
            printf( _("   -r n              : video frame rate\n") );
            printf( _("   -x n              : video width\n") );
            printf( _("   -y n              : video hight\n") );
            printf( _("   -q n              : video signal source [0=tuner, 1=composite, 2=S-video]\n") );
            printf( _("   -R n              : audio sample rate\n") );
            printf( _("   -C n              : audio channels [1, 2]\n") );
            printf( _("   -B n              : audio bits [8, 16]\n") );
            printf( _("   -m                : monitor audio\n") );
            printf( "\n" );
            printf( _("   -c n              : number of frames\n") );
            printf( _("   -i n              : interpolate n=1 : temporal interpolation\n") );
            printf( _("                                   n=2 : line interpolation\n") );
            printf( _("   -d                : deinterlace\n") );
            exit(0);
        }
    }

    // Eingabedatei und Ausgabedatei gesetzt
    if( !ausgabe ) {
        fprintf( stderr, _("Help: %s -h\n"), argv[0] );
        exit(0);
    }

    // Codec fr Audio und Video-Ausgabe festlegen
    if( Ausgabe != NULL ) {
        Ausgabe->SetVCodec( &V_Optionen );
        Ausgabe->SetACodec( &A_Optionen );
    }
}








/*******************************************************************************/
/***** Filterfunktionen ********************************************************/
/*******************************************************************************/
void YUV420P_h_interpolate_frame( VideoFrame *frame )
{
    unsigned char *yuv[3];
    unsigned char *H1, *H2;
    //unsigned char *Z;
    unsigned char P;
    int w=0, h=0;

    if( frame->ColorModel() != YUV420P )
        return;

    frame->Size( w, h );
    if( frame->FramePointer( yuv[0], yuv[1], yuv[2] ) <= 0 )
        return;

    //yuv[0] = frame;                                             // y - Plane
    //yuv[1] = yuv[0] + w * h;                                    // u - Plane
    //yuv[2] = yuv[1] + w/2 * ((h + 1)/2) ;                       // v - Plane

    // y - Plane interpolieren
    for( int i=0; i < h-1; i++ ) {
	//Z = yuv[0] + i*w;
	H1 = yuv[0] + i*w;
	for( int j=0; j < w; j++ ) {
	//for( H1 = Z; H1 < Z + w; H1++ ) {
	    //H1 = Z + j;
	    H2 = H1 + w;
	    
	    P = (*H2 + *H1) >> 1;
	    *H1 = P;
	    
	    H1++;
	}
    }
    
}






void YUV420P_h_interpolate_time( VideoFrame *frame )
{
    unsigned char *yuv[3];
    unsigned char *H1, *H2;
    //unsigned char *Z1, *Z2;
    //unsigned char *N1, *N2;
    //int P;
    unsigned char P;
    int w=0, h=0;

    if( frame->ColorModel() != YUV420P )
        return;

    frame->Size( w, h );
    if( frame->FramePointer( yuv[0], yuv[1], yuv[2] ) <= 0 )
        return;


    if( previous_Y == NULL ) {
        previous_Y = new unsigned char[ w * h ];
        memcpy( previous_Y, yuv[0], w * h );
    }
    
    //yuv[0] = frame;                                             // y - Plane
    //yuv[1] = yuv[0] + w * h;                                    // u - Plane
    //yuv[2] = yuv[1] + w/2 * ((h + 1)/2) ;                       // v - Plane


    // y - Plane interpolieren
    for( int i=0; i < h; i++ ) {
	//Z1 = yuv[0] + i*w;
	//Z2 = previous_Y + i*w;
	H1 = yuv[0] + i*w;
	H2 = previous_Y + i*w;
	for( int j=0; j < w; j++ ) {
	    //H1 = Z1 + j;
	    //H2 = Z2 + j;

	    //P = ( *H2 - *H1 );	    
	    //if( P < 10 && P > -10 ) {
	    if( (*H1 ^ *H2) < 8 ) {
	    
		//N1 = H1 + w;
		//N2 = H2 + w;
		
		//P = ( *H2 + *H1 + *N1 + *N2 ) >> 2;
		P = ( *H2 + *H1 ) >> 1;
		*H2 = *H1;
		*H1 = P;

	    } else {
		*H2 = *H1;
	    }
	    H1++;
	    H2++;

	}
    }
    
    // memcpy( previous_Y, yuv[0], w*h );

}






/****************************************************************************************/
/***** Deinterlacer                                                                 *****/
/****************************************************************************************/
void YUV420P_deinterlace(VideoFrame *frame,
        int SpatialTolerance, int TemporalTolerance, int mode)
{
    int i,j,weave;
    int cnt = 0;
    int width = 0, height = 0;

    unsigned char *M0, *T0, *b0, *M1, *T1, *B1, *frame_buf;

    if( frame->ColorModel() != YUV420P )
        return;

    frame->Size( width, height );
    if( frame->FramePointer( frame_buf ) <= 0 )
        return;


    if( previous_Y == NULL ) {
        previous_Y = new unsigned char[ width * height ];
        memcpy( previous_Y, frame_buf, width * height );
    }

    switch (mode) {
    case 1:
        for (i=1; i<height-1; i+=2) {
            M1 = frame_buf + i*width;
            T1 = M1 - width;
            B1 = M1 + width;
            M0 = previous_Y + i*width;
            T0 = M0 - width;
            b0 = M0 + width;
            for (j=0; j<width; j++, M0++, T0++, b0++, M1++, T1++, B1++) {
                register int f;
                f = *T1 - *M1;
                f *= f;
                weave = (f < SpatialTolerance);
                if (!weave) {
                    f = *B1 - *M1;
                    f *= f;
                    weave = (f < SpatialTolerance);
                    if (!weave) {
                        f = *M0 - *M1;
                        f *= f;
                        weave = (f < TemporalTolerance);
                        f = *T0 - *T1;
                        f *= f;
                        weave = (f < TemporalTolerance) && weave;
                        f = *b0 - *B1;
                        f *= f;
                        weave = (f < TemporalTolerance) && weave;
                        if (!weave) {
                            f = *T1 + *B1;
                            *M1 = f>>1;
                            cnt++;
                        }
                    }
                }
            }
        }

        break;
    case 2:
        for (i=1; i<height-1; i+=2) {
            M1 = frame_buf + i*width;
            T1 = M1 - width;
            B1 = M1 + width;
            for (j=0; j<width; j++, M1++, T1++, B1++) {
                register int f = *T1 + *B1;
                *M1 = f>>1;
            }
        }
        break;
    }
    /* just copy the last line */
    M1 = frame_buf + (height - 1)*width;
    T1 = M1 - width;
    memcpy(M1,T1,width);

    memcpy( previous_Y, frame_buf, width*height );
}
