/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/


/********************************************************************\
*****   lesen und schreiben in Avi Dateien                       *****
*****   architekturunabhngige Methoden, um int/uint - Werte     *****
*****   in Avi-Dateien zu schreiben bzw. zu lesen                *****
\********************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "RIFF_File.h"
#include "VideoFile/Errors.h"
     


RIFF_File::RIFF_File( const char *datei, bool io_read )
{
    rd = true;
    file = 0;
    
    if( io_read ) {
        rd = true;

        file = open( datei, O_RDONLY );
        if( file < 0 ) {
            Errors errorstat = Errors::Open;
            throw( errorstat );
        }
        
        // Dateigre feststellen
        total_bytes = lseek( file, 0, SEEK_END );
        lseek( file, 0, SEEK_SET );
        
    } else {
        rd = false;

        file = open( datei, O_WRONLY | O_TRUNC | O_CREAT, 0644 );
        if( file < 0 ) {
            Errors errorstat = Errors::Open;
            throw( errorstat );
        }
        
        total_bytes = 0;
    }
}







RIFF_File::~RIFF_File()
{
    if( file > 0 )
        close( file );
}







/**uint64_t lesen
  */
bool RIFF_File::read( uint64_t &x )
{
    if( !rd )
        return false;

    uint8_t Z[8];

    if( !read( Z, 8 ) )
        return false;

    read_int( x, Z );

    return true;
}




/**uint32_t lesen
  */
bool RIFF_File::read( uint32_t &x )
{
    if( !rd )
        return false;

    uint8_t Z[4];

    if( !read( Z, 4 ) )
        return false;
    
    read_int( x, Z );

    return true;
}





/**uint16_t lesen
  */
bool RIFF_File::read( uint16_t &x )
{
    if( !rd )
        return false;

    uint8_t Z[2];

    if( !read( Z, 2 ) )
        return false;

    read_int( x, Z );

    return true;
}



/**uint8_t lesen
  */
bool RIFF_File::read( uint8_t &x )
{
    if( !rd )
        return false;

    if( !read( &x, 1 ) )
        return false;

    return true;
}






/**uint64_t schreiben
  */
bool RIFF_File::write( uint64_t x )
{
    if( rd )
        return false;

    uint8_t  Z[8];
    write_int( x, Z );

    if( write( Z, 8 ) )
        return true;

    return false;
}




/**uint32_t schreiben
  */
bool RIFF_File::write( uint32_t x )
{
    if( rd )
        return false;

    uint8_t  Z[4];
    write_int( x, Z );

    if( write( Z, 4 ) )
        return true;
        
    return false;   
}





/**uint16_t schreiben
  */
bool RIFF_File::write( uint16_t x )
{
    if( rd )
        return false;

    uint8_t Z[2];
    write_int( x, Z );

    if( write( Z, 2 ) )
        return true;

    return false;
}




/**uint8_t schreiben
  */
bool RIFF_File::write( uint8_t x )
{
    if( rd )
        return false;

    if( write( &x, 1 ) )
        return true;

    return false;
}






/**Chunk-Header lesen - liest <br>
  *   uint32_t Typ<br>
  *      und<br>
  *   uint32_t Length
  */
bool RIFF_File::read_chunk_header( uint32_t &type, uint32_t &length )
{
    if( !read( type ) )
        return false;

    if( !read( length ) )
        return false;

    return true;
}





/**n Bytes lesen
  */
bool RIFF_File::read( uint8_t *buf, size_t n )
{
    if( !rd )
        return false;

    if( ::read( file, buf, n ) == (ssize_t)n )
        return true;

    return false;
}






/**n Bytes schreiben
  */
bool RIFF_File::write( uint8_t *buf, size_t n )
{
    if( rd )
        return false;
        
    if( ::write( file, buf, n ) == (ssize_t)n )
        return true;

    return false;
}






/**Bytes bis zum Dateiende
  */
off_t RIFF_File::bytes_left(void)
{
    return total_bytes - lseek64( file, 0, SEEK_CUR );
}







/**Bytes berspringen
  */
bool RIFF_File::skip( off_t bytes )
{
    lseek64( file, bytes, SEEK_CUR );

    return true;
}



/**Datei positionieren
  */
bool RIFF_File::seek( off_t offset, int mode )
{
    lseek64( file, offset, mode );

    return true;
}





/**Datei-Position
  */
off_t RIFF_File::pos( void )
{
    return lseek64( file, 0, SEEK_CUR );
}
