/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                                                **
**                                                                           **
** Based on QT4Linux by Adam Williams.                                       **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <stdio.h>


#include "Ima4Codec.h"
#include "debug.h"
#include "VideoFile/translate.h"



// Known by divine revelation 
#define BLOCK_SIZE 0x22
#define SAMPLES_PER_BLOCK 0x40



/***** Konstruktor / Destruktor *****/
Ima4Codec::Ima4Codec( A_Track *Track )
:A_Codec( Track )
{
    // Variablen initialisieren
    puffer = NULL;
    puffer_gr = 0;
    samples_in_buffer = 0;
    

    // Allocate a work buffer
    work_buffer = new int16_t*[Channels()];
    for( int i=0; i < Channels(); i++ ) 
        work_buffer[i] = new int16_t[SAMPLES_PER_BLOCK];

    // Allocate last sample buffer
    last_samples = new int[Channels()];
    
    for( int i=0; i < Channels(); i++ ) 
        last_samples[i] = 0;

    // Allocate last index buffer
    last_indexes = new int[Channels()];
    for( int i=0; i < Channels(); i++ )
        last_indexes[i] = 0;

    // Information ber die Bitrate und Blockgre des Codecs
    SetBitsPerSecond( (double)(8 * Channels() * SampleRate() * BLOCK_SIZE ) / SAMPLES_PER_BLOCK );
    SetFrameSamples( SAMPLES_PER_BLOCK );
}




Ima4Codec::~Ima4Codec()
{
    for( int i=0; i < Channels(); i++ )
        delete [] work_buffer[i];
        
    delete [] work_buffer;
    delete [] last_samples;
    delete [] last_indexes;
                
    delete [] puffer;
}






// Audiodaten schreiben
int Ima4Codec::encode( float **daten, int samples )
{
    // + 1 Block vom letzten Aufruf
    long bytes = (samples/SAMPLES_PER_BLOCK + 1) * Channels() * BLOCK_SIZE;
    int16_t sample;
    int channels = Channels();
    int i, j, k, start_sample = samples_in_buffer;

    
    // Puffer anfordern
    if( bytes > puffer_gr ) {
        delete [] puffer;
        puffer = new unsigned char[bytes];
        puffer_gr = bytes;
    }
    
    unsigned char *output_ptr = puffer;
    bytes = 0;
    
    // Samples in Blcken codieren
    // beginne mit restlichen Daten vom letzten Aufruf
    if( samples_in_buffer > 0 ) {
        for(j = 0; j < channels; j++) {
            for(k=0; k < SAMPLES_PER_BLOCK - samples_in_buffer && k < samples; k++) {
                work_buffer[j][k+samples_in_buffer] = (int16_t)( daten[j][k] * (1<<15) );
            }
            
            if( k + samples_in_buffer == SAMPLES_PER_BLOCK ) {
                ima4_encode_block( output_ptr, work_buffer[j], j );
                output_ptr += BLOCK_SIZE;
                bytes += BLOCK_SIZE;
            }
        }
        
        samples_in_buffer += k;
        if( samples_in_buffer == SAMPLES_PER_BLOCK )
            samples_in_buffer = 0;
        else
            return 1;
    }
    
    // Blcke codieren
    for( i = start_sample; i + SAMPLES_PER_BLOCK < samples; i += SAMPLES_PER_BLOCK ) {
        for( j = 0; j < channels; j++ ) {
            for( k=0; k < SAMPLES_PER_BLOCK; k++ ) {
                work_buffer[j][k] = (int16_t)( daten[j][k + i] * (1<<15) );
            }
            
            ima4_encode_block( output_ptr, work_buffer[j], j );

            output_ptr += BLOCK_SIZE;
            bytes += BLOCK_SIZE;
        }
    }
    
    // restliche Samples im Puffer speichern
    if( i < samples ) {
        for(j = 0; j < channels; j++) {
            for(k=0; k + i < samples; k++) {
                work_buffer[j][k] = (int16_t)( daten[j][k + i] * (1<<15) );
            }
        }
        samples_in_buffer = k;
    }

    
    // Puffer in Datei schreiben
    return track->writeData( puffer, bytes, bytes / (channels * BLOCK_SIZE) * SAMPLES_PER_BLOCK );
}






// Audiodaten lesen
int Ima4Codec::decode( float **daten, int samples )
{
    long bytes = (Channels() * samples * BLOCK_SIZE) / SAMPLES_PER_BLOCK;
    int16_t sample;
    int channels = Channels();
    int i,j,k;

    // Puffer anfordern
    if( bytes > puffer_gr ) {
        delete [] puffer;
        puffer = new unsigned char[bytes];
        puffer_gr = bytes;
    }
    
    unsigned char* block_ptr = puffer;

    // Puffer aus Datei lesen
    if( !track->readBytes( puffer, bytes ) )
        return 0;

    // Datenblcke decodieren
    for(i = 0; i < samples; i += SAMPLES_PER_BLOCK) {
        for(j=0; j < channels; j++) {
            ima4_decode_block( work_buffer[j], block_ptr );
            block_ptr += BLOCK_SIZE;
            
            for(k=0; k<SAMPLES_PER_BLOCK; k++) {
                sample = work_buffer[j][k];
                daten[j][k + i] = (float)sample / (float)(1<<15);
            }
        }
    }
    
    return 1;
}





















const int Ima4Codec::ima4_step[89] = 
{
    7, 8, 9, 10, 11, 12, 13, 14, 16, 17,
    19, 21, 23, 25, 28, 31, 34, 37, 41, 45,
    50, 55, 60, 66, 73, 80, 88, 97, 107, 118,
    130, 143, 157, 173, 190, 209, 230, 253, 279, 307,
    337, 371, 408, 449, 494, 544, 598, 658, 724, 796,
    876, 963, 1060, 1166, 1282, 1411, 1552, 1707, 1878, 2066,
    2272, 2499, 2749, 3024, 3327, 3660, 4026, 4428, 4871, 5358,
    5894, 6484, 7132, 7845, 8630, 9493, 10442, 11487, 12635, 13899,
    15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794, 32767
};

const int Ima4Codec::ima4_index[16] = 
{
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8
};






void Ima4Codec::ima4_decode_sample(int *predictor, int *nibble, int *index, int *step)
{
	int difference, sign;

// Get new index value 
	*index += ima4_index[*nibble];

	if(*index < 0) *index = 0; 
	else 
	if(*index > 88) *index = 88;

// Get sign and magnitude from *nibble 
	sign = *nibble & 8;
	*nibble = *nibble & 7;

// Get difference 
	difference = *step >> 3;
	if(*nibble & 4) difference += *step;
	if(*nibble & 2) difference += *step >> 1;
	if(*nibble & 1) difference += *step >> 2;

// Predict value 
	if(sign) 
	*predictor -= difference;
	else 
	*predictor += difference;

	if(*predictor > 32767) *predictor = 32767;
	else
	if(*predictor < -32768) *predictor = -32768;

// Update the step value 
	*step = ima4_step[*index];
}






void Ima4Codec::ima4_decode_block( int16_t *output, uint8_t *input )
{
	int predictor, index, step;
	int nibble, nibble_count;
	uint8_t *input_end = input + BLOCK_SIZE;

        // Get the chunk header 
	predictor = *input++ << 8;
	predictor |= *input++;

	index = predictor & 0x7f;
	if(index > 88) index = 88;

	predictor &= 0xff80;
	if(predictor & 0x8000) predictor -= 0x10000;
	step = ima4_step[index];

        // Read the input buffer sequentially, one nibble at a time 
	nibble_count = 0;
	while(input < input_end)
	{
		nibble = nibble_count ? (*input++  >> 4) & 0x0f : *input & 0x0f;

		ima4_decode_sample(&predictor, &nibble, &index, &step);
		*output++ = predictor;

		nibble_count ^= 1;
	}
}


void Ima4Codec::ima4_encode_sample(int *last_sample, int *last_index, int *nibble, int next_sample)
{
	int difference, new_difference, mask, step;

	difference = next_sample - *last_sample;
	*nibble = 0;
	step = ima4_step[*last_index];
	new_difference = step >> 3;

	if(difference < 0)
	{
		*nibble = 8;
		difference = -difference;
	}

	mask = 4;
	while(mask)
	{
		if(difference >= step)
		{
			*nibble |= mask;
			difference -= step;
			new_difference += step;
		}

		step >>= 1;
		mask >>= 1;
	}

	if(*nibble & 8)
		*last_sample -= new_difference;
	else
		*last_sample += new_difference;

	if(*last_sample > 32767) *last_sample = 32767;
	else
	if(*last_sample < -32767) *last_sample = -32767;

	*last_index += ima4_index[*nibble];

	if(*last_index < 0) *last_index = 0;
	else
	if(*last_index > 88) *last_index= 88;
}


void Ima4Codec::ima4_encode_block( unsigned char *output, int16_t *input, int channel )
{
    int i, nibble_count = 0, nibble, header;

    // Get a fake starting sample 
    header = last_samples[channel];
    
    // Force rounding. 
    if(header < 0x7fc0)
        header += 0x40;
    
    if(header < 0)
        header += 0x10000;
    
    header &= 0xff80;
    *output++ = (header & 0xff00) >> 8;
    *output++ = (header & 0x80) + (last_indexes[channel] & 0x7f);

    for(i = 0; i < SAMPLES_PER_BLOCK; i++) {
        ima4_encode_sample(&(last_samples[channel]), 
                           &(last_indexes[channel]), 
                           &nibble, 
                           *input);

        if(nibble_count)
            *output++ |= (nibble << 4);
        else
            *output = nibble;

        input += 1;
        nibble_count ^= 1;
    }
}
