/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <stdio.h>
#include <assert.h>

#include "LameCodec.h"
#include "debug.h"
#include "VideoFile/translate.h"
#include "plugins/RIFFio/Wave_Format.h"
#include "mp3Format.h"





/***** Konstruktor / Destruktor *****/
LameCodec::LameCodec( A_Track *Track )
:A_Codec( Track )
{
    // Variablen initialisieren
    puffer = NULL;
    puffer_gr = 0;
    bitrate = 128;

    format_info = NULL;
    

    // Encoder initialisieren
    // Achtung: Lame darf nicht in mehreren Instanzen gleichzeitig initialisiert werden!
    //          Auch wenn hier eine eigene Datenstruktur fr Lame vorhanden ist, werden
    //          offensichtlich intern noch weitere Daten verwendet!

    if( iomode == VideoFile::write ) {
                
        lame_init(&encoder);
        encoder.silent                  = 1;
        encoder.padding_type            = 2;

        encoder.VBR                     = 0;
        // encoder.VBR = vbr_off;
        encoder.VBR_q                   = 5;
        encoder.VBR_min_bitrate_kbps    = 56;
        encoder.VBR_max_bitrate_kbps    = 160;

        encoder.in_samplerate   =   SampleRate();
        encoder.num_channels    =   Channels();
        encoder.lowpassfreq     = 0;                /* freq in Hz. 0=lame choses. -1=no filter */
        encoder.highpassfreq    = 0 ;
        if( Channels() == 1 ) {                   /* mode 0,1,2,3 stereo,jstereo,dual channel,mono */
            encoder.mode=3;
            encoder.brate=64;
        } else {
            encoder.mode=1;
            encoder.brate=128;
        }

        lame_init_params(&encoder);
        
        
        // Puffer fr den Encoder
        encoder_puffer = new (int16_t*)[ Channels() ];
        encoder_puffer_gr = encoder.framesize;
        encoder_puffer_start = 0;
        for( int i=0; i < Channels(); i++ ) {
            encoder_puffer[i] = new int16_t[encoder_puffer_gr];
        }
    

        // Optionen registrieren
        AddOption( "Bitrate", _("average bitrate in kbps") ) = encoder.brate;
        AddOption( "VBR", _("compress with variable bitrate 0:no / 1:yes") ) = encoder.VBR;
        AddOption( "VBR_Quality", _("quantisation for variable bitrate (1: high quality - 9: low quality)") ) = encoder.VBR_q;
        AddOption( "VBR_MinBitrate", _("lowest bitrate (kbps) for VBR") )   = encoder.VBR_min_bitrate_kbps;
        AddOption( "VBR_MaxBitrate", _("highest bitrate (kbps) for VBR") )   = encoder.VBR_max_bitrate_kbps;
    }



    if( iomode == VideoFile::read ) {
        // Puffer fr den decoder
        decoder_puffer = new char[ decoder_puffer_gr ];
        decoder_puffer_pos = 0;
        decoder_puffer_rest = 0;
        
        // Decoder (mpeglib) initialisieren
        InitMP3( &decoder );

        // ersten Frame lesen
        decode( NULL, 0 );

        // Info ber mp3-Frame
        SetChannels( decoder.fr.stereo );
        SetSampleRate( freqs[decoder.fr.sampling_frequency] );
        bitrate = tabsel_123[decoder.fr.lsf][decoder.fr.lay-1][decoder.fr.bitrate_index];
        SetBitsPerSecond( (double)bitrate * 1000 );
        SetFrameSamples( (decoder.fr.lsf == 0) ? 1152 : 576 );
    }

}




LameCodec::~LameCodec()
{
    int bytes;

    if( iomode == VideoFile::read ) {
        ExitMP3( &decoder );
        delete [] decoder_puffer;
    }

    
    if( iomode == VideoFile::write ) {
        if( puffer_gr > 0 ) {
            // restliche Daten ausgeben
            bytes = lame_encode_finish( &encoder, (char *)puffer, puffer_gr );

            //if( bytes < 0 ) {
            //    file->Fehler( "~LameCodec: lame_encode_finish fehlgeschlagen!\n" );
            //} else
            if( bytes > 0 ) {
                // Puffer in Datei schreiben
                track->writeData( puffer, bytes, encoder.framesize );
            }

            delete [] puffer;
        }

        for( int i=0; i < Channels(); i++ ) {
            delete [] encoder_puffer[i];
        }
        delete [] encoder_puffer;
    }

    delete [] format_info;
}









int LameCodec::SeekSample( int64_t sample )
{
    if( iomode != VideoFile::read )
        return 0;

    ExitMP3( &decoder );        
    decoder_puffer_rest = 0;
    decoder_puffer_pos = 0;
    //remove_buf( &decoder );
    InitMP3( &decoder );

    return A_Codec::SeekSample( sample );
}





// Audiodaten schreiben
int LameCodec::encode( float **daten, int sampels )
{
    // Maximale Gre der kompremierten Ausgabe ( Obergrenze )
    long bytes = Channels() * 2 * sampels;
    int c=0, i=0;
    int sample = 0;

    // Puffer anfordern
    if( bytes > puffer_gr ) {
        delete [] puffer;
        puffer = new unsigned char[bytes];
        puffer_gr = bytes;
    }

    
    c = (Channels() > 1) ? 1 : 0;
    
    // schreibe einzelne Datenblcke
     do {
        // konvertieren
        for( i=0; i < encoder.framesize - encoder_puffer_start && i + sample < sampels; i++ ) {
            for( int j=0; j < Channels(); j++ ) {
                encoder_puffer[j][i] = (int16_t)( daten[j][i+sample] * (1<<15) );
            }
        }
        sample += i;
        encoder_puffer_start += i;
        if( encoder_puffer_start >= encoder.framesize )
            encoder_puffer_start = 0;
    
        bytes = lame_encode_buffer( &encoder, encoder_puffer[0], encoder_puffer[c],
                                    i, (char*)puffer, puffer_gr );

        if( bytes < 0 ) {
            ERROR_TXT( _("lame_encode_buffer failed!") );
            return 0;
        } else if( bytes > 0 ) {
            // Puffer in Datei schreiben
            if( !track->writeData( puffer, bytes, encoder.framesize ) )
                return 0;
        }
    } while( bytes > 0 );

    // keine Daten zu schreiben
    return 1;
}





// Audiodaten lesen
int LameCodec::decode( float **daten, int sampels )
{
    int i=0, j, size=0, stat=0;
    long bytes=64;
    char buf[128];

    do {
        // noch alte Daten im Puffer?
        if( decoder_puffer_rest == 0 ) {
            // einen Frame lesen und dekodieren
            do {
                // Puffer aus Datei lesen
                if( !track->readBytes( (unsigned char*)buf, bytes ) )
                    return 0;

                // Frame decodieren
                stat = decodeMP3( &decoder, buf, bytes, decoder_puffer, decoder_puffer_gr, &size );
                if( decoder.fr.stereo > 0 ) {
                    //channels = decoder.fr.stereo;
                    //decoder_puffer_rest = size / ( 2 * channels );
                    decoder_puffer_rest = size / ( 2 * decoder.fr.stereo );
                } else {
                    decoder_puffer_rest = 0;
                }
                decoder_puffer_pos = 0;
                assert( decoder_puffer_rest <= 1152 );
            } while( stat == MP3_NEED_MORE );
        }

        if( stat != MP3_OK )
            return 0;

        // Kanle in einen Datenstrom multiplexen
        while( i < sampels && decoder_puffer_rest > 0 ) {
            for( j=0; j<Channels(); j++ ) {
                daten[j][i] = (float)((int16_t *)decoder_puffer)[ decoder_puffer_pos * Channels() + j ] / (float)(1<<15);
            }
            decoder_puffer_rest--;
            decoder_puffer_pos++;
            i++;
        }
    } while( i < sampels );

    return 1;
}






int LameCodec::SetOption( OptionEntry *opt )
{
    DEBUG_TXT_3( "setze Option %s = %s", (const char*)opt->Name(), (const char*)*opt );

    if( opt->Name() == "Bitrate" ) {
        encoder.brate = (int)*opt;
        bitrate = encoder.brate;
        if( iomode == VideoFile::write )
            lame_init_params(&encoder);

    } else if( opt->Name() == "VBR" ) {
        if( (int)*opt > 0 )
            encoder.VBR = 1;
        else
            encoder.VBR = 0;
        if( iomode == VideoFile::write )
            lame_init_params(&encoder);

    } else if( opt->Name() == "VBR_Quality" ) {
        encoder.VBR_q = (int)*opt;
        if( encoder.VBR_q < 0 )
            encoder.VBR_q = 0;
        else if( encoder.VBR_q > 9 )
            encoder.VBR_q = 9;

        if( iomode == VideoFile::write )
            lame_init_params(&encoder);

    } else if( opt->Name() == "VBR_MinBitrate" ) {
        encoder.VBR_min_bitrate_kbps = (int)*opt;
        if( iomode == VideoFile::write )
            lame_init_params(&encoder);

    } else if( opt->Name() == "VBR_MaxBitrate" ) {
        encoder.VBR_max_bitrate_kbps = (int)*opt;
        if( iomode == VideoFile::write )
            lame_init_params(&encoder);

    } else {
        return 0;
    }

    
    
    // fr Avi-Dateien immer variable Bitrate angeben
    if( iomode == VideoFile::write && encoder.VBR == 0 && file->ModulName() != "Avi" ) {
        SetBitsPerSecond( (double)bitrate * 1000 );
    } else if( iomode == VideoFile::write ) {
        SetBitsPerSecond( 0 );
    }
    
    
    
    if( iomode == VideoFile::write ) {
        SetFrameSamples( encoder.framesize );
        
        if( encoder.framesize > encoder_puffer_gr ) {
            encoder_puffer_gr = encoder.framesize;
            for( int i=0; i < Channels(); i++ ) {
                delete [] encoder_puffer[i];
                encoder_puffer[i] = new int16_t[encoder_puffer_gr];
            }
        }
    }
    
    return 1;
}






int LameCodec::GetExtraInfo( uint8_t *&info, unsigned int &size )
{
    if( file->ModulName() == "Avi" ) {
        size_t n = size;
        WaveFormat w_format;
        Mp3Format mp3_format;

        w_format.Read( info, n );
        
        // wformat->nAvgBytesPerSec    = encoder.brate*125;
        w_format.BitsPerSample      = 0;

        mp3_format.ID               = 1;
        mp3_format.Flags            = 2;
        mp3_format.BlockSize        = encoder.framesize;
        mp3_format.FramesPerBlock   = 1;
        mp3_format.CodecDelay       = 1393;

        if( w_format.DataSize > 0 ) {
            delete [] w_format.Data;
        }
        n = mp3_format.Size();
        w_format.Data = new uint8_t[n];
        mp3_format.Write( w_format.Data, n );
        w_format.DataSize = mp3_format.Size();

        if( size < w_format.Size() ) {
            format_info = new uint8_t[w_format.Size()];
            info = format_info;
        }
        n = w_format.Size();
        w_format.Write( info, n );
        size = w_format.Size();

        return 1;
    }
    return 0;
}
