/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** Based on QT4Linux by Adam Williams.                                       **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <stdio.h>


#include "UlawCodec.h"
#include "debug.h"
#include "VideoFile/translate.h"




#define uBIAS 0x84
#define uCLIP 32635




/***** Konstruktor / Destruktor *****/
UlawCodec::UlawCodec( A_Track *Track )
:A_Codec( Track )
{
    // Variablen initialisieren
    puffer = NULL;
    puffer_gr = 0;
    
    SetBitsPerSecond( (double)(8 * Channels() * SampleRate()) );
    SetFrameSamples( 1 );

    // Tables initialisieren    
    tables.ulawtoint16_table=NULL;
    tables.ulawtoint16_ptr=NULL;
    tables.int16toulaw_table=NULL;
    tables.int16toulaw_ptr=NULL;

    ulaw_init_ulawtoint16();
    ulaw_init_int16toulaw();
}




UlawCodec::~UlawCodec()
{
    if( tables.ulawtoint16_table )
        delete [] tables.ulawtoint16_table; 
    
    if( tables.int16toulaw_table )
        delete [] tables.int16toulaw_table; 
                
    delete [] puffer;
}






// Audiodaten schreiben
int UlawCodec::encode( float **daten, int samples )
{
    int i,j;
    long bytes = (Channels() * samples);
    int16_t sample;

    // Puffer anfordern
    if( bytes > puffer_gr ) {
        delete [] puffer;
        puffer = new unsigned char[bytes];
        puffer_gr = bytes;
    }


    for( i=0; i<samples; i++ ) {
        for( j=0; j<Channels(); j++ ) {
            sample = (int16_t)( daten[j][i] * (1<<15) );
            puffer[(i*Channels()+j)] = tables.int16toulaw_ptr[sample];
        }
    }

    // Puffer in Datei schreiben
    return track->writeData( puffer, bytes, samples );
}






// Audiodaten lesen
int UlawCodec::decode( float **daten, int samples )
{
    int i,j;
    long bytes = (Channels() * samples);
    int16_t sample;

    // Puffer anfordern
    if( bytes > puffer_gr ) {
        delete [] puffer;
        puffer = new unsigned char[bytes];
        puffer_gr = bytes;
    }

    // Puffer aus Datei lesen
    if( !track->readBytes( puffer, bytes ) )
        return 0;

    // Kanle in einen Datenstrom multiplexen    
    for( i=0; i<samples; i++ ) {
        for( j=0; j<Channels(); j++ ) {
            sample = tables.ulawtoint16_ptr[ puffer[(i*Channels()+j)] ];
            daten[j][i] = (float)sample / (float)(1<<15);
        }
    }
    
    return 1;
}







int UlawCodec::ulaw_init_ulawtoint16( void )
{

    int i;

    if(!tables.ulawtoint16_table) {
        static int exp_lut[8] = { 0, 132, 396, 924, 1980, 4092, 8316, 16764 };
        int sign, exponent, mantissa, sample;
        unsigned char ulawbyte;

        tables.ulawtoint16_table = new int16_t[256];
        tables.ulawtoint16_ptr = tables.ulawtoint16_table;
        for( i = 0; i < 256; i++ ) {
            ulawbyte = (unsigned char)i;
            ulawbyte = ~ulawbyte;
            sign = (ulawbyte & 0x80);
            exponent = (ulawbyte >> 4) & 0x07;
            mantissa = ulawbyte & 0x0F;
            sample = exp_lut[exponent] + (mantissa << (exponent + 3));
            
            if(sign != 0)
                sample = -sample;
                
            tables.ulawtoint16_ptr[i] = sample;
        }
    }
    
    return 0;
}



int UlawCodec::ulaw_init_int16toulaw( void )
{
    int i;

    if(!tables.int16toulaw_table) {
        int sign, exponent, mantissa;
        unsigned char ulawbyte;
        int sample, i;
        int exp_lut[256] = {0,0,1,1,2,2,2,2,3,3,3,3,3,3,3,3,
                            4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
                            5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
                            5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
                            6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
                            6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
                            6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
                            6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
                            7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                            7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                            7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                            7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                            7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                            7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                            7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                            7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7};

        tables.int16toulaw_table = new unsigned char[65536];
        tables.int16toulaw_ptr = tables.int16toulaw_table + 32768;

        for(i = -32768; i < 32768; i++) {
            sample = i;
            // Get the sample into sign-magnitude. 
            sign = (sample >> 8) & 0x80;                    // set aside the sign 
            if(sign != 0)
                sample = -sample;                          // get magnitude 
            if(sample > uCLIP)
                sample = uCLIP;                            // clip the magnitude 

            // Convert from 16 bit linear to ulaw. 
            sample = sample + uBIAS;
            exponent = exp_lut[(sample >> 7) & 0xFF];
            mantissa = (sample >> (exponent + 3)) & 0x0F;
            ulawbyte = ~(sign | (exponent << 4) | mantissa);
#ifdef ZEROTRAP
            if (ulawbyte == 0)
                ulawbyte = 0x02;                            // optional CCITT trap 
#endif

            tables.int16toulaw_ptr[i] = ulawbyte;
        }
    }
    
    return 0;
}
