/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <limits.h>

#include "debug.h"
#include "AviFile.h"
#include "VideoFile/codecplugins.h"
#include "VideoFile/translate.h"


/***** Konstruktor / Destruktor *****/
AviFileIO::AviFileIO( const String Dateiname, IO_Mode mode )
:VideoFile( Dateiname, "Avi", mode )
{
    char *datei_name;

    // Variablen initialisieren
    wDatei = NULL;
    rDatei = NULL;

    datei_name = new char[ Dateiname.length() + 1 ];
    strcpy( datei_name, Dateiname.chars() );

    // Avi Datei ffnen
    if( iomode == read ) {
        // Datei zum lesen ffnen
        try {
            rDatei = new AviReadFile( datei_name );
        }
        
        catch(...) {
            ErrorStatus() = Errors::FileType;
            throw( ErrorStatus() );
        }

        // Anzahl der Tracks in der Datei
        SetVideoTracks( (int)rDatei->VideoStreamCount() );
        SetAudioTracks( (int)rDatei->AudioStreamCount() );
        
    } else if( iomode == write ) {
        // Datei zum schreiben ffnen
        int flags = 0;
        int modus = 0644;

        try {
            wDatei = new AviWriteFile( datei_name, flags, modus );
        }
        
        catch(...) {
            ErrorStatus() = Errors::Open;
            throw( ErrorStatus() );
        }
        
    } else {
        // falscher IO-Modus
        ErrorStatus() == Errors::IO_Mode;
        throw( ErrorStatus() );
    }

    delete [] datei_name;
}




AviFileIO::~AviFileIO()
{
    // Tracks schlieen
    deleteTracks();

    // Datei schlieen
    delete wDatei;
    delete rDatei;
}








/***** Video-Funktionen *********************************************/
// Video codec initialisieren
// Rckgabewert: false - Fehler
//               true  - Erfolg
V_Track *AviFileIO::openVideoTrack( double Rate, int x, int y, const char *Codec )
{
    if( iomode != write )
        return NULL;
    return new AviFileVtr( this, Rate, x, y, Codec );
}

// Track ffnen
V_Track *AviFileIO::openVideoTrack( int track )
{
    if( iomode != read )
        return NULL;

    if( VideoTracks() <= track )
        return NULL;

    return new AviFileVtr( this, track );
}








/***** Audiofunktionen **********************************************/

// Audio codec initialisieren
A_Track *AviFileIO::openAudioTrack( int Sample_rate, int Channels, const char *Codec )
{
    if( iomode != write )
        return NULL;
    return new AviFileAtr( this, Sample_rate, Channels, Codec );
}

// Track ffnen
A_Track *AviFileIO::openAudioTrack( int track )
{
    if( iomode != read )
        return NULL;

    if( AudioTracks() <= track )
        return NULL;

    return new AviFileAtr( this, track );
}









/*****************************************************************************/
/***** Videotrack-Funktionen *************************************************/
/*****************************************************************************/
// Video Track initialisieren
// Rckgabewert: false - Fehler
//               true  - Erfolg
AviFileVtr::AviFileVtr( AviFileIO *File, double rate, int x, int y, const char *Codec_name )
:Track( File, 0, Codec_name), V_Track( File, rate, x, y, Codec_name )
{
    file = File;

    FrameBuf = NULL;
    BufGr = 0;
    keyframe_ok = true;

    
    wDatei = file->wDatei;
    rDatei = file->rDatei;

    // Informationen ber Bildformat setzen
    BITMAPINFOHEADER binfo;
    memset( &binfo, 0, sizeof(BITMAPINFOHEADER) );

    binfo.biWidth = Width();
    binfo.biHeight = Height();

    // Codec erstellen
    // passenden Codec zuordnen und erstellen
    codec = AV_Codecs->CreateVCodec( this, "Avi", CodecName() );

    if( codec == NULL ) {
        ERROR_TXT( _("Video codec \"%s\" not implemented!"), CodecName().chars() );
        codec = new V_Codec( this );
    }

    strncpy( (char *)&binfo.biCompression, CodecID().chars(), 4 );

    wStream = wDatei->AddVideoStream( (char*)&binfo, sizeof(BITMAPINFOHEADER), Rate() );
    rStream = NULL;

    assert( wStream != NULL );
}





AviFileVtr::AviFileVtr( AviFileIO *File, int track )
:Track(File, track), V_Track( File, track )
{
    file = File;

    wDatei = file->wDatei;
    rDatei = file->rDatei;
    wStream = NULL;
    rStream = rDatei->GetStream( track, AviStream::Video );
    keyframe_ok = true;

    // Informationen aus Datei auslesen
    BITMAPINFOHEADER binfo;
    rStream->GetVideoFormatInfo( &binfo );

    breite = binfo.biWidth;
    hoehe = binfo.biHeight;

    char codecname[ 5 ];
    strncpy( codecname, (char *)&binfo.biCompression, 4 );
    codecname[4] = 0;
    SetCodecName( codecname );

    SetFrames( rStream->GetEndPos() );
    SetRate( 1 / rStream->GetFrameTime() );

    BufGr = breite * hoehe * 3;
    FrameBuf = new char[ BufGr ];

    // Codec erstellen
    // passenden Codec zuordnen und erstellen
    codec = AV_Codecs->CreateVCodec( this, "Avi", CodecName() );

    if( codec == NULL ) {
        ERROR_TXT( _("Video codec \"%s\" not implemented!"), CodecName().chars() );
        codec = new V_Codec( this );
    }
}



AviFileVtr::~AviFileVtr()
{
    delete [] FrameBuf;
    delete codec;
    codec = NULL;
}





// zu Bildposition springen
int AviFileVtr::Seek( long block )
{
    if( IOMode() == VideoFile::read ) {
        rStream->Seek( block );
        SetPosition( block );
        return 1;
    }

    return 0;
}



// letzter Keyframe
long AviFileVtr::GetKeyFrame( long bild )
{
    long i;
    if( IOMode() != VideoFile::read )
        return bild;

    if( keyframe_ok ) {
        i = rStream->GetPrevKeyFrame( bild );
        // DEBUG_TXT( "Keyframe %i zu Bild %i", i, bild );
        if( i >= 0 ) {
            return i;
        }
        keyframe_ok = false;
    }
    return bild;
}




// ist dies ein Keyframe?
bool AviFileVtr::IsKeyFrame( void )
{
    int flags=0;
    rStream->GetFrameFlags( &flags );

    return flags & AVIIF_KEYFRAME;
}





/***** Rohdaten lesen und schreiben *****************************************************/
// Rckgabewert: 1   -    OK
//               <1  -    Fehler
int AviFileVtr::readRaw( unsigned char *&data, long &bytes )
{
    unsigned gr, blocks=0;

    // gehe zu aktuellem Bild
    if( rStream == NULL )
        return 0;

    //if( !Seek(pos) )
    //    return 0;

    rStream->ReadDirect( FrameBuf, BufGr, 1, blocks, gr);
    bytes = gr;

    // Speicherbereich anforden
    if( BufGr <= gr ) {
        delete [] FrameBuf;
        BufGr = gr + 1;
        FrameBuf = new char[BufGr];
    }

    if( blocks == 1 ) {
        data = (unsigned char *)FrameBuf;
        //pos++;
        //printf( "Pos=%i, BufGr=%i, Data=%i\n", Pos, BufGr, (int)data[1000] );
        return 1;
    } else {
        data = NULL;
        return 0;
    }

    return 0;
}






int AviFileVtr::writeRaw( unsigned char *data, long bytes, int keyframe )
{
    int flags = 0;

    if( keyframe & IS_KEYFRAME )
        flags = AVIIF_KEYFRAME;

    wStream->AddChunk( (char*)data, bytes, flags );
    
    // Header auffrischen
    //wDatei->WriteHeaders();

    return 1;
}








/*****************************************************************************/
/***** Audiotrack - Funktionen ***********************************************/
/*****************************************************************************/
AviFileAtr::AviFileAtr( AviFileIO *File, int Sample_rate, int Channels, const char *Codec_name )
:Track(File, 0, Codec_name), A_Track( File, Sample_rate, Channels, Codec_name )
{
    file = File;

    puffer = NULL;
    puffer_gr = 0;
    samplesize = 0;

    wDatei = file->wDatei;
    rDatei = file->rDatei;
    wStream = NULL;
    rStream = NULL;

    WAVEFORMATEX wformat, *wformat_ptr;
    wformat_ptr = &wformat;
    unsigned int wformat_size = sizeof(WAVEFORMATEX);
    memset( wformat_ptr, 0, wformat_size );

    // initialisiere mit 16 Bit por Sample
    // (wird spter vom Codec berichtigt)
    int sampsize = 16 / 8 * Channels;
    wformat.nChannels = Channels;
    wformat.nSamplesPerSec = Sample_rate;
    wformat.nAvgBytesPerSec = sampsize * Sample_rate;
    wformat.nBlockAlign = 1;
    wformat.wBitsPerSample = 16;
    wformat.cbSize = 0;

    if( CodecName() != "" ) {
        wformat.wFormatTag = WAVE_FORMAT_UNKNOWN;
        codec = AV_Codecs->CreateACodec( this, "Avi", CodecName() );

        // in codec_name ist nun die Formatnummer fr den Codec
        // als String mit Format "%04x" gespeichert
        // - das wird spter in Nummer umgewandelt

        // der Codec kann wformat modifizieren
        // (das Avi-Format ist wirklich schwachsinnig!!!)
        if( codec != NULL ) {
            codec->GetExtraInfo( (uint8_t *)wformat_ptr, wformat_size );
        }
    }

    if( codec == NULL ) {
        ERROR_TXT( _("Audio codec \"%s\" not implemented!"), CodecName().chars() );
        codec = new A_Codec( this );
    }

    // codec_name in Formatnunner umwandeln
    wformat_ptr->wFormatTag = strtol( CodecID().chars(), NULL, 16 );
    DEBUG_TXT( _("Audioformat %04x"), wformat_ptr->wFormatTag );
    wStream = wDatei->AddAudioStream( (char *)wformat_ptr, wformat_size );
}




AviFileAtr::AviFileAtr( AviFileIO *File, int track )
:Track(File, track), A_Track( File, track )
{
    file = File;

    wDatei = file->wDatei;
    rDatei = file->rDatei;
    wStream = NULL;
    rStream = rDatei->GetStream( track, AviStream::Audio );
    assert( rStream != NULL );

    puffer = NULL;
    puffer_gr = 0;
    samplesize = 0;

    // Informationen aus Datei auslesen
    WAVEFORMATEX winfo;
    rStream->GetAudioFormatInfo( &winfo, NULL );

    unsigned short format = winfo.wFormatTag;
    SetChannels( winfo.nChannels );
    SetSampleRate( winfo.nSamplesPerSec );
    bytes_per_second = winfo.nAvgBytesPerSec;
    SetBitsPerSecond( (double)bytes_per_second * 8.0 );
    //winfo.nBlockAlign;
    //bits = winfo.wBitsPerSample;

    //winfo.cbSize;

    total_samples = (int64_t)( rStream->GetEndTime() * SampleRate() );
    if( (samplesize = rStream->GetSampleSize()) == 0 ) {
        SetRate( 1 / rStream->GetFrameTime() );
    }


    // passenden Codec zuordnen und erstellen
    char codecname[5];
    DEBUG_TXT( _("Create audio codec for format %04x"), format );
    sprintf( codecname, "%04x", format );
    SetCodecName( codecname );

    codec = AV_Codecs->CreateACodec( this, "Avi", CodecName() );

    if( codec == NULL ) {
        ERROR_TXT( _("Format \"%04x\" not implemented!"), format );
        codec = new A_Codec( this );
    }

    // vom Codec gendertes Format bernehmen
    // (Codec kann z.B. andere Channel-Anzahl benutzen)
    (A_Format)*this = (A_Format)*codec;
}





AviFileAtr::~AviFileAtr()
{
    // Formatinfo aus Codec aktualisieren
    if( IOMode() == VideoFile::write ) {
        char *format = NULL;
        unsigned int size = wStream->GetFormatInfo( format );
        codec->GetExtraInfo( (uint8_t*)format, size );
    }

    delete codec;
    codec = NULL;
}





// zu Sample springen
int AviFileAtr::Seek( long block )
{
    if( IOMode() != VideoFile::read )
        return 0;

    if( samplesize == 0 ) {
        // Block-Zugriff
        rStream->Seek( block );
    } else {
        // Sample-Zuriff
        int64_t sample = (int64_t)(( (double)bytes_per_second * (double)block ) / (Rate() * samplesize) );

        rStream->Seek( sample );
    }
    SetPosition( block );

    return 1;
}






/***** Rohdaten lesen und schreiben *****************************************************/
// Rckgabewert: 1   -    OK
//               <1  -    Fehler
int AviFileAtr::readRaw( unsigned char *&data, long &bytes )
{
    unsigned gr, blocks=0, samp=1;
    int stat;

    if( rStream == NULL )
        return 0;

    // Speicherbereich anforden
    if( samplesize == 0 ) {
        gr = rStream->GetDataSize();
        samp = 1;
    } else {
        gr = (long)( bytes_per_second / Rate() );
        samp = gr / samplesize;
    }

    if( puffer_gr <= gr ) {
        delete [] puffer;
        puffer_gr = gr + 1;
        puffer = new unsigned char[puffer_gr];
    }

    stat = rStream->ReadDirect( (char *)puffer, puffer_gr, samp, blocks, gr);

    bytes = gr;
    data = puffer;

    if( gr == 0 ) {
        return 0;
    }

    if( stat == 0 )
        return 1;

    return 0;
}






int AviFileAtr::writeRaw( unsigned char *data, long bytes, int samples )
{
    wStream->AddChunk( (char*)data, bytes, 16 );

    // Bitrate aktualisieren
    bytes_per_second = ( BytesWritten() * SampleRate() ) / Frames();
    wStream->SetBytesPerSec( bytes_per_second );

    //Framerate aktualisieren
    wStream->SetFrameRate( Rate() );

    return 1;
}
