/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>

#include "Avi.h"
#include "Avi_Track.h"
#include "Avi_Vtr.h"
#include "Avi_Atr.h"
#include "types/AVI_Codes.h"


#include "debug.h"
#include "VideoFile/codecplugins.h"
#include "VideoFile/translate.h"




// maximale Gre fr einen RIFF-Chunk
#define MAX_RIFF_SIZE           1024*1024*1024

// zum Testen
//#define MAX_RIFF_SIZE           10*1024*1024

// Platz pro Track im Header
#define HEADER_SIZE             2048                






/***** Konstruktor / Destruktor *****/
AviIO::AviIO( const String datei, IO_Mode mode )
:VideoFile( datei, "Avi", mode )
{
    // initialisieren
    size_invalid = false;
    open_dml_format = false;
    chunks_written = 0;
    

    // Datei ffnen
    if( iomode == read ) {
        if( datei != "" ) {
            avi_datei = new RIFF_File( datei.chars(), true );
            assert( avi_datei != NULL );
        } else {
            ErrorStatus() = Errors::Open;
            throw( ErrorStatus() );
        }

        
        // prfe Dateityp
        if( !sync_avi_start() ) {
            ErrorStatus() = Errors::FileType;
            throw( ErrorStatus() );
        }

        //lese Tracks und Index ein
        if( !parse_avi_file() ) {
            ErrorStatus() = Errors::FileType;
            throw( ErrorStatus() );
        }

    } else if( iomode == write ) {
        
        if( datei != "" ) {
            avi_datei = new RIFF_File( datei.chars(), false );
            assert( avi_datei != NULL );
        } else {
            ErrorStatus() = Errors::Open;
            throw( ErrorStatus() );     
        }
        
    } else {
        ErrorStatus() = Errors::IO_Mode;
        throw( ErrorStatus() );
    }

}




AviIO::~AviIO()
{
    // Tracks schlieen
    deleteTracks();

    // Header aktualisieren
    if( iomode == write && riff_list.size() > 0 && movi_list.size() > 0 ) {
        write_index();
        write_headers();
    }

    // Liste mit Avi-Tracks schlieen   
    while( track_list.size() > 0 ) {
        delete track_list.back();
        track_list.pop_back();
    }
    

    // Liste mit RIFF-Headern schlieen
    while( riff_list.size() > 0 ) {
        delete riff_list.back();
        riff_list.pop_back();
    }

    // Liste mit MOVI-Headern schlieen
    while( movi_list.size() > 0 ) {
        delete movi_list.back();
        movi_list.pop_back();
    }

    // Datei schlieen
    delete avi_datei;   
}








/***** Video-Track *********************************************/
// Video codec initialisieren
// Rckgabewert:  Zeiger auf Videotrack
V_Track *AviIO::openVideoTrack( double Rate, int x, int y, const char *Codec )
{
    if( iomode == write ) {
        return new AviVtr( this, Rate, x, y, Codec, VideoTracks() );
    }
    return NULL;
}



// Track ffnen
V_Track *AviIO::openVideoTrack( int track )
{
    if( iomode != read )
        return NULL;

    if( track >= VideoTracks() ) {
        ERROR_TXT( _("Video track %i not found!"), track );
        return NULL;
    }

    return new AviVtr( this, track );
}





/***** Audio-Track *********************************************/
// Audio codec initialisieren
// Rckgabewert:  Zeiger auf Audiotrack
A_Track *AviIO::openAudioTrack( int Sample_rate, int Channels, const char *Codec )
{
    if( iomode == write ) {
        return new AviAtr( this, Sample_rate, Channels, Codec, AudioTracks() );
    }
        
    return NULL;
}



// Track ffnen
A_Track *AviIO::openAudioTrack( int track )
{
    if( iomode != read )
        return NULL;

    if( track >= AudioTracks() ) {
        ERROR_TXT( _("Audio track %i not found!"), track );
        return NULL;
    }

    return new AviAtr( this, track );
}











/***** Dateityp prfen *******************************************************/
bool AviIO::sync_avi_start( void )
{
    uint32_t type, length;
    
    if( !avi_datei->read_chunk_header( type, length) ) {
        DEBUG_TXT( _("File conains no data!") );
        return false;
    }

    if( type != FORMAT_CODE_RIFF ) {
        DEBUG_TXT( _("File contains no 'RIFF' header!") );
        return false;
    }

    // Grenangabe in Header ungltig
    if( length <= 4 ) {
        length = avi_datei->bytes_left() + 4;
        size_invalid = true;
    }

    avi_datei->read( type );

    if( type != FORMAT_CODE_AVI ) {
        DEBUG_TXT( _("File-Typ is not 'AVI'.") );
        return false;
    }

    // RIFF-Chunk in Liste aufnehmen
    RIFF_Header *riff = new RIFF_Header;
    riff->Type = FORMAT_CODE_RIFF;
    riff->Length = length;
    riff->DataType = type;
    riff->Offset = avi_datei->pos() - riff->Size();

    riff_list.push_back( riff );

    DEBUG_TXT( _("Opened file - 'AVI' file format.") );
    return true;
}









/***** Datei einlesen ********************************************************/

bool AviIO::parse_avi_file()
{
    uint32_t type = 0;
    uint32_t length = 0;
    bool index_found = false;
    bool extended_avi = false;
    RIFF_Header *movi = NULL, *riff = NULL;


    while( avi_datei->bytes_left() >= 8 && avi_datei->read_chunk_header(type, length) ) {
        //DEBUG_TXT( "type = 0x%08x   length = %u", type, length );
        
        if( length > avi_datei->bytes_left() ) {
            //ERROR_TXT( _("Error in Avi file!") );
            //return false;
            length = avi_datei->bytes_left();
        }
        
        switch(type) {
        case CHUNK_CODE_LIST:
            // LIST-Typ einlesen
            avi_datei->read( type );

            if( length<4 && (type != LIST_TYPE_AVIMOVIE || !size_invalid) )
                return false;

            length -= 4;

            switch( type ) {
            case LIST_TYPE_AVIMOVIE:
                DEBUG_TXT( _("Movi data found.") );
                
                // movi-Chunk in Liste aufnehmen
                movi = new RIFF_Header;
                movi->Type = CHUNK_CODE_LIST;
                movi->Length = length + 4;
                movi->DataType = type;
                movi->Offset = avi_datei->pos() - movi->Size();
                movi_list.push_back( movi );

                if (length < 8 && size_invalid) {
                    // Wenn die Greninformation nicht stimmt,
                    // dehne den movi-Chunk bis zum Dateiende aus
                    movi->Length = avi_datei->bytes_left() + 4;
                }

                break;
                
            case LIST_TYPE_AVIHEADER:
                if( open_dml_format | extended_avi )
                    break;
                    
                if( !parse_headers( length ) )
                    return false;
                  
                break;

            default:
                break;
            }

            break;


            
        case CHUNK_CODE_AVINEWINDEX:
            if( open_dml_format | extended_avi )
                break;

            DEBUG_TXT_3( "Avi index found." );
            if( movi_list.size() > 0 ) {
                if( !(index_found = read_index( length )) )
                    return false;
            } else {
                return false;
            }
            break;


        case FORMAT_CODE_RIFF:
            if( length < 4 )
                return false;

            // RIFF-Typ einlesen
            avi_datei->read( type );

            // RIFF-Chunk in Liste aufnehmen
            riff = new RIFF_Header;
            riff->Type = FORMAT_CODE_RIFF;
            riff->Length = length;
            riff->DataType = type;
            riff->Offset = avi_datei->pos() - riff->Size();

            riff_list.push_back( riff );

            length -= 4;

            if (length < 8 && size_invalid) {
                // Wenn die Greninformation nicht stimmt,
                // dehne den RIFF-Chunk bis zum Dateiende aus
                riff->Length = avi_datei->bytes_left() + 4;
            }

            // keine weiteren Header mehr einlesen
            extended_avi = true;
           
            switch( type ) {
            case FORMAT_CODE_AVI:
                ERROR_TXT( _("Found RIFF type 'AVI ' at offset 0x%08x."), avi_datei->pos() - 12l );
                ERROR_TXT( _("But an extended Avi block with type 'AVIX' is expected.") );
                // in RIFF-Chunk weiterlesen
                length = 0;
                break;
            case FORMAT_CODE_AVIX:
                DEBUG_TXT( _("Found RIFF type 'AVIX' at offset 0x%08x."), avi_datei->pos() - 12l );
                // in RIFF-Chunk weiterlesen
                length = 0;
                break;
            default:
                ERROR_TXT( _("Unknown RIFF typ at offset 0x%08x"), avi_datei->pos() - 12l );
                return false;
            }
            break;

        case CHUNK_CODE_PADDING_DATA:   // JUNK
            DEBUG_TXT_3( "Found padding data." );
            break;


        }

        
        if( length ) {
            if( !avi_datei->skip( length + (length&1) ) )
            break;
        }

        // Bei Avidateien mit falscher Greninformation nur bis zum
        // Index-Block lesen
        if( size_invalid && type == CHUNK_CODE_AVINEWINDEX)
            break;
    }

    if( movi_list.size() == 0 ) {
        ERROR_TXT( _("No movi datablock found!") );
        return false;
    }

    INFO_TXT( _("The Avi file contains a total of %i RIFF chunks."), riff_list.size() );
    INFO_TXT( _("The Avi file contains a total of %i movi chunks."), movi_list.size() );
    
    INFO_TXT( _("%i Tracks found."), track_list.size() );
    DEBUG_PROC( {
        if( open_dml_format ) {
            DEBUG_TXT( _("Extended index:") );
            for( unsigned int i=0; i<track_list.size(); i++ ) {
                fprintf( stderr, "    Track %02x: %i Frames, %i Bytes\n",
                         i, track_list[i]->Index->size(), track_list[i]->Index->DataBytes() );
            }
        }
    } );


    return true;

}









/******* Header-Block einlesen ***********************************************/
bool AviIO::parse_headers( uint32_t &length_left )
{

    uint32_t type;
    uint32_t length;
    
    DEBUG_TXT( _("Read header chunk") );
    while( length_left >= 8 && avi_datei->read_chunk_header( type, length) ) {
        //DEBUG_TXT( "type = 0x%08x   length = %u", type, length );

        length_left -= 8;

        if( length > length_left) {
            ERROR_TXT( _("Error in Avi file!") );
            return false;
        }

        length_left -= ( length + (length & 1) );


        switch(type) {
        case CHUNK_CODE_AVIMAINHDR:
            DEBUG_TXT_3( "AVI-Header gefunden." );
            /* lese Avi-Header ein */
            if( !avi_header.ReadStream(avi_datei, length) )
                return false;

            DEBUG_TXT_3( "AVI-Header gelesen." );
            DEBUG_PROC_3( avi_header.printInfo() );
            
            break;


        case CHUNK_CODE_LIST:
            // Typ des LIST-Segments einlesen
            avi_datei->read( type );

            if( length < 4 /*&& (type != LIST_TYPE_AVIMOVIE || !size_invalid)*/ ) {
                ERROR_TXT( _("Error in Avi file!") );
                return false;
            }

            // Gre des gelesenen Typs von restlicher Datenlnge abziehen
            length -= 4;

            switch( type ) {
            case LIST_TYPE_STREAMHEADER:
                DEBUG_TXT_3( "Stream-Header Abschnitt gefunden" );
                
                if( !parse_stream_header( length ) )
                    return false;
                    //accept_index_only = false;
                
                break;

            case LIST_TYPE_ODML:
                if( !parse_odml_header( length ) )
                    return false;

                break;

            default:
                DEBUG_TXT_3( "unbekannter LIST-Typ 0x%08x gefunden", type );
                break;
            }

            break;
                
        case CHUNK_CODE_PADDING_DATA:   // JUNK
            DEBUG_TXT_3( "Found padding data." );
            break;


        case CHUNK_CODE_SEGMENT_HINT:   // VirtualDub segment hint block
            DEBUG_TXT_3( "VirtualDub segment hint block found." );
            break;

        default:
            DEBUG_TXT_3( "Unknown chunk code 0x08x found.", type ); 
        }

        avi_datei->skip( length + (length & 1) );

    }
    DEBUG_TXT_3( "End reading header chunk." );

    return true;
}














/**Stream-Header einlesen
  */
bool AviIO::parse_stream_header( uint32_t &length_left )
{
    AviTrack *track;

    // einen euen Track anlegen
    // die Track-Nummer folgt aus den schon existierenden Tracks
    track = new AviTrack( this, track_list.size() );

    // Header einlesen
    if( !track->ReadHeaders( length_left ) ) {
        delete track;
        return false;
    }

    // Track in Liste einfgen
    track_list.push_back( track );

    // Video-/Audio-Tracks hochzhlen
    if( track->Type() == STREAM_CODE_VIDEO ) {
        SetVideoTracks( VideoTracks() + 1 );
    } else if( track->Type() == STREAM_CODE_AUDIO ) {
        SetAudioTracks( AudioTracks() + 1 );
    }

    return true;    
}






/**OpenDML-Header einlesen
  */
bool AviIO::parse_odml_header( uint32_t &length_left )
{
    uint32_t type;
    uint32_t length;

    DEBUG_TXT( _("Read OpenDML header chunk.") );
    while( length_left >= 8 && avi_datei->read_chunk_header( type, length) ) {
        //DEBUG_TXT( "type = 0x%08x   length = %u", type, length );

        length_left -= 8;

        if( length > length_left) {
            ERROR_TXT( _("Error in Avi file!") );
            return false;
        }

        length_left -= ( length + (length & 1) );


        switch(type) {
        case CHUNK_CODE_EXT_AVIHEADER:
            DEBUG_TXT_3( "Erweiterten AVI-Header gefunden." );
            /* lese Avi-Header ein */
            if( !avi_ext_header.ReadStream(avi_datei, length) )
                return false;

            //open_dml_format = true;

            DEBUG_TXT_3( "Erweiterten AVI-Header gelesen." );
            DEBUG_PROC_3( avi_ext_header.printInfo() );
            break;
        default:
            break;
        }
        
        avi_datei->skip( length + (length & 1) );
    }
    
    DEBUG_TXT_3( "End reading OpenDML header chunk." );

    return true;
}












/******** Index lesen ********************************************************/
bool AviIO::read_index( uint32_t &length )
{
    bool abs = true;
    AviIndexEntry entry;
    uint64_t offset = 0;


    // Index-Eintrag einlesen
    entry.ReadStream( avi_datei, length );

    // Die Adresse "ChunkOffset" sollte eigentlich relativ zu Anfang des ersten
    // 'movi'-Chunks sein (Offset vor 'movi').
    // Natrlich gibt es einige Programme, die sich nicht daran halten wollen...
    // (z.B. VitrtualDub?)
    // Hier werden absolute Adressen erkannt und entsprechend umgewandelt.

    if( abs && entry.ChunkOffset < movi_list[0]->DataOffset() - 4 )
        abs = false;

    if( !abs ) {
        offset = movi_list[0]->DataOffset() - 4;
    }

    // Indexeintrag zu jeweiligem Track hinzufgen
    if( track_list.size() > entry.StreamNum() ) {
        track_list[entry.StreamNum()]->Index->AddEntry( entry, offset );
    }

    while( length >= entry.Size() ) {
        entry.ReadStream( avi_datei, length );

        // Indexeintrag zu jeweiligem Track hinzufgen
        if( track_list.size() > entry.StreamNum() ) {
            track_list[entry.StreamNum()]->Index->AddEntry( entry, offset );
        }
    }

    DEBUG_TXT( "Avi-Index gelesen" );
    DEBUG_PROC( {
        for( unsigned int i=0; i<track_list.size(); i++ ) {
            fprintf( stderr, "    Track %02x: %i Frames, %i Bytes\n",
                     i, track_list[i]->Index->size(), track_list[i]->Index->DataBytes() );
        }
    } );
    return true;
}












/******** Index neu aufbauen *************************************************/
bool AviIO::rebuild_index( void )
{
    uint32_t type, length;
    uint64_t offset = 0;
    uint16_t track_num = 0, track_id = 0;

    INFO_TXT( _("Rebuilding index...") );

    avi_datei->seek( movi_list[0]->DataOffset() );

    while( avi_datei->bytes_left() >= 8 && avi_datei->read_chunk_header(type, length) ) {
        offset = avi_datei->pos() - 8l;

        track_num = RIFF_ITypes::stream_num( type );
        track_id = (uint16_t)(type >> 16);

        if( track_id == CHUNK_TYPE_DIBBITS || track_id == CHUNK_TYPE_DIBCOMPRESSED ||
            track_id == CHUNK_TYPE_PALCHANGE || track_id == CHUNK_TYPE_WAVEBYTES ) {

            // Indexeintrag zu jeweiligem Track hinzufgen
            if( track_list.size() > track_num ) {
                track_list[track_num]->Index->AddEntry( offset, length );
            }

            avi_datei->skip( length + (length & 1) );
        } else {
            DEBUG_TXT( "Unbekannter Track-Typ bei Offset 0x%x gefunden", offset );
            break;
        }

    }

    return true;
}














/*****************************************************************************/
/***** Funktionen zum Schreiben **********************************************/
/*****************************************************************************/

/**Avi-Header schreiben
  */
bool AviIO::write_avi_header( size_t &length )
{
    unsigned int i=0;
    
    // Daten im Header aktualisieren
    // ersten Videotrack finden
    for( i=0; i<track_list.size(); i++ ) {
        if( track_list[i]->Type() == STREAM_CODE_VIDEO )
            break;
    }
    
    // kein Videotrack gefunden -> benutze ersten Track
    if( i >= track_list.size() )
        i=0;

    if( track_list.size() > 0 && track_list[i]->Frames() > 0 ) {
    
        avi_header.MicroSecPerFrame = (uint32_t)( 1000000.0 / track_list[i]->FrameRate() );
        avi_header.MaxBytesPerSec = (uint32_t)(BytesWritten() * track_list[i]->FrameRate() / track_list[i]->Frames() );
        
        // Anzahl der Bilder im ersten RIFF-Chunk
        if( riff_list.size() == 1 )
            avi_header.TotalFrames = track_list[i]->Frames();  
        avi_header.Streams = track_list.size();
        avi_header.Width = track_list[i]->Width();
        avi_header.Height = track_list[i]->Height();

        // gesamte Anzahl der Bilder wird im extended-Header gespeichert
        //avi_ext_header.TotalFrames = track_list[i]->Frames();;
    }

    avi_datei->write( CHUNK_CODE_AVIMAINHDR );
    avi_datei->write( avi_header.Size() );
    length -= 8;
    if( !avi_header.WriteStream( avi_datei, length ) )
        return false;
    
    return true;
}







/**Avi-Header schreiben
  */
bool AviIO::write_avi_ext_header( size_t &length )
{
    unsigned int i=0;

    // Daten im Header aktualisieren
    // ersten Videotrack finden
    for( i=0; i<track_list.size(); i++ ) {
        if( track_list[i]->Type() == STREAM_CODE_VIDEO )
            break;
    }
    if( i >= track_list.size() )
        i=0;

    if( track_list.size() > 0 && track_list[i]->Frames() > 0 ) {
        // gesamte Anzahl der Bilder wird im extended-Header gespeichert
        avi_ext_header.TotalFrames = track_list[i]->Frames();;
    }

    if( length < avi_ext_header.Size() + 8 )
        return false;

    avi_datei->write( CHUNK_CODE_EXT_AVIHEADER );
    avi_datei->write( avi_ext_header.Size() );
    length -= 8;
    if( !avi_ext_header.WriteStream( avi_datei, length ) )
        return false;

    return true;
}









/**Sreibzugriff auf den Movi-Chunk vorbereiten
  */
bool AviIO::write_to_movi_chunk( size_t bytes )
{
    // ist die Datei schon initialisiert?
    if( riff_list.size() == 0 ) {
        RIFF_Header *riff_chunk = new RIFF_Header;
        riff_chunk->Type = FORMAT_CODE_RIFF;
        riff_chunk->DataType = FORMAT_CODE_AVI;
        riff_chunk->Offset = 0;

        riff_list.push_back( riff_chunk );

        // erstmal die Header-Informationen in die Datei schreiben
        // dabei wird auch ein MOVI-Chunk erstellt
        if( !write_headers() )
            return false;
    }


                
    // ist noch Platz im aktuellen RIFF-Chunk
    if( riff_list.back()->Length + bytes >= MAX_RIFF_SIZE ) {
        
        // den Index ans Ende des Blocks schreiben
        if( ! open_dml_format )  {
            // zuerst einen openDML-Index schreiben       
            if( !write_odml_index() )
                return false;
            
            // standard-Index schreiben
            if( !write_index() )
                return false;
            
            open_dml_format = true;
        } else {
            // openDML-Index schreiben
            if( !write_odml_index() )
                return false;
        }
        

        // die Header-Informationen in die Datei schreiben
        // dabei wird auch der letzte RIFF- und MOVI-Header aktualisiert
        if( !write_headers() )
            return false;
        
        // einen neuen RIFF-Chunk am Dateiende erstellen
        // dies entspricht der OpenDML Erweiterung des AVI-Formats
        // Der Typ ist diesmal 'AVIX' fr Avi-Extended
        // Dieser Chunk enthlt nur einen movi-Chunk und keine Header
        avi_datei->seek( 0, SEEK_END );
        
        RIFF_Header *riff_chunk = new RIFF_Header;
        riff_chunk->Type = FORMAT_CODE_RIFF;
        riff_chunk->DataType = FORMAT_CODE_AVIX;
        riff_chunk->Offset = avi_datei->pos();
        riff_list.push_back( riff_chunk );

        // einen neuen movi-Chunk im RIFF-Chunk erstellen
        RIFF_Header *movi_chunk = new RIFF_Header;
        movi_chunk->Type = CHUNK_CODE_LIST;
        movi_chunk->DataType = LIST_TYPE_AVIMOVIE;
        movi_chunk->Offset = riff_chunk->DataOffset();
        movi_list.push_back( movi_chunk );

        // bisherige Gesamtgre des RIFF-Chunks - der enthlt bisher nur
        // den leeren movi-Chunk
        riff_chunk->SetDataSize( movi_chunk->Size() );

        // Chunks in Datei schreiben
        size_t b = riff_chunk->Size() + movi_chunk->Size();
        if( !riff_chunk->WriteStream( avi_datei, b ) )
            return false;
            
        if( !movi_chunk->WriteStream( avi_datei, b ) )
            return false;   
    }

    // nach 1000 geschriebenen Daten-Blcken die Header aktualisieren
    if( (++chunks_written % 1000) == 0 ) {
        if( !write_headers() )
            return false;
    }
    
    // der Chunk kann geschrieben werden
    // bytes zu RIFF- und MOVI-Chunk zuzhlen
    riff_list.back()->Length += bytes + (bytes & 1);
    movi_list.back()->Length += bytes + (bytes & 1);
    
    return true;
}







/**Header schreiben
  */
bool AviIO::write_headers( void )
{
    size_t size = 0;
    size_t length = 0;
    RIFF_Header header_chunk, stream_header_chunk, odml_header_chunk;
    header_chunk.Type = CHUNK_CODE_LIST;
    header_chunk.DataType = LIST_TYPE_AVIHEADER;
    stream_header_chunk.Type = CHUNK_CODE_LIST;
    stream_header_chunk.DataType = LIST_TYPE_STREAMHEADER;
    odml_header_chunk.Type = CHUNK_CODE_LIST;
    odml_header_chunk.DataType = LIST_TYPE_ODML;

        
    // ist die Datei schon geffnet?
    if( riff_list.size() == 0 )
        return false;
    
    
    // wenn noch kein movi-Chunk vorhanden ist, erstelle jetzt einen
    if( movi_list.size() == 0 ) {
        RIFF_Header *movi_chunk = new RIFF_Header;
        movi_chunk->Type = CHUNK_CODE_LIST;
        movi_chunk->DataType = LIST_TYPE_AVIMOVIE;

        // genug Platz fr Indexeintrge freihalten
        movi_chunk->Offset = riff_list[0]->DataOffset() + header_chunk.Size()
                           + HEADER_SIZE * (track_list.size() + 1);
        movi_chunk->Offset += movi_chunk->Offset & 1;

        movi_list.push_back( movi_chunk );

        // bisherige Gesamtgre des RIFF-Chunks berechnen
        riff_list[0]->SetDataSize( movi_list[0]->Offset /*+ movi_list[0]->Size() - riff_list[0]-Size()*/);
    }



    // Echte Gre des Headerblocks berechnen
    size = avi_header.Size() + 8;
    if( open_dml_format ) {
        size += avi_ext_header.Size() + 8 + 12;
    }

    // Gre der Header fr die einzelnen Tracks zusammenzhlen
    // Der Header-Block ist jeweils in einem LIST-Chunk enthalten (Headergre 12)
    for( unsigned int i=0; i<track_list.size(); i++ ) {
        size += 12 + track_list[i]->Size();
    }

    // Der gesamte Header-Block ist in einem LIST-Chunk enthalten (Headergre 12)
    header_chunk.SetDataSize( size );

    // Der Header-Block darf sich mit dem ersten Daten-Block nicht berschneiden
    if( riff_list[0]->DataOffset() + header_chunk.Size()
      + header_chunk.DataSize() > movi_list[0]->Offset ) {
        return false;
    }
  
    // Der Header-Block nimmt den gesamten Platz bis zum ersten movi-Chunk ein
    header_chunk.SetDataSize( movi_list[0]->Offset - riff_list[0]->DataOffset() - header_chunk.Size() );

    

    
    // beginne mit dem Schreiben
    // Datei auf den Anfang des ersten RIFF-Headers positionieren
    // Der Header-Block flogt dierekt nach dem ersten Riff-Header
    length = movi_list[0]->Offset;
    //DEBUG_TXT( "Schreibe Header-Block mit der Lnge %i Bytes", length );
    
    avi_datei->seek( riff_list[0]->Offset );
    if( !riff_list[0]->WriteStream( avi_datei, length ) )
        return false;

    // Avi-Header schreiben
    if( !header_chunk.WriteStream( avi_datei, length ) )
        return false;
        
    if( !write_avi_header( length ) )
        return false;

    // Stream-Header und Stream-Format schreiben
    for( unsigned int i=0; i<track_list.size(); i++ ) {
        stream_header_chunk.Offset = avi_datei->pos();
        stream_header_chunk.SetDataSize( track_list[i]->Size() );
        if( !stream_header_chunk.WriteStream( avi_datei, length ) )
            return false;
        
        // Lngenangabe im Stream-Header aktualisieren (nur Lnge im ersten RIFF-Chunk)
        if( riff_list.size() == 1 )
            track_list[i]->UpdateLength();  
            
        if( !track_list[i]->WriteHeaders( length ) )
            return false;
    }

    // OpenDML-Header schreiben
    if( open_dml_format ) {
        odml_header_chunk.Offset = avi_datei->pos();
        odml_header_chunk.SetDataSize( avi_ext_header.Size() + 8 );
        if( !odml_header_chunk.WriteStream( avi_datei, length ) )
            return false;

        if( !write_avi_ext_header( length ) )
            return false;
    }
    

    // restlichen Platz auffllen
    if( length > 8 ) {
        avi_datei->write( CHUNK_CODE_PADDING_DATA );
        avi_datei->write( length - 8 );
        length -= 8;

        uint8_t c[2] = {0,0};
        while( length > 0 ) {
            avi_datei->write( c, 2 );
            length -= 2;
        }
    }

    // letzten RIFF- und movi-Header aktualisieren
    if( riff_list.size() > 1 ) {
        length = 12;
        avi_datei->seek( riff_list.back()->Offset );
        if( !riff_list.back()->WriteStream( avi_datei, length ) )
            return false;
    }

    if( movi_list.size() >= 1 ) {
        length = 12;
        avi_datei->seek( movi_list.back()->Offset );
        if( !movi_list.back()->WriteStream( avi_datei, length ) )
            return false;
    }

    
    return true;    
}











/***** Index schreiben *******************************************************/

/**Index schreiben
  */
bool AviIO::write_index( void )
{
    unsigned int i=0, next_track=0;
    uint64_t offset = 0;
    unsigned long *index_pos;
    AviIndexEntry entry;
    AviTrack *tr = NULL;
    long n = 0;
    uint32_t size = 0;


    // OpenDML - Index schreiben
    if( open_dml_format ) {
        return write_odml_index();
    }

            
    // Standard-Index schreiben
    // Die Indexeintrge werden aus den einzelnen Tracks zusammengemischet.
    // Dabei werden die Eintrge nach steigendem Dateioffset sortiert.
    // Es wird vorausgesetzt, da die Indexeintrge eines einzelnen Tracks
    // bereits sortiert sind.
    index_pos = new unsigned long[track_list.size()];
    
    for( i=0; i<track_list.size(); i++ ) {
        index_pos[i] = 0;
        n += track_list[i]->Index->size();
    }
    size = n * entry.Size();
    
    // Header und Gre des Index schreiben
    avi_datei->write( CHUNK_CODE_AVINEWINDEX );
    avi_datei->write( size );
    riff_list.back()->Length += size + 8;
    
    while( n > 0 ) {
        offset = 0;
        
        for( i=0; i<track_list.size(); i++ ) {
            if( track_list[i]->Index->size() <= index_pos[i] )
                continue;
            
            if( offset == 0 || track_list[i]->Index->at(index_pos[i])->AbsFileOffset < offset ) {
                offset = track_list[i]->Index->at(index_pos[i])->AbsFileOffset;
                next_track = i;
            }
        }

        // keine weiteren Indexeintrge mehr
        assert( track_list[next_track]->Index->size() > index_pos[next_track] );

        // Eintrag schreiben
        tr = track_list[next_track];
        
        entry.ID = tr->TrackID();
        entry.Flags = tr->Index->at(index_pos[next_track])->IsKeyframe() ? AVI_INDEX_KEYFRAME : 0;
        entry.ChunkOffset = (uint32_t)( offset - movi_list[0]->DataOffset() + 4 );
        entry.ChunkLength = tr->Index->at(index_pos[next_track])->ChunkSize;

        if( !entry.WriteStream( avi_datei, size ) ) {
            delete [] index_pos;
            return false;
        }
        
        index_pos[next_track]++;
        n--;
    }

    delete [] index_pos;
    return true;
}










/**OpenDML Index schreiben
  */
bool AviIO::write_odml_index( void )
{
    unsigned int i;
    size_t size = 0;
    
    for( i=0; i<track_list.size(); i++ ) {
        size += track_list[i]->WriteIndex();
    }

    // Gre zum aktuellen movi-chunk hinzufgen
    riff_list.back()->Length += size;
    movi_list.back()->Length += size;

    return true;
}
