/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                                                **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/


#include "Avi_Atr.h"
#include "debug.h"
#include "VideoFile/codecplugins.h"
#include "VideoFile/translate.h"
#include "types/AVI_Codes.h"






/*****************************************************************************/
/***** Audiotrack-Funktionen *************************************************/
/*****************************************************************************/

AviAtr::AviAtr( AviIO *File, int Sample_rate, int Channels, const char *Codec_name, int track )
:Track( File, track, Codec_name ), A_Track( File, Sample_rate, Channels, Codec_name )
{
    file = File;

    buffer_size = 0;
    buffer_data = NULL;
    sample_size = 0;
    format_info = NULL;
                
    // zugehrigen Avi-Track anlegen
    avi_track = new AviTrack( file, file->track_list.size() );
    file->track_list.push_back( avi_track );

    avi_track->SetType( STREAM_CODE_AUDIO );

 
    codec = AV_Codecs->CreateACodec( this, "Avi", CodecName() );

    if( codec == NULL ) {
        ERROR_TXT( _("Audio codec \"%s\" not implemented!"), CodecName().chars() );
        codec = new A_Codec( this );
    }

    // vom Codec gendertes Format bernehmen
    SetAFormat( *codec );
    
    
    // codec_name in Formatnunner umwandeln
    avi_track->SetFormatCode( strtol( CodecID().chars(), NULL, 16 ) );

    // speichere Formatinformationen des Audiotracks
    avi_track->SetSampleRate( Sample_rate );
    avi_track->SetChannels( Channels );
    avi_track->SetBitsPerSample( Bits() );
    avi_track->SetFrameRate( Rate(), (int)( BitsPerSecond() / 8.0 + 0.5 ), FrameSamples(), FrameBytes() );
        
    // Bitrate wird aus tatschlich geschriebenen Daten begechnet
    //avi_track->SetBytesPerSec( (int)(BitsPerSecond() / 8) );

    // Der Codec kann WaveFormat verndern und eigene Daten speichern
    size_t n = 0;
    uint8_t *info = NULL;

    CodecExtraInfo( info, n );
    Codec()->GetExtraInfo( info, n );
    avi_track->wave_format->Read( info, n );
}







AviAtr::AviAtr( AviIO *File, int track )
: Track(File, track), A_Track( File, track )
{
    file = File;

    buffer_size = 0;
    buffer_data = NULL;
    format_info = NULL;
    
    // Avi-Track finden
    avi_track = NULL;
    int t = track;
    unsigned int i = 0;
    for( i = 0; i < file->track_list.size(); i++ ) {
        if( file->track_list[i]->Type() == STREAM_CODE_AUDIO && --t < 0 )
            break;
    }
    avi_track = file->track_list[i];

    
    bytes_per_sec = avi_track->BytesPerSec();
    SetBitsPerSecond( bytes_per_sec * 8 );
    sample_size = avi_track->SampleSize();
    SetSampleRate( avi_track->SampleRate() );
    SetChannels( avi_track->Channels() );
    SetBits( avi_track->BitsPerSample() );
    
    if( bytes_per_sec > 0 && sample_size > 0 )
        SetFrameSamples( (int)(SampleRate() * avi_track->BlockAlign() / bytes_per_sec + 0.5) );
    else
        SetFrameSamples( 1 );
        
    total_samples = avi_track->Samples();
    SetRate( avi_track->FrameRate() );
    SetFrames( avi_track->Frames() );
    
    
    
    // passenden Codec zuordnen und erstellen
    char codecname[5];
    sprintf( codecname, "%04x", avi_track->FormatCode() );
    SetCodecName( codecname );

    codec = AV_Codecs->CreateACodec( this, "Avi", CodecName() );

    if( codec == NULL ) {
        ERROR_TXT( _("Audio codec \"%s\" not implemented!"), CodecName().chars() );
        codec = new A_Codec( this );
    }

    // vom Codec gendertes Format bernehmen
    SetAFormat( *codec );
}






AviAtr::~AviAtr()
{
    if( IOMode() == VideoFile::write ) {
        // vom Codec gendertes Format bernehmen
        SetAFormat( *codec );
    
        // speichere Formatinformationen des Audiotracks
        //avi_track->SetBlockAlign( FrameBytes() );
        avi_track->SetBitsPerSample( Bits() );
        
        // Framerate aktualisieren
        avi_track->SetFrameRate( Rate(), (int)( BitsPerSecond() / 8.0 + 0.5 ), FrameSamples(), FrameBytes() );

        
        // Der Codec kann WaveFormat verndern und eigene Daten speichern
        size_t n = 0;
        uint8_t *info = NULL;

        CodecExtraInfo( info, n );
        Codec()->GetExtraInfo( info, n );
        avi_track->wave_format->Read( info, n );
    }

    delete [] format_info;    
    delete codec;
    codec = NULL;
}




int AviAtr::Seek( long block )
{
    if( /*sample_size == 0*/ true ) {
        // Die Unterteilung in Blcke der Avidatei entspticht der
        // logischen Unterteilung fr die Positionierung
        if( block < Frames() ) {
            SetPosition( block );
            avi_track->SetNextRead( block, 0 );

            return 1;
        }
    } else {
        // finde anhand von sample_size und sample_rate die echte Position
        // in der Avi-Datei

        // Startzeit des angeforderten Blocks
        double time = (double)block / Rate();
        if( time > TotalTime() )
            return 0;
            
        // Offset im Stream
        off_t offset = (off_t)((double)bytes_per_sec * time);

        // finde echten Datenblock mit diesem Offset
        avi_track->SetNextRead( 0, offset );

        // Positionierung ausgefhrt
        SetPosition( block );
    }
    
    return 0;
}











// Rohdaten lesen und schreiben
int AviAtr::readRaw( unsigned char *&data, long &bytes )
{
    if( Position() >= Frames() ) {
        return 0;
    }

    //if( sample_size == 0 ) {
        // Datengre des Blocks feststellen
        bytes = avi_track->Index->at(Position())->ChunkSize;
        avi_track->SetNextRead( Position(), 0 );
    //} else {
    //    bytes = (long)(bytes_per_sec / Rate());
    //}
        
    // Speicher fr Daten erstellen
    if( buffer_size < bytes ) {
        delete [] buffer_data;
        buffer_data = new uint8_t[bytes];
        buffer_size = bytes;
    }
    data = buffer_data;

    // Daten lesen
    if( !avi_track->ReadData( data, bytes ) )
        return 0;
    

    
    return 1;
}






int AviAtr::writeRaw( unsigned char *data, long bytes, int samp )
{
 
    // Daten schreiben
    if( !avi_track->WriteData( data, bytes, 0 ) )
        return 0;

    // Framerate aktualisieren
    avi_track->SetFrameRate( Rate(), (int)( BitsPerSecond() / 8.0 + 0.5 ), FrameSamples(), FrameBytes() );

    return 1;
}






int AviAtr::CodecExtraInfo( uint8_t *&info, unsigned int &size )
{
    delete [] format_info;

    size_t n = avi_track->wave_format->Size();
    format_info = new uint8_t[n];
    avi_track->wave_format->Write( format_info, n );

    size = avi_track->wave_format->Size();
    info = format_info;

    return 1;
}

