/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/


/********************************************************************\
*****   Audio / Video - Track fr Avi Dateien                    *****
\********************************************************************/

#include <string.h>

#include "Avi_Track.h"
#include "types/AVI_Codes.h"
#include "VideoFile/translate.h"
#include "debug.h"


AviTrack::AviTrack( AviIO *File, int track )
{
    stream_header = new StreamHeader;
    Index = new FrameIndex;

    bitmap_info = NULL;
    wave_format = NULL;
    odml_index = NULL;

    file = File;
    avi_datei = file->avi_datei;

    track_num = track;

    next_chunk = 0;
    next_chunk_offset = 0;
}




AviTrack::~AviTrack()
{
    delete Index;
    
    delete stream_header;

    delete bitmap_info;

    delete wave_format;

    delete odml_index;
}










/**Bytes fr Header-Informationen
  */
size_t AviTrack::Size(void)
{
    size_t size = stream_header->Size() + 8;

    if( bitmap_info != NULL )
        size += bitmap_info->Size() + 8;

    if( wave_format != NULL )
        size += wave_format->Size() + 8;

    if( odml_index != NULL )
        size += odml_index->Size() + 8;

    return size;
}


    


    
/**Header-Informationen schreiben
  */
bool AviTrack::WriteHeaders( size_t &length )
{
    if( length < Size() )
        return false;
        
    // Werte aktualisieren
    if( wave_format != NULL ) {
        wave_format->AvgBytesPerSec = (uint32_t)( Index->DataBytes() * FrameRate() / Index->size() );
    }
    
    // Kennung und Gre von Stream-Header schreiben
    avi_datei->write( CHUNK_CODE_STREAMHEADER );
    avi_datei->write( stream_header->Size() );
    length -= 8;

    // Stream-Header schreiben
    stream_header->WriteStream( avi_datei, length );


    if( stream_header->StreamType == STREAM_CODE_VIDEO && bitmap_info != NULL ) {
    
        // Kennung und Gre von Stream-Format schreiben
        avi_datei->write( CHUNK_CODE_STREAMFORMAT );
        avi_datei->write( bitmap_info->Size() );
        length -= 8;            

        // Bitmap-Info schreiben
        bitmap_info->WriteStream( avi_datei, length );
    } else if( stream_header->StreamType == STREAM_CODE_AUDIO && wave_format != NULL ) {

        // Kennung und Gre von Stream-Format schreiben
        avi_datei->write( CHUNK_CODE_STREAMFORMAT );
        avi_datei->write( wave_format->Size() );
        length -= 8;

        // Wave-Format schreiben
        wave_format->WriteStream( avi_datei, length );
    }

    if( odml_index != NULL ) {

        // OpenDML-Superindex schreiben
        avi_datei->write( CHUNK_CODE_EXTENDED_INDEX );
        avi_datei->write( odml_index->Size() );
        length -= 8;

        odml_index->WriteStream( avi_datei, length );
    }

    return true;
}





/**Header-Informationen lesen
  */
bool AviTrack::ReadHeaders( size_t &length_left )
{
    uint32_t type = 0, length = 0;
    
    while( length_left > 8 && avi_datei->read_chunk_header( type, length ) ) {

        length_left -= 8;

        if( length > length_left) {
            ERROR_TXT( _("Error in Avi file!") );
            return false;
        }

        length_left -= ( length + (length & 1) );

        switch(type) {
        case CHUNK_CODE_STREAMHEADER:
            DEBUG_TXT_3( "Stream-Header gefunden" );
            // Stream-Header lesen
            if( !stream_header->ReadStream( avi_datei, length ) )
                return false;

            DEBUG_TXT_3( "Stream-Header gelesen" );
            DEBUG_PROC_3( stream_header->printInfo() );
                
            break;

        case CHUNK_CODE_STREAMFORMAT:
            DEBUG_TXT_3( "Stream-Format gefunden" );

            if( stream_header->StreamType == STREAM_CODE_VIDEO ) {
            
                bitmap_info = new BitmapInfo;
                // Bitmap-Info lesen
                if( !bitmap_info->ReadStream( avi_datei, length ) )
                    return false;

                DEBUG_TXT_3( "Bitmap-Info gelesen" );
                DEBUG_PROC_3( bitmap_info->printInfo() );
                
            } else if( stream_header->StreamType == STREAM_CODE_AUDIO ) {

                wave_format = new WaveFormat;
                // Wave-Format lesen
                if( !wave_format->ReadStream( avi_datei, length ) )
                    return false;

                DEBUG_TXT_3( "Wave-Format gelesen" );
                DEBUG_PROC_3( wave_format->printInfo() );
            }
            
            break;

        case CHUNK_CODE_EXTENDED_INDEX:
            DEBUG_TXT_3( "Erweiterten OpenDML-Index gefunden" );

            // Index lesen
            odml_index = new ODML_Index;
            if( !odml_index->ReadStream( avi_datei, length ) )
                return false;

            DEBUG_TXT_3( "Index gelesen" );
            DEBUG_PROC_3( odml_index->printInfo() );

            // Eintrge zum Index hinzufgen
            Index->AddIndex( *odml_index, avi_datei );

            file->open_dml_format = true;
            break;
            
        case CHUNK_CODE_PADDING_DATA:   // JUNK
            DEBUG_TXT_3( "Found padding data." );
            break;
            
        default:
            DEBUG_TXT_3( "unbekannter Chunk-Code 0x%08x gefunden", type );
        }

        avi_datei->skip( length + (length & 1) );
    }

    return true;
}






/**Alle noch nicht geschriebenen Indexeintrge in einen Indexchunk
  *ans Ende der Datei schreiben<br>
  *Rckgabewert: geschriebene Bytes
  */
size_t AviTrack::WriteIndex( void )
{
    ODML_Index *index;
    ODML_SuperIndexEntry *se;
    off_t offset;
    size_t length;

    // Index erstellen
    index = Index->ODML_GetIndex( TrackID() );

    // keine Eintrge zu schreiben
    if( index == NULL )
        return 0;
    
    // Datei positionieren
    // gehe ans Ende der Datei
    if( !avi_datei->seek( 0, SEEK_END ) )
        return 0;
    offset = avi_datei->pos();

    // Index schreiben
    avi_datei->write( IndexID() );
    avi_datei->write( index->Size() );
    length = index->Size();
    if( !index->WriteStream( avi_datei, length ) ) {
        size_t left = 8 + index->Size() - length;
        delete index;
        return left;
    }

    // Eintrag in Superindex erstellen
    se = new ODML_SuperIndexEntry;
    se->ChunkOffset = offset;
    se->DataSize = index->Size();
    se->Duration = index->Entries();  // time span in stream ticks ???

    if( odml_index == NULL ) {
        // neuen Superindex erstellen
        odml_index = new ODML_Index;

        odml_index->SetIndexType( AVI_INDEX_OF_INDEXES );
        odml_index->ChunkId = TrackID();
    }

    odml_index->AddIndex( se );

    delete index;
    return se->DataSize + 8;
}





/**Lngenangabe im Streamheader aktualisieren
  */
void AviTrack::UpdateLength( void )
{
    if( stream_header->SampleSize == 0 ) {
        // Azahl der Frames (Video und Audio bei SampleSize=0)
        stream_header->Length = Index->size();
    } else {
        // oder Lnge in Sekunden = Length/Rate (Audio)
        stream_header->Length = Index->DataBytes() / stream_header->SampleSize;
    }
}













/*****************************************************************************/
/***** Funktionen fr Audio- und Video-Tracks                            *****/
/*****************************************************************************/


/**Typ des Track setzen
  */
void AviTrack::SetType( uint32_t type )
{
    stream_header->StreamType = type;

    if( type == STREAM_CODE_VIDEO && bitmap_info == NULL ) {
        bitmap_info = new BitmapInfo;
    } else if( type == STREAM_CODE_AUDIO && wave_format == NULL ) {
        wave_format = new WaveFormat;
    }
}




/**Typ des Tracks
  */
uint32_t AviTrack::Type( void )
{
    return stream_header->StreamType;
}





/**Codecname fr Videotracks setzen
  */
void AviTrack::SetCodecName( const String &codec )
{
    uint32_t codec_code;
    uint8_t Z[4];
    
    strncpy( (char*)Z, codec.chars(), 4 );
    
    RIFF_ITypes::read_int( codec_code, Z );

    // Codec in Datenstrukturen schreiben
    stream_header->StreamHandler = codec_code;

    if( bitmap_info != NULL )
        bitmap_info->Compression = codec_code;
}






/**Codecname fr Videotracks
  */
const String AviTrack::CodecName( void )
{
    uint8_t Z[5];
    Z[4] = 0;
    String codec_name;

    if( bitmap_info != NULL ) {
        RIFF_ITypes::write_int( bitmap_info->Compression, Z );
    } else {
        RIFF_ITypes::write_int( stream_header->StreamHandler, Z );
    }

    codec_name = (char*)Z;

    return codec_name;
}






/**Framerate setzen
  */
void AviTrack::SetFrameRate( double rate, int byte_rate, int frame_samples, int frame_bytes )
{

    if( frame_bytes == 0 || byte_rate == 0 ) {
    
        // fr Videotracks oder bei Audiotracks mit variabler Bitrate    
        // Rate/Scale = Frames/sec (fr Video und Audio bei SampleSize=0)
        stream_header->Rate  = 1000000000l;
        stream_header->Scale = (uint32_t)((double)stream_header->Rate / rate); // Nanosec. pro Frame
        stream_header->SampleSize = 0;
        
        if( wave_format != NULL ) {
            // das hier ist fr den Windoof-Mediaplayer
            // der braucht das zum positionieren - absolut schwachsinnig!!!
            // Der Wert ist nicht zuverlssig, wenn sample_rate/frame_rate nicht ganzzahlig!
            // BlockAlign hat hier offensichtlich die Bedeutung von Samples pro Block ???
            // Ist diese Bedeutung irgendwo dokumentiert ???
            // Hier wird ein gemittelter Wert verwendet.
            wave_format->BlockAlign = (int16_t)( (double)wave_format->SamplesPerSec / rate + 0.5 );
        }
    } else {
    
        // fr Audiotracks mit fester Bitrate
        // bei einigen Codecs geht die Einstellung fr variable Bitrate nicht
        // Rate/Scale = (Bytes / sec) / SampleSize fr Audio bei SampleSize!=0 
        stream_header->Rate  = byte_rate;
        stream_header->Scale = frame_bytes;
        
        // Bie Bedeutung von SampleSize ist Blockgre, in der Daten gelesen werden knnen
        // Mit Samples hat das nichts zu tun ???
        stream_header->SampleSize = frame_bytes;
        
        if( wave_format != NULL ) {
            // einige Codecs brauchen diesen Wert
            // BlockAlign hat hier die Bedeutung von Bytes pro Block ???
            wave_format->BlockAlign = frame_bytes;
        }
    }   
}





/**echte Framerate
  */
double AviTrack::FrameRate( void )
{
    if( stream_header->SampleSize == 0 && stream_header->Scale > 0 ) {
        return (double)stream_header->Rate / (double)stream_header->Scale;
    } else if( stream_header->SampleSize == 0 ) {
        return (double)stream_header->Rate;
    }

    // Framerate aus Anzahl der Blcke und der Gesamtzeit berechnen
    if( stream_header->Length > 0 && Rate() > 0 )
        return (double)Index->size()  / ( (double)(Index->DataBytes() / stream_header->SampleSize) / Rate() );

    return 25.0;
}





/**Framerate oder Samplerate aus StreramHeader
  */
double AviTrack::Rate( void )
{
    if( stream_header->Scale > 0 ) {
        return (double)stream_header->Rate / (double)stream_header->Scale;
    }

    return (double)stream_header->Rate;
    
}





/**Bildgre setzen
  */
void AviTrack::SetSize( int x, int y )
{
    stream_header->right = x;
    stream_header->bottom = y;
    
    if( bitmap_info != NULL ) {
        bitmap_info->Width = x;
        bitmap_info->Height = y;
        bitmap_info->SizeImage = (x * y * bitmap_info->BitCount) / 3;
    }
}




/**Breite des Videobildes
  */
int AviTrack::Width( void )
{
    if( bitmap_info != NULL ) {
        return bitmap_info->Width;
    }

    return 0;
}




/**Hhe des Videobildes
  */
int AviTrack::Height( void )
{
    if( bitmap_info != NULL ) {
        return bitmap_info->Height;
    }

    return 0;
}




/**Anzahl der Bilder
  */
long AviTrack::Frames()
{
    return Index->size();
}




/**Formatcode fr Audiotracks
  */
uint32_t AviTrack::FormatCode( void )
{
    if( wave_format != NULL )
        return wave_format->Format;

    return 0;
}



/**Formatcode fr Audiotracks
  */
void AviTrack::SetFormatCode( uint32_t code )
{
    if( wave_format != NULL )
        wave_format->Format = code;
}





/**Samplerate fr Audiotracks
  */
uint32_t AviTrack::SampleRate( void )
{
    if( wave_format != NULL )
        return wave_format->SamplesPerSec;

    return 0;
}




/**Samplerate fr Audiotracks
  */
void AviTrack::SetSampleRate( uint32_t rate )
{
    if( wave_format != NULL )
        wave_format->SamplesPerSec = rate;
}





/**Anzahl der Knle fr Audiotracks
  */
void AviTrack::SetChannels( int n )
{
    if( wave_format != NULL )
        wave_format->Channels = n;
}




/**Anzahl der Knle fr Audiotracks
  */
int AviTrack::Channels( void )
{
    if( wave_format != NULL )
        return wave_format->Channels;

    return 0;
}




/**Anzahl der Bits pro Sample
  */
void AviTrack::SetBitsPerSample( int n )
{
    if( wave_format != NULL )
        wave_format->BitsPerSample = n;
}




/**Anzahl der Bits pro Sample
  */
int AviTrack::BitsPerSample( void )
{
    if( wave_format != NULL )
        return wave_format->BitsPerSample;

    return 0;
}






/**Block Align - Blockgre in Bytes
  */
int AviTrack::BlockAlign( void )
{
    if( wave_format != NULL )
        return wave_format->BlockAlign;

    return 0;
}




/**Bytes pro Sekunde fr Audiotracks
  */
uint32_t AviTrack::BytesPerSec( void )
{
    if( stream_header->SampleSize > 0 && stream_header->Scale > 0 ) {
        return stream_header->Rate / stream_header->Scale * stream_header->SampleSize;
    }
    
    if( wave_format != NULL ) {
        return  wave_format->AvgBytesPerSec;   
    }
    
    return 0;
}



/**Anzahl der Samples
  */
uint64_t AviTrack::Samples( void )
{
    if( wave_format == NULL || Rate() == 0 )
        return 0;
        
    if( stream_header->SampleSize > 0 ) {
        return (uint64_t)( (double)stream_header->Length / Rate() * wave_format->SamplesPerSec );
    }
    
    return (uint64_t)( (double)Index->size() / Rate() * wave_format->SamplesPerSec );
}


    




/**Position im Stream fr den nchsen Lesezugriff festlegen
  */
void AviTrack::SetNextRead( uint64_t block, off_t offset )
{
    next_chunk = block;
    next_chunk_offset = offset;
}



    

/**Daten lesen
  */
bool AviTrack::ReadData( uint8_t *data, size_t bytes )
{
    uint64_t offset = 0;
    int64_t size = 0, n = 0;
    
  
    // lese bytes Zeichen ein
    // wenn in einem Block nicht genug Daten vorhanden sind, lese im
    // nchsten Block des gleichen Tracks weiter
    do {
        if( next_chunk >= Index->size() )
                return false;

        // Dateiposition des Datenblocks
        do {
            offset = Index->at(next_chunk)->AbsFileOffset;
            size   = Index->at(next_chunk)->ChunkSize;
            if( next_chunk_offset >= size ) {
                next_chunk_offset -= size;
                next_chunk++;
            }

            if( next_chunk >= Index->size() )
                return false;
            
        } while( next_chunk_offset >= size );

        // Datei positionieren
        // berspringe den Header des Blocks (8 Byte)
        // berspringe Daten am Anfang bis zur aktuelle Position
        if( !avi_datei->seek( offset + 8 + next_chunk_offset ) )
            return false;

        // lese Daten aus dem Block    
        n = bytes > (size - next_chunk_offset) ? (size - next_chunk_offset) : bytes;
        if( !avi_datei->read( data, n ) )
            return false;

        bytes -= n;
        data += n;
        next_chunk_offset += n;

        if( next_chunk_offset >= size ) {
            next_chunk_offset = 0;
            next_chunk++;
        }
        
    } while( bytes > 0 );

    return true;    
}






/**Daten schreiben
  */
bool AviTrack::WriteData( uint8_t *data, size_t bytes, uint32_t flags )
{
    uint64_t offset;
    uint8_t c = 0;

    // Schreibzugriff in Datei initialisieren
    if( !file->write_to_movi_chunk( bytes + (bytes&1) + 8 ) )
        return false;
    
    // Datei positionieren
    // gehe ans Ende der Datei
    if( !avi_datei->seek( 0, SEEK_END ) )
        return false;
    offset = avi_datei->pos();

    if( !avi_datei->write( TrackID() ) )
        return false;

    // Datengre schreiben
    if( !avi_datei->write( (uint32_t)bytes ) )
        return false;

    // Daten schreiben
    if( !avi_datei->write( data, bytes ) )
        return false;
        
    // es muss immer eine gerade Anzahl von Bytes geschrieben werdren
    // wenn bytes ungerade ist, wird noch eine 0 angehngt
    if( bytes & 1 ) {
        if( !avi_datei->write( &c, 1 ) )
            return false;
    }

    // Indexeintrag erstellen
    Index->AddEntry( offset, bytes, flags );

    return true;
}






/**Track-ID
  */
uint32_t AviTrack::TrackID( void )
{
    char c[5];
    uint8_t z[4];
    uint32_t id = 0;
    
    // Track-ID schreiben
    if( Type() == STREAM_CODE_VIDEO ) {
        sprintf( c, "%02xdc", track_num );
    } else if( Type() == STREAM_CODE_AUDIO ) {
        sprintf( c, "%02xwb", track_num );
    } else {
        DEBUG_TXT( "Kann unbekannten Track-Typ 0x%08x nicht schreiben", Type() );
        return 0;
    }
    memcpy( z, c, 4 );

    RIFF_ITypes::read_int( id, z );

    return id;
}






/**ID fr einen OpenDML Index-Chunk
  */
uint32_t AviTrack::IndexID( void )
{
    char c[5];
    uint8_t z[4];
    uint32_t id = 0;

    // Track-ID schreiben
    sprintf( c, "ix%02x", track_num );
    memcpy( z, c, 4 );

    RIFF_ITypes::read_int( id, z );

    return id;
}
