/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/


#include "Frame_Index.h"
#include "types/AVI_Codes.h"
#include "ODML_Index.h"
#include "debug.h"





/*****************************************************************************/
/***** Index-Eintrag                                                     *****/
/*****************************************************************************/
FrameIndexEntry::FrameIndexEntry()
{
    AbsFileOffset = 0;
    ChunkSize = 0;
    Flags = 0;
}




bool FrameIndexEntry::IsKeyframe( void )
{
    return Flags & AVI_INDEX_KEYFRAME;
}











/*****************************************************************************/
/***** Index-verwaltung                                                  *****/
/*****************************************************************************/

FrameIndex::FrameIndex()
{
    bytes = 0;
    keyframes = 0;
    entries_written = 0;
}




FrameIndex::~FrameIndex()
{
    // Index lschen
    while( size() > 0 ) {
        delete back();
        pop_back();
    }
}







        

/**Einen Indexeintrag anhand eines vorhandenen AviIndexEntry erstellen.
  */
void FrameIndex::AddEntry( AviIndexEntry &e, uint64_t offset )
{
    FrameIndexEntry *entry;

    entry = new FrameIndexEntry;

    entry->AbsFileOffset = e.ChunkOffset + offset;
    entry->ChunkSize     = e.ChunkLength;
    entry->Flags         = e.Flags;

    push_back( entry );

    bytes += entry->ChunkSize;

    if( entry->IsKeyframe() )
        keyframes++;
}





/**Einen Indexeintrag erstellen
  */   
void FrameIndex::AddEntry( uint64_t offset, uint32_t length, uint32_t flags )
{
    FrameIndexEntry *entry;

    entry = new FrameIndexEntry;

    entry->AbsFileOffset = offset;
    entry->ChunkSize     = length;
    entry->Flags         = flags;

    push_back( entry );

    bytes += entry->ChunkSize;

    if( entry->IsKeyframe() )
        keyframes++;
}




/**Einen Indexeintrag anhand eines vorhandenen OpenDML-Index-Eintrags erstellen.
  */
void FrameIndex::AddEntry( ODML_StandardIndexEntry &e, uint64_t offset )
{
    FrameIndexEntry *entry;

    entry = new FrameIndexEntry;

    entry->AbsFileOffset = e.DataOffset + offset;
    entry->ChunkSize     = e.DataSize & 0x7FFFFFFF;
    // Bit 31 ist gesetzt, wenn dies _kein_ Keyframe ist
    entry->Flags         = (e.DataSize & (1<<31)) ? 0 : AVI_INDEX_KEYFRAME;

    push_back( entry );

    bytes += entry->ChunkSize;

    if( entry->IsKeyframe() )
        keyframes++;
}








/**Einen OpenDML-Index hinzufgen.
  */
bool FrameIndex::AddIndex( ODML_Index &index, RIFF_File *file )
{
    unsigned int i;
    
    if( (index.IndexType & 0x0f) == AVI_INDEX_OF_CHUNKS ) {
        // Index ist ein Chunk-Index
        // hnge Eintrge an den Index an
        for( i=0; i < index.Entries(); i++ ) {
            AddEntry( index.GetFrame(i), index.BaseOffset - 8 );
        }

        DEBUG_TXT( "%i Eintrge zum Index hinzugefgt.", index.Entries() );
        return true;
        
    } else if( (index.IndexType & 0x0f) == AVI_INDEX_OF_INDEXES ) {
        // Index zeigt auf weitere Indizes
        ODML_Index *idx = NULL;
            
        // speichere Dateiposition
        off_t pos = file->pos();
        
        for( i=0; i < index.Entries(); i++ ) {
            idx = new ODML_Index;

            if( index.GetIndex(i).ChunkOffset == 0 )
                continue;
                
            // Index einlesen
            DEBUG_TXT( "Lese Index-Bolck bei Offset 0x%08x.", index.GetIndex(i).ChunkOffset );
            // Im Superindex zeigt ChunkOffset auf den Anfang des Chunks
            file->seek( index.GetIndex(i).ChunkOffset + 8 );
            size_t length = index.GetIndex(i).DataSize;
            if( !idx->ReadStream( file, length ) ) {
                delete idx;
                file->seek( pos );
                return false;
            }

            // Index anhngen
            if( !AddIndex( *idx, file ) ) {
                delete idx;
                file->seek( pos );
                return false;
            }
        }

        delete idx;

        // zu alter Position zurckgehen
        file->seek( pos );

        return true;
    }

    return false;
}







/**Einen OpenDML-Index erstellen
  */
ODML_Index *FrameIndex::ODML_GetIndex( uint32_t chunk_id )
{
    // Anzahl der Eintrge
    unsigned long n = size() - entries_written;
    if( n <= 0 )
        return NULL;

    FrameIndexEntry *e;
    ODML_Index *index = new ODML_Index;

    // Werte initialisieren
    index->SetIndexType( AVI_INDEX_OF_CHUNKS );
    index->ChunkId = chunk_id;
    index->BaseOffset = (*(begin() + entries_written))->AbsFileOffset;

    // Eintrge hinzufgen
    while( entries_written < size() ) {
        ODML_StandardIndexEntry *entry = new ODML_StandardIndexEntry;

        e = *(begin() + entries_written);
        // DataOffset zeigt auf den Beginn der Framedaten und
        // nicht auf den Beginn des Chunks.
        entry->DataOffset = (uint32_t)( e->AbsFileOffset + 8 - index->BaseOffset );
        entry->DataSize   = e->ChunkSize;
        // Bit 31 ist gesetzt, wenn dies _kein_ Keyframe ist
        if( !e->IsKeyframe() )
            entry->DataSize |= 1<<31;

        index->AddFrame( entry );
        entries_written++;
    }

    return index;
}







/**Sind die Keyframe-Informationen vorhanden?
  */
bool FrameIndex::KeyframesOk( void )
{
    return keyframes > 0;
}
