/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/




#include <string.h>
#include <assert.h>

#include "FilePuffer.h"



FilePuffer::FilePuffer( int file, size_t gr )
{
    groesse = gr;
    puffer = new unsigned char[groesse];
    position = 0;
    eof = false;
    stream = file;
    file_offset = 0;
    file_length = 0;
    
    // Dateigre bestimmen
    if( CanSeek() ) {
        file_length = lseek64( stream, 0, SEEK_END);
        lseek64( stream, file_offset, SEEK_SET );
    }
}




FilePuffer::~FilePuffer()
{
    delete [] puffer;
}




/**Daten in Puffer laden<br>
  *Rckgabe:  1  Daten gelesen<br>
  *           0  keine Daten gelesen / eof<br>
  *          -1  Fehler
  */
int FilePuffer::NeedData( size_t n)
{
    if( n <= bytes )
        return 1;

    //Puffer fllen
    int stat = Fuellen();
    if( stat <= 0 )
        return stat;

    if( n <= bytes )
        return 1;

    return 0;

}






/**Puffer weiterschieben
  */
void FilePuffer::ClearData( size_t n )
{
    if( n >= bytes ) {
        // Puffer neu fllen
        Fuellen();
        if( n >= bytes ) {
            bytes = 0;
            position = 0;
            return;
        }
    }

    position += n;
    if( position >= groesse )
        position -= groesse;

    bytes -= n;
}







/**Puffer leeren
  */
void FilePuffer::Reset( void )
{
    bytes = 0;
    position = 0;
}






/**Finde "muster" im Puffer<br>
  *Rckgabe: Anfangsposition<br>
  *          oder -1, wenn nicht gefunden
  */
long FilePuffer::Finden( size_t offset, const unsigned char *muster, size_t n )
{
    if( bytes <= offset )
        Fuellen();
        
    int stat = FifoPuffer::Finden( offset, muster, n );

    if( stat < 0 ) {
        // Puffer neu fllen
        if( eof )
            return -1;
        Fuellen();
        return FifoPuffer::Finden( offset, muster, n );
    }

    return stat;
}







long FilePuffer::Finden( size_t offset, uint32_t muster, uint32_t maske, uint32_t &gefunden )
{
    if( bytes <= offset )
        Fuellen();

    int stat = FifoPuffer::Finden( offset, muster, maske, gefunden );

    if( stat < 0 ) {
        // Puffer neu fllen
        if( eof )
            return -1;
        Fuellen();
        return FifoPuffer::Finden( offset, muster, maske, gefunden );
    }

    return stat;
}







/**Zeiger auf Datenfeld an Stelle "offset"
  */
uint8_t FilePuffer::Data( size_t offset )
{
    //assert( offset < groesse );
        // Puffer fllen, wenn ntig
    if( offset >= (size_t)Bytes() )
        Fuellen();
        
    size_t pos = position + offset;
    if( pos >= groesse )
        pos -= groesse;
        
    return puffer[pos];
}








/**Daten aus Puffer kopieren
  */
int FilePuffer::GetData( size_t offset, uint8_t *data, size_t n )
{
    if( n + offset > bytes )
        Fuellen();
    if( n + offset > bytes ) {
        if( eof )
            return 0;
        return -1;
    }

    size_t pos = 0;

    long bytes = n;
    long anfang = position + offset;
    
    if( anfang >= (long)groesse )
        anfang -= groesse;

    if( anfang + bytes > (long)groesse )
        bytes = groesse - anfang;

    if( bytes > 0 ) {
        memcpy( data, puffer + anfang, bytes );
        pos += bytes;
    }

    bytes = n - bytes;
    
    if( bytes > 0 ) {
        memcpy( data + pos, puffer, bytes );
    }
    
    return 1;
}




/**Eingabedatei positionieren
  */
int FilePuffer::Seek( off_t offset )
{
    off64_t alter_offset = file_offset;
    
    if( !CanSeek() )
        return -1;
    
    if( offset > file_length ) {
        eof = true;
        return 0;
    }
            
    if( file_offset - bytes == offset )
        return 1;
        
    // Puffer leeren
    Reset();

    // positionieren
    file_offset = lseek64( stream, offset, SEEK_SET );

    // eof zurcksetzen
    eof = false;

    if( file_offset < offset ) {
        file_offset = alter_offset;
        return -1;
    }

    // Puffer neu fllen
    if( Fuellen() < 0 )
        return -1;

    return 1;
}





/**Offset einstellen
  */
void FilePuffer::set_offset( size_t bytes_read )
{
    // keine echte Datei
    if( !CanSeek() ) {
        file_offset += bytes_read;
        return;
    }

    // echte Datei
    file_offset = lseek64( stream, 0, SEEK_CUR );
}





/**Gre der Datei
  */
off_t FilePuffer::FileSize( void )
{
    if( !CanSeek() )
        return -1;
        
    return file_length;
}













/**Ring-Puffer aus Eingabedatei fllen<br>
  *Rckgabe:  1  Daten gelesen<br>
  *           0  keine Daten gelesen / eof<br>
  *          -1  Fehler
  */
int FilePuffer::Fuellen( void )
{
    long read_bytes=0, read_bytes2=0;
    int stat = 1;

    // eof zurcksetzern
    if( eof )
        return 0;

    // Fehlerstatus lschen und Datei neu positionieren
    if( CanSeek() && file_offset != lseek64( stream, 0, SEEK_CUR ) ) {
        if( file_offset >= file_length ) {
            eof = true;
            return 0;
        }
        
        if( lseek64( stream, file_offset, SEEK_SET ) < 0 ) {
            return -1;
        }
    }

    // Endposition der Daten
    size_t end_pos = position + bytes;
    if( end_pos >= groesse )
        end_pos -= groesse;

    // Restlnge bis zum Ende
    size_t rest = groesse - bytes;
    if( end_pos + rest > groesse )
        rest = groesse - end_pos;

    // Restlnge vom Anfang
    size_t anfang = groesse - bytes - rest;

    // Der Puffer bildet einen Ring. Nachdem der Bereich bis zum Ende aufgefllt
    // ist, wird am Anfang weitergemacht.
    //
    //  Puffer:   |__________________###################______________|
    //                               |<---- bytes ---->|
    //            |<---- anfang ---->|                 |<--- ende --->|
    //                               |>
    //                    daten werden hier gelesen
    //
    // oder       |##############____________#########################|
    //            ------------->|            |<-------------- bytes ---
    //                          |<-- ende -->|
    //                                       |>
    //                                     lesen
    // einen Block am Ende vom Puffer enlesen
    if( rest > 0 ) {
        read_bytes = read( stream, (char*)&(puffer[end_pos]), rest );

        if( read_bytes < rest ) {
            if( read_bytes > 0 ) {
                eof = true;
                stat = 1;
            } else if( read_bytes == 0 ) {
                eof = true;
                stat = 0;
            } else {
                stat = -1;
                read_bytes = 0;
            }
        }
        bytes += read_bytes;
    }

    // einen Block am anfang vom Puffer einlesen
    if( anfang > 0 && stat > 0 && ! eof ) {
        read_bytes2 = read( stream, (char*)puffer, anfang );
        
        if( read_bytes2 < rest ) {
            eof = true;
            if( read_bytes2 < 0 ) {
                stat = -1;
                read_bytes2 = 0;
            } else if( read_bytes > 0 || read_bytes2 > 0 ) {
                stat = 1;
            } else {
                stat = 0;
            }
        }
        bytes += read_bytes2;
    }

    set_offset( read_bytes + read_bytes2 );
    return stat;
}
