/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "Mpeg.h"
#include "Mpeg_Vtr.h"
#include "Mpeg_Atr.h"
#include "MpegCodes.h"
#include "debug.h"
#include "VideoFile/codecplugins.h"
#include "VideoFile/translate.h"


#define MAX_OFFSET 3000
#define MPLEX_PACK_BUF 50

/***** Konstruktor / Destruktor *****/
MpegIO::MpegIO( const String datei, IO_Mode mode, int mpegtype )
:VideoFile( datei, "Mpeg", mode )
{
    // Variablen initialisieren
    mpeg_datei = NULL;
    puffer = NULL;
    system = NULL;
    wsystem = NULL;
    mpeg_start_offset = 0;
    mpeg_system_type = mpegtype;
        
    for( int i=0; i<16; i++ ) {
        vstream[i] = NULL;
        astream[i] = NULL;
    }

    // Datei ffnen
    if( iomode == read ) {
        if( datei != "" ) {
            mpeg_datei = open( datei.chars(), O_RDONLY );
            if( mpeg_datei < 0 ) {
                ErrorStatus() = Errors::Open;
                throw( ErrorStatus() );
            }
            puffer = new FilePuffer( mpeg_datei, FILE_PUFFER_GR );
        } else {
            puffer = new FilePuffer( STDIN_FILENO, FILE_PUFFER_GR );
        }

        
        // prfe Dateityp
        if( !sync_block_start() ) {
            ErrorStatus() = Errors::FileType;
            throw( ErrorStatus() );
        }

    } else if( iomode == write ) {
        
        if( datei != "" ) {
            mpeg_datei = open( datei.chars(),  O_WRONLY | O_TRUNC | O_CREAT, 0644 );
            if( mpeg_datei < 0 ) {
                ErrorStatus() = Errors::Open;
                throw( ErrorStatus() );
            }
        } else {
            mpeg_datei = STDIN_FILENO;
        }

        wsystem = new System_Write( mpeg_datei, mpeg_system_type );

    } else {
        ErrorStatus() = Errors::IO_Mode;
        throw( ErrorStatus() );
    }

}




MpegIO::~MpegIO()
{
    // restliche Packete in Stream schreiben
    if( iomode == write ){
        multiplex_puffer();
    }
        
    // Tracks schlieen
    deleteTracks();

    for( int i=0; i<16; i++ ) {
        delete vstream[i];
        delete astream[i];
    }
     
    delete system;

    delete puffer;

}








/***** Video-Track *********************************************/
// Video codec initialisieren
// Rckgabewert:  Zeiger auf Videotrack
V_Track *MpegIO::openVideoTrack( double Rate, int x, int y, const char *Codec )
{
    if( iomode == write ) {
        return new MpegVtr( this, Rate, x, y, Codec );
    }
    return NULL;
}



// Track ffnen
V_Track *MpegIO::openVideoTrack( int track )
{
    if( iomode != read )
        return NULL;

    if( track >= 16 || vstream[track] == NULL ) {
        ERROR_TXT( _("Video track %i not found!"), track );
        return NULL;
    }

    return new MpegVtr( this, track );
}





/***** Audio-Track *********************************************/
// Audio codec initialisieren
// Rckgabewert:  Zeiger auf Audiotrack
A_Track *MpegIO::openAudioTrack( int Sample_rate, int Channels, const char *Codec )
{
    if( iomode == write )
        return new MpegAtr( this, Sample_rate, Channels, Codec );
    return NULL;
}



// Track ffnen
A_Track *MpegIO::openAudioTrack( int track )
{
    if( iomode != read )
        return NULL;

    if( track >= 16 || astream[track] == NULL ) {
        ERROR_TXT( _("Audio track %i not found"), track );
        return NULL;
    }

    return new MpegAtr( this, track );
}








/***** Anfang des Streams suchen **********************************************/
bool MpegIO::sync_block_start( void )
{
    int stat;
    char buf[16];
    off_t max_offset = MAX_OFFSET;

    // Dateityp prfen
    stat = puffer->NeedData(16);
    if( stat <= 0 )
        return false;
    
    // CDXAfmt testen
    if( puffer->GetData( 0, (uint8_t *)buf, 4 ) < 1 )
        return false;
    if( strncmp( buf, "RIFF", 4 ) == 0 ) {
        if( puffer->GetData( 8, (uint8_t *)buf, 7 ) < 1 )
            return false;
        if( strncmp( buf, "CDXAfmt", 7 ) == 0 ) {
            // CDXA-Format
            // suche nach Start der Daten - maximaler Offset ist die Puffergre
            mpeg_start_offset = puffer->Finden( 16, CHUNK_START, 3 );
            if( mpeg_start_offset < 0 )
                return false;
            max_offset += mpeg_start_offset;
        } else {
            // unbekanntes RIFF-Format
            return false;
        }
    }

    // Suche nach einem Pack-Header oder einem PES-Header
    mpeg_start_offset = puffer->Finden( mpeg_start_offset, CHUNK_START, 3 );
    if( mpeg_start_offset < 0 || mpeg_start_offset > max_offset ) {
        return false;
    }
    
    if( puffer->Data( mpeg_start_offset + 3 ) <= 0xB9 ) {
        // kein System Stream
        return false;
    }
    
    puffer->ClearData( mpeg_start_offset );    
    system = new System( puffer, mpeg_start_offset );

    // Das erste Paket einlesen
    stat = system->NextPack();
    if( stat <= 0 )
        return false;

    mpeg_system_type = system->MpegType();

    // Puffer fllen und Tracks finden
    stat = FillStreamPuffer();
    if( stat <= 0 )
        return false;
    
    return true;
}





/**Stream-Puffer aus Eingabedatei fllen<br>
  *Rckgabe:  1  Daten gelesen<br>
  *           0  keine Daten gelesen / eof<br>
  *          -1  Fehler
  */
int MpegIO::FillStreamPuffer( MpegTrack *tr )
{
    size_t bytes = 0;
    uint8_t *data;
    int i = 0, j = 0, stat = 0;
    StreamPuffer *stream = NULL;
    bool ein_track = false;
    System *sys = system;

    if( tr != NULL && tr->system != system ) {
        // Stream Puffer hat eingenen System-Handler
        // flle nur diesen einen Stream-Puffer
        stream = tr->stream;
        sys = tr->system;
        ein_track = true;
        
    }

    // i: Pakete verworfen
    // j: Pakete gelesen
     
    while( j < 100 * (ein_track? 1:(VideoTracks() + AudioTracks() + 2)) ) {
        // Abbruchbedingung - Stream anscheinend fehlerhaft
        if( i++ > 1000 )
            return -1;
            
        if( !ein_track ) {
            // Track auswhlen
            stream = finde_stream_puffer( sys->Tracktype(), sys->Tracknumber() );
        } else {
            // Nur Pakete fr gewhlten Stream kopieren
            if( sys->Tracktype() != stream->Tracktype() || sys->Tracknumber() != stream->Tracknumber() ) {
                // gehe zum nchsten Paket               
                goto weiter;
            }
        }
        
        // kopiere Daten, wenn noch Platz im Puffer ist
        if( stream != NULL && (!puffer->CanSeek() || tr != NULL) && stream->FreePacks() > 0 ) {
            // Datenpaket kopieren
            bytes = sys->Bytes();
            stream->AddPack( data, bytes, sys->Timestamp(), sys->FileOffset() );
            sys->GetPackData( data, bytes );
            // DEBUG_TXT_3( "Timestamp: %f", system->Timestamp() );
            
            i=0;
            j++;
        } else if ( stream != NULL && puffer->CanSeek() ) {
            // Datei nur nach Tracks durchsuchen
            // erhhe Zhler der gelesen Blcke
            i=0;
            j++;
        }

weiter:
        // gehe zum nchsten Paket
        stat = sys->NextPack();
        if( stat <= 0 && j == 0 )
            return stat;
        else if( stat <= 0 && j > 0 )
            return 1;
    }

    return 1;
}







/**kleinste Anzahl der Packs aus allen Puffern
  */
/*
int MpegIO::stream_puffer_packs( void )
{
    int n = PACK_PUFFER + 1;
    
    for( int i=0; i<16; i++ ) {
        if( vstream[i] != NULL && vstream[i]->Packs() < n )
            n = vstream[i]->Packs();

        if( astream[i] != NULL && astream[i]->Packs() < n )
            n = astream[i]->Packs();
    }

    if( n > PACK_PUFFER )
        n = 0;
        
    return n;
}
*/



/**Stream Puffer zu Track finden / anlegen
  */
StreamPuffer *MpegIO::finde_stream_puffer( int type, int num )
{
    StreamPuffer **stream;
    int n=0;
    
    if( type == TRACK_CODE_MPEG ) {
        stream = vstream;
    } else if( type == TRACK_CODE_MP3 || type == TRACK_CODE_AC3 ) {
        stream = astream;
    } else {
        return NULL;
    }

    for( int i=0; i<16; i++ ) {
        if( stream[i] == NULL ) {
            //neuen Puffer anlegen
            stream[i] = new StreamPuffer( PACK_PUFFER );
            if( num >= 0 ) {
                stream[i]->SetTrackID( type, num );
                DEBUG_TXT( _("Found track typ %i number %i."), type, num );
            } else {
                stream[i]->SetTrackID( type, n );
            }
            return stream[i];

            //Anzahl der Tracks aktualisieren
            if( stream == vstream )
                SetVideoTracks( n );
            else if( stream == astream )
                SetAudioTracks( n );
            
        } else {
            if( num >= 0 && stream[i]->Tracktype() == type && stream[i]->Tracknumber() == num ) {
                return stream[i];
            } else if( num < 0 && stream[i]->Tracktype() == type ) {
                n++;
            }
        }
    }

    return NULL;
}






/**Stream-Puffer zu einem Stream multiplexen
  */
int MpegIO::multiplex_puffer( MpegTrack *tr )
{
    double next_timestamp = -1;
    int packs;
    StreamPuffer **stream, *next_stream=NULL;
    
    do{
        packs = 0;
        next_timestamp = -1;
        next_stream = NULL;
        
        // ersten Timestamp von allen Paketen finden
        stream = astream;
        for( int i=0; i<16; i++ ) {
            if( stream[i] == NULL ) {
                break;
            }

            if( stream[i]->Packs() > 0 && ( next_timestamp < 0 || stream[i]->Timestamp() < next_timestamp ) ) {
                next_timestamp = stream[i]->Timestamp();
                next_stream = stream[i];
            }
            packs += stream[i]->Packs();
        }
        
        // ersten Timestamp von allen Paketen finden
        stream = vstream;
        for( int i=0; i<16; i++ ) {
            if( stream[i] == NULL ) {
                break;
            }

            if( stream[i]->Packs() > 0 && (next_timestamp < 0 || stream[i]->Timestamp() < next_timestamp) ) {
                next_timestamp = stream[i]->Timestamp();
                next_stream = stream[i];
            }
            packs += stream[i]->Packs();
        }

        // nchstes Pack schreiben - nur wenn in allen Streams
        // mehr als MPLEX_PACK_BUF Pakete sind
        if( next_stream != NULL && ( packs >= MPLEX_PACK_BUF || (tr==NULL && packs > 0) ) ) {
            wsystem->WritePack( next_stream );
        }
    } while( packs >= MPLEX_PACK_BUF || (tr==NULL && packs > 0) );

    return 1;
}
