/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <assert.h>
#include <unistd.h>

#include "MpegTrack.h"
#include "Mpeg.h"
#include "debug.h"
#include "VideoFile/translate.h"




MpegTrack::MpegTrack( MpegIO *File, StreamPuffer *buf, int track )
: Track( File, track )
{
    init( File, buf );
}



MpegTrack::MpegTrack( MpegIO *File, double rate, const char *Codec_name, int track )
: Track( File, track, Codec_name )
{
    SetRate( rate );
    init( File, NULL );
}



MpegTrack::~MpegTrack()
{
    if( IOMode() == VideoFile::read && puffer->CanSeek() ) {
        delete system;
        delete puffer;
    }
}




void MpegTrack::init( MpegIO *File, StreamPuffer *buf )
{
    max_pos = 0;

    stream = buf;
    file = File;
    
    if( File->iomode == VideoFile::read && File->puffer->CanSeek() ) {
        puffer = new FilePuffer( File->mpeg_datei, FILE_PUFFER_GR );
        puffer->ClearData( File->mpeg_start_offset );

        system = new System( File->system, puffer, File->mpeg_start_offset );
        system->NextPack();

        //Streampuffer lschen und neu fllen
        stream->Reset();
        File->FillStreamPuffer( this );

    } else {
        puffer = File->puffer;
        system = File->system;
    }
}







int MpegTrack::Seek( long block )
{
    if( file->iomode != VideoFile::read )
        return 0;
    
    DEBUG_TXT( _("Seek to block %i"), block );

    // nur positionieren, wenn ntig
    if( block == Position() )
        return 1;

    //suche nach einem Eintrag in der Keytable
    off_t offset = -1;
    long key = key_table.GetKeyFrame( block, offset );
    int stat = 1;

    if( offset >= 0 && key >= 0) {
        stat=seek_time( key / Rate(), offset );
    } else {
        block -= 25;
        if( block < 0 )
            block = 0;
            
        stat=seek_time( block / Rate(), -1 );
    }

    if( stat < 0 ) {
        ERROR_TXT( _("Can not seek stream!") );
        return 0;
    }

    // Tatschliche Position des Streams im Puffer bestimmen
    SetPosition( (uint32_t)(stream->Timestamp() * Rate() ) );

    //durchsuche Stream vom Anfang, wenn direkte Positionierung nicht geht
    if( stat == 0 && Position() > block ) {
        // gehe zum Anfang
        if( (stat = seek_time( 0, 0 )) > 0 )
            SetPosition( 0 );
    }

    return 1;
}









/**Stream Puffer positionieren<br>
  *Rckgabe:  1  erfolgreich positioniert<br>
  *           0  kann nicht positionieren - andere Methode probieren!<br>
  *          -1  Fehler
  */
int MpegTrack::seek_time( double sec, off_t offset )
{
    //suche Block im Puffer
    if( stream->HaveTimestamps() && sec > stream->Timestamp() ) {
        while( stream->Packs() > 0 && sec > stream->Timestamp() ) {
            stream->ClearPack();
        }
        if( sec <= stream->Timestamp() )
            return 1;
    } else if( !system->HaveTimestamps() && offset < 0 && puffer->CanSeek() ) {
        // Track mu selber positionieren - nur bei echten Dateien mglich!
        // Sollte nie vorkommen, da ein System-Stream immer Timestamps haben mu!!!
        return 0;
    }

    // mache nur fr echte Dateien weiter
    if( !puffer->CanSeek() ) {
        return -1;
    }

    if( offset >= 0 ) {
        if( system->SeekOffset( offset ) <= 0 )
            return -1;

    } else if( system->HaveTimestamps() ) {
        if( system->SeekTime( sec ) <= 0 )
            return -1;
    } else {
        return -1;
    }

    // Puffer leeren und neu fllen
    stream->Reset();

    //Synchronisiere Puffer
    if( stream->HaveTimestamps() && sec > stream->Timestamp() ) {

        if( file->FillStreamPuffer( this ) <= 0 )
            return -1;

        while( stream->Packs() > 0 && sec > stream->Timestamp() ) {
            stream->ClearPack();
        }
        
        if( sec <= stream->Timestamp() )
            return 1;

        // kann Timestamp nicht im Puffer finden -> Tracks leigen mehr
        // Pakete als Puffergre auseinander
        ERROR_TXT( _("Error: can not sync stream %02x - buffer too small."),
            stream->Tracktype() & stream->Tracknumber() );
    }
        
    if( file->FillStreamPuffer( this ) > 0 )
        return 1;

    return -1;
}
