/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>

#include "Mpeg_Atr.h"
#include "MpegCodes.h"
#include "debug.h"
#include "VideoFile/codecplugins.h"
#include "VideoFile/translate.h"

#define MAX_OFFSET 1024

// Puffer festlegen - maximale Gre 1152 Samples/Block * 320 kbps bei 8 kHz Mono
// mit Faktor 2 Sicherheit
#define PUFFER_GR  1152 / 8000 * 320000 * 2








/*****************************************************************************/
/***** Audiotrack-Funktionen *************************************************/
/*****************************************************************************/

MpegAtr::MpegAtr( MpegIO *File, int Sample_rate, int Channels, const char *Codec_name, int track )
: Track( File, track, Codec_name ), MpegTrack( File, 25, Codec_name ), A_Track( File, Sample_rate, Channels, Codec_name )
{
    puffer_gr = PUFFER_GR;
    puffer = new uint8_t[puffer_gr];
    framesize = 1;
    layer = 1;
    mp3_block_start = 0;

    if( CodecName() == "" ) {
        SetCodecName( "mp2" );
    }

    codec = AV_Codecs->CreateACodec( this, "Mpeg", CodecName() );

    if( codec != NULL && codec->CodecName() == "ac3" ) {
        stream = file->finde_stream_puffer( TRACK_CODE_AC3, track );
    } else {
        stream = file->finde_stream_puffer( TRACK_CODE_MP3, track );
    }

    if( codec == NULL || stream == NULL ) {
        ERROR_TXT( _("Audio codec \"%s\" not implemented!"), CodecName().chars() );
        delete codec;
        codec = new A_Codec( this );
    }
}





MpegAtr::MpegAtr( MpegIO *File, int track )
: Track( File, track ), MpegTrack( File, File->astream[track], track ), A_Track( File, track )
{
    puffer_gr = PUFFER_GR;
    puffer = new uint8_t[puffer_gr];
    framesize = 1;
    layer = 1;
    mp3_block_start = 0;


    if( stream->Tracktype() == TRACK_CODE_MP3 ) {
        // MP3 //
        // MP3-Header finden   
        bool stat = sync_block_start_mp3();
        assert( stat );

        switch( layer ) {
        case 1:
            SetCodecName( "mp1" );
            break;
        case 2:
            SetCodecName( "mp2" );
            break;
        default:
            SetCodecName( "mp3" );
        }
        
        codec = AV_Codecs->CreateACodec( this, "Mpeg", CodecName() );
    } else if( stream->Tracktype() == TRACK_CODE_AC3 ) {
        // AC3 //
        // AC3-Header finden
        bool stat = sync_block_start_ac3();
        assert( stat );

        SetCodecName( "ac3" );
        codec = AV_Codecs->CreateACodec( this, "Mpeg", CodecName() );        
    } else {
        SetCodecName( "" );
        codec = NULL;
    }
     
    if( codec == NULL ) {
        ERROR_TXT( _("Audio codec \"%s\" not implemented!"), CodecName().chars() );
        codec = new A_Codec( this );
    }

    // vom Codec gendertes Format bernehmen
    (A_Format)*this = (A_Format)*codec;
    
    //block_rate = (double)sample_rate / (double)framesize;
    SetFrames( (int64_t)(system->TotalTime() * Rate()) );
}





MpegAtr::~MpegAtr()
{

    delete codec;
    codec = NULL;

    delete [] puffer;
}




int MpegAtr::Seek( long block )
{
    int stat = MpegTrack::Seek( block );
    if( stat < 1 )
        return stat;

    uint8_t *buf;
    long n;
    while( Position() < block ) {
        if( ! readData( buf, n ) ) {
            return 0;
        }
    }

    return 1;
}





// Rohdaten lesen und schreiben
int MpegAtr::readRaw( unsigned char *&data, long &bytes )
{
    long ende;
    
    bytes = 0;

    if( stream->Tracktype() == TRACK_CODE_MP3 ) {
        ende = get_block_mp3();
    } else if( stream->Tracktype() == TRACK_CODE_AC3 ) {
        ende = get_block_ac3();
    } else {
        return 0;
    }

    if( ende < 0 )
        return 0;
        
    bytes = ende;


    // Position in Keytable einreihen
    if( (Position() % 30) == 0 ) {
        // Fge I-Frame in Key-Table ein
        key_table.AddKeyFrame( Position(), stream->FileOffset() );
    }

    // merke die maximale Position, fr die schon Key-Frame informationen existieren
    max_pos = (Position() > max_pos)? Position() : max_pos;

    
    // Daten in Puffer kopieren
    if( bytes > puffer_gr ) {
        delete [] puffer;
        puffer_gr = bytes;
        puffer = new uint8_t[puffer_gr];
    }
    stream->GetData( 0, puffer, bytes );
    data = puffer;
    stream->ClearData( bytes );
            
    return 1;
}






int MpegAtr::writeRaw( unsigned char *data, long bytes, int samp )
{
    uint8_t *buf = NULL;
    
    if( IOMode() != VideoFile::write )
        return 0;

    double timestamp = Position() / Rate();

    stream->AddPack( buf, bytes, timestamp );
    memcpy( buf, data, bytes );

    // Streams multiplexen
    file->multiplex_puffer( this );

    return 1;
}












/******************************************************************************/
/***** MP3 Funktionen                                                     *****/
/******************************************************************************/

/***** Anfang des Streams suchen **********************************************/
bool MpegAtr::sync_block_start_mp3( void )
{
    uint32_t head = 0;
    int i = 0;
    int offset = 10;
    
    while( i < offset ) {
        if( stream->Bytes() < 4 ) {
            if( file->FillStreamPuffer(this) < 1 )
                return false;

            if( stream->Bytes() < 4 )
                return false;
        }

        head  = (uint32_t)stream->Data(0) << 24;
        head |= (uint32_t)stream->Data(1) << 16;
        head |= (uint32_t)stream->Data(2) << 8;
        head |= (uint32_t)stream->Data(3);

        // DEBUG_TXT( "Head = %8x", head );

        // Header auswerten
        if( decode_header_mp3( head ) ) {
            mp3_block_start = head;
            DEBUG_TXT( _("MP%i stream found: %i channels, %i Hz"), layer, Channels(), SampleRate() );
            break;
        }

        // kein gltiger Header gefunden -> ein Zeichen weitergehen
        stream->ClearData( 1 );

        i++;
    }

    if( i >= offset )
        return false;

    return true;
}






/***** Header dekodieren ******************************************************/
#define MPA_STEREO  0
#define MPA_JSTEREO 1
#define MPA_DUAL    2
#define MPA_MONO    3

int MpegAtr::decode_header_mp3( uint32_t head )
{
    const uint16_t mpa_freq_tab[3] = { 44100, 48000, 32000 };
    
    int lsf;
    int mpeg25;
    int sample_rate_index;
    int mode;


    if( (head & 0xffe00000) != 0xffe00000 )
        return 0;

    if( head & (1<<20) ) {
        lsf = (head & (1<<19)) ? 0x0 : 0x1;
        mpeg25 = 0;
    } else {
        lsf = 1;
        mpeg25 = 1;
    }

    // Sample Rate
    if( ((head>>10)&0x3) == 0x3) {
        return 0;
    }
    sample_rate_index = (head >> 10) & 3;
    SetSampleRate( mpa_freq_tab[sample_rate_index] >> (lsf + mpeg25) );
    
    //bitrate_index = (header >> 12) & 0xf;
    //padding = (header >> 9) & 1;
    //extension = (header >> 8) & 1;
    mode = (head >> 6) & 3;
    //mode_ext = (header >> 4) & 3;
    //copyright = (header >> 3) & 1;
    //original = (header >> 2) & 1;
    //emphasis = header & 3;

    (mode == MPA_MONO)? SetChannels( 1 ): SetChannels( 2 );
   
    layer = 4-((head>>17)&3);

    framesize = (lsf == 0) ? 1152 : 576;
    SetFrameSamples( framesize );
    SetRate( (double)SampleRate() / (double)framesize );
    
    return 1;
}





int MpegAtr::get_block_mp3( void )
{
    int begin, ende;
    uint32_t head=0;
    
    // Suche Start-Header - prfe, ob Header mit altem Header bereinstimmt
    begin = stream->Finden( 0, mp3_block_start & 0xfffffd00, 0xfffffd00, head );
    // begin = puffer_finden( 0, 0xffe00000, 0xffe00000, head );

    if( begin < 0 ) {
        // Puffer wieder auffllen
        if( file->FillStreamPuffer(this) <= 0 )
            return -1;

        // nochmal suchen
        begin = stream->Finden( 0, mp3_block_start & 0xfffffd00, 0xfffffd00, head );

        if( begin < 0 )
            return -1;
    }


    // Header gefunden!
    if( begin > 0 ) {
        DEBUG_TXT( _("Resync stream %i bytes"), begin );
        stream->ClearData( begin );

        // Puffer wieder auffllen
        file->FillStreamPuffer(this);
    }


    // ende = puffer_finden( 4, 0xffe00000, 0xffe00000, head );
    ende = stream->Finden( 4, mp3_block_start & 0xfffffd00, 0xfffffd00, head );
    if( ende < 0 ) {
        // Puffer wieder auffllen
        if( file->FillStreamPuffer(this) < 0 )
            return -1;

        // nochmal suchen
        ende = stream->Finden( 4, mp3_block_start & 0xfffffd00, 0xfffffd00, head );
    }

    if( ende < 0 ) {
        //letzter Block - bis Zum Ende des Puffers
        ende = stream->Bytes();
    }


    return ende;
}












/******************************************************************************/
/***** AC3 Funktionen                                                     *****/
/******************************************************************************/

/***** Anfang des Streams suchen **********************************************/
bool MpegAtr::sync_block_start_ac3( void )
{
    uint8_t head[7];
    int ende;
    double rate = Rate();
    int channels = Channels(), samplerate = SampleRate(), bit_rate = 0;

    ende = get_block_ac3();

    if( ende < 0 )
        return false;

    // Header auswerten
    stream->GetData( 0, head, 7 );
    if( decode_header_ac3( head, rate, samplerate, channels, bit_rate ) > 0 ) {
        SetRate( rate );
        SetSampleRate( samplerate );
        SetChannels( channels );
        SetFrameSamples( (int)(samplerate / rate) );
        SetBitsPerSecond( bit_rate );
        DEBUG_TXT( _("AC3 stream found: %i channels, %i Hz"), Channels(), SampleRate() );
        return true;
    }

    return false;
}






/***** Header dekodieren ******************************************************/
/* AC3 - Header parsen
   Code aus a52dec
   http://liba52.sourceforge.net/
*/
#define A52_CHANNEL 0
#define A52_MONO 1
#define A52_STEREO 2
#define A52_3F 3
#define A52_2F1R 4
#define A52_3F1R 5
#define A52_2F2R 6
#define A52_3F2R 7
#define A52_CHANNEL1 8
#define A52_CHANNEL2 9
#define A52_DOLBY 10
#define A52_CHANNEL_MASK 15
#define A52_LFE 16

int MpegAtr::decode_header_ac3( uint8_t *buf, double &blockrate, int &samplerate, int &ch, int &bit_rate )
{
    static int rate[] = { 32,  40,  48,  56,  64,  80,  96, 112,
                         128, 160, 192, 224, 256, 320, 384, 448,
                         512, 576, 640};
    static uint8_t halfrate[12] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 3};    
    static uint8_t lfeon[8] = {0x10, 0x10, 0x04, 0x04, 0x04, 0x01, 0x04, 0x01};
    static const int ac3_channels[8] = { 2, 1, 2, 3, 3, 4, 4, 5 };
    
    int frmsizecod;
    int bitrate;
    int half;
    int data_size = 0;
    int acmod;
    int flags;

    if ((buf[0] != 0x0b) || (buf[1] != 0x77))   /* syncword */
        return 0;

    if (buf[5] >= 0x60)         /* bsid >= 12 */
        return 0;

    half = halfrate[buf[5] >> 3];

    /* acmod, dsurmod and lfeon */
    acmod = buf[6] >> 5;
    flags = ((((buf[6] & 0xf8) == 0x50) ? A52_DOLBY : acmod) |
             ((buf[6] & lfeon[acmod]) ? A52_LFE : 0));
    
    frmsizecod = buf[4] & 63;
    if (frmsizecod >= 38)
        return 0;
    bitrate = rate [frmsizecod >> 1];
    bit_rate = (bitrate * 1000) >> half;

    
    switch (buf[4] & 0xc0) {
    case 0:     /* 48 KHz */
        samplerate = 48000 >> half;
        data_size = 4 * bitrate;
        break;
    case 0x40:
        samplerate = 44100 >> half;
        data_size = 2 * (320 * bitrate / 147 + (frmsizecod & 1));
        break;
    case 0x80:
        samplerate = 32000 >> half;
        data_size = 6 * bitrate;
        break;
    default:
        return 0;
    }

    blockrate = (double)bit_rate / (8.0 * data_size);

    ch = ac3_channels[ (flags & 7) ];
    if( flags & A52_LFE )
        ch++;
    
    return data_size;
}







int MpegAtr::get_block_ac3( void )
{
    uint8_t head[7];
    int i = 0, b=0, c=0, d=0, ende=-1;
    int offset = 10000;
    double a=0;

    while( i < offset ) {
        if( stream->Bytes() < 7 ) {
            if( file->FillStreamPuffer(this) < 1 )
                return -1;

            if( stream->Bytes() < 7 )
                return -1;
        }

        stream->GetData( 0, head, 7 );

        // Header auswerten
        if( (ende = decode_header_ac3( head, a, b, c, d )) > 0 ) {
            break;
        }

        // kein gltiger Header gefunden -> ein Zeichen weitergehen
        stream->ClearData( 1 );

        i++;
    }

    if( i >= offset )
        return -1;

    if( i > 0 ){
        DEBUG_TXT( _("Resync stream %i bytes"), i );
    }

    // sicherstellen, da genug Daten im Puffer sind
    if( (int)stream->Bytes() < ende ) {
        if( file->FillStreamPuffer(this) < 1 )
            return -1;

        if( (int)stream->Bytes() < ende )
            return -1;
    }
    
    return ende;
}
