/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>

#include "Mpeg_Vtr.h"
#include "debug.h"
#include "VideoFile/codecplugins.h"
#include "VideoFile/translate.h"
#include "MpegCodes.h"


#define MAX_OFFSET 100

// Puffer festlegen - maximale Gre: 2 umkompremierte Bilder mit 800*600 Punkten RGB24
#define PUFFER_GR 2 * 800*600*3

/* picture coding type */
#define I_TYPE 1
#define P_TYPE 2
#define B_TYPE 3
#define D_TYPE 4




const double MpegVtr::mpeg_framerates[] = {
    00.000, 23.976,
    24.000,                      /* 24fps movie */
    25.000,                      /* PAL */
    29.970,                      /* NTSC */
    30.000, 50.000,
    59.940, 60.000,
    -1
};




/*****************************************************************************/
/***** Videotrack-Funktionen *************************************************/
/*****************************************************************************/

MpegVtr::MpegVtr( MpegIO *File, double Rate, int x, int y, const char *Codec_name, int track )
:Track( File, track, Codec_name ), MpegTrack( File, Rate, Codec_name, track ), V_Track( File, Rate, x, y, Codec_name )
{
    data_puffer_gr = 0;
    data_puffer = NULL;
    max_pos = 0;
    sequence_header = NULL;
    sequence_header_gr = 0;
    sequence_header_gelesen = false;
    chunk = NULL;

    // Variablen fr Timestamp-Berechnung
    group_start_pic=0; group_start_field=0; fields_presented=0; decoding_order=0;

    // Stream-Puffer festlegen
    stream = file->finde_stream_puffer( TRACK_CODE_MPEG, track );
            
    if( CodecName() == "" ) {
        if( File->mpeg_system_type == 1 )
            SetCodecName( "Mpeg-1" );
        else if( File->mpeg_system_type == 2 )
            SetCodecName( "Mpeg-2" );
    }
    
    codec = AV_Codecs->CreateVCodec( this, "MpegVideo", CodecName() );

    if( codec == NULL ) {
        ERROR_TXT( _("Video codec \"%s\" not implemented!"), CodecName().chars() );
        codec = new V_Codec( this );
    }
}





MpegVtr::MpegVtr( MpegIO *File, int track )
: Track(File, track), MpegTrack( File, File->vstream[track], track ), V_Track( File, track )
{
    data_puffer_gr = PUFFER_GR;
    data_puffer = new uint8_t[data_puffer_gr];
    max_pos = 0;
    sequence_header = NULL;
    sequence_header_gr = 0;
    sequence_header_gelesen = false;

    chunk = new Chunk( file, this );
    
    SetCodecName( "Mpeg-1/2" );

    // Sequence Header aus Track auslesen
    assert( sync_header() > 0 );
    decode_header( sequence_header, sequence_header_gr );
    SetFrames( (int64_t)(system->TotalTime() * Rate()) );
    
    codec = AV_Codecs->CreateVCodec( this, "MpegVideo", CodecName() );

    if( codec == NULL ) {
        ERROR_TXT( _("Video codec \"%s\" not implemented!"), CodecName().chars() );
        codec = new V_Codec( this );
    }

    // vom Codec gendertes Format bernehmen
    (V_Format)*this = (V_Format)*codec;
}





MpegVtr::~MpegVtr()
{

    delete codec;
    codec = NULL;

    delete chunk;
}




int MpegVtr::Seek( long block )
{
    int stat = MpegTrack::Seek( block );
    if( stat < 1 )
        return stat;
    
    // Auf Anfang von nchstem Bild synchronisieren
    int code;
    while( (code = chunk->Startcode()) != PICTURE_START_CODE && code >= 0 ) {
        if( chunk->Skip() < 1)
            return 0;
    }
    if( code < 0 )
        return 0;
            
    while( Position() < block ) {
        if( !skip_frame() ) {
            ERROR_TXT( _("Can not seek stream!") );
            return 0;
        }
    }
    
    //sequence_header_gelesen = header_gelesen;
    sequence_header_gelesen = false;

    return 1;
}






/**liefert den letzten Keyframe zur angegebenen Position
  */
long MpegVtr::GetKeyFrame( long bild )
{
    long long offset = 0;
    long key = 0;

    if( max_pos >= bild ) {
        key = key_table.GetLastKeyFrame( bild, offset );
        // Kein Keyframe fr diese Position vorhanden
        if( key < 0 )
            key = bild;

        //DEBUG_TXT( "Keyframe %i fr Bild %i", key, bild );
        return key;
    }

    return bild;
}






// Rohdaten lesen und schreiben
int MpegVtr::readRaw( unsigned char *&data, long &bytes )
{
    size_t offset = 0;
    size_t n = 0;
    int code = 0;

    if( IOMode() != VideoFile::read )
        return 0;
        
    // Ist schonmal ein Sequenceheader gelesen worden?
    if( !sequence_header_gelesen ) {
        DEBUG_TXT( _("Sequence header not read.") );
        // enthlt der aktuelle Block schon einen Srequence Header?
        if( chunk->Startcode() != SEQUENCE_HEADER_CODE ) {
            DEBUG_TXT( _("Copy sequence header.") );
            
            memcpy( data_puffer, sequence_header, sequence_header_gr );
            offset += sequence_header_gr;
        }
        sequence_header_gelesen = true;
    }

    // kopiere alles bis zum ersten Bild
    while( (code = chunk->Startcode()) != PICTURE_START_CODE && code >= 0 ) {
        n = data_puffer_gr - offset;
        if( chunk->Lesen( data_puffer + offset, n ) < 1 )
            return 0;
        offset += n;
    }

    if( code < 0 )
        return 0;

    n = data_puffer_gr - offset;
    if( chunk->Lesen( data_puffer + offset, n ) < 1 )
        return 0;
    offset += n;

    
    // kopiere alles bis zum Ende des Bildes
    while( (code = chunk->Startcode()) != PICTURE_START_CODE && code >= 0 ) {
        n = data_puffer_gr - offset;
        if( chunk->Lesen( data_puffer + offset, n ) < 1 ) {
            bytes = offset;
            data = data_puffer;
            return 1;
        }
        offset += n;
    }

    bytes = offset;
    data = data_puffer;
    
    if( code < 0 )
        return 1;

                                
    // prfe, ob nchstes Bild ein I-Frame ist
    if( test_keyframe() ) {
        // Fge I-Frame in Key-Table ein
        key_table.AddKeyFrame( Position()+1, stream->FileOffset() );
    }

    // merke die maximale Position, fr die schon Key-Frame informationen existieren
    max_pos = (Position()+1 > max_pos)? Position()+1 : max_pos;

    return 1;
}






int MpegVtr::writeRaw( unsigned char *data, long bytes, int keyframe )
{
    uint8_t *buf = NULL;
    double pts=0, dts=0;
    
    if( IOMode() != VideoFile::write )
        return 0;

    //pts = Position() / Rate();

    calc_timestamp( data, bytes, pts, dts );

    // Daten zum Stream hinzufgen
    stream->AddPack( buf, bytes, pts );
    memcpy( buf, data, bytes );

    // Streams multiplexen
    file->multiplex_puffer( this );
    
    return 1;
}







/***** Prfe, ob Frame ein I-Frame ist ***************************************/
bool MpegVtr::test_keyframe(void)
{
    if( chunk->Startcode() != PICTURE_START_CODE )
        return false;

    uint8_t data = chunk->Data(5);       
    int coding_type = (data >> 3) & 7;

    return (coding_type == I_TYPE);
}





/***** Ein Bild berspringen *************************************************/
int MpegVtr::skip_frame( void )
{
    int code = 0;

    // einen Block einlesen
    
    // berspringe alles bis zum ersten Bild
    while( (code = chunk->Startcode()) != PICTURE_START_CODE && code >= 0 ) {
        if( chunk->Skip() < 1)
            return 0;
    }

    if( code < 0 )
        return 0;

    if( chunk->Skip() < 1 )
        return 0;

    // berspringe alles bis zum Ende des Bildes
    while( (code = chunk->Startcode()) != PICTURE_START_CODE && code >= 0 ) {
        if( chunk->Skip() < 1 )
            return 1;
    }

    if( code < 0 )
        return 1;

    // prfe, ob nchstes Bild ein I-Frame ist
    if( test_keyframe() ) {
        // Fge I-Frame in Key-Table ein
        key_table.AddKeyFrame( Position()+1, stream->FileOffset() );
    }

    // merke die maximale Position, fr die schon Key-Frame informationen existieren
    max_pos = (Position()+1 > max_pos)? Position()+1 : max_pos;

    SetPosition( Position() +1 );

    return 1;
}









/***** finde einen Sequence-Header und kopiere ihn ****************************/
int MpegVtr::sync_header( void )
{
    int i=0, code=0;
    
    // berspringe alles bis zum Sequence-Header
    while( (code = chunk->Startcode()) != SEQUENCE_HEADER_CODE && code >= 0 ) {
        if( chunk->Skip() < 1)
            return 0;

        // breche nach zu vielen Versuchen ab
        if( i++ > 100 )
            return 0;
    }

    if( code < 0 )
        return 0;

    // Sequence Header gefunden -> kopieren ( Sequence-Header und
    // alles bis zum ersten Bild )
    size_t n=0;
    sequence_header_gr=0;
    while( (code = chunk->Startcode()) != PICTURE_START_CODE && code >= 0 ) {
        n = data_puffer_gr - sequence_header_gr;
        if( n==0 )
            return 0;
        if( chunk->Lesen( data_puffer + sequence_header_gr, n ) < 1 )
            return 0;
        sequence_header_gr += n;
    }
    // sequence_header_gr = chunk->Bytes();
    delete [] sequence_header;
    sequence_header = new uint8_t[sequence_header_gr];
    memcpy( sequence_header, data_puffer, sequence_header_gr );
    
    return 1;
}

    



/***** Header dekodieren ******************************************************/
int MpegVtr::decode_header( unsigned char *head, int bytes )
{
    int frame_rate_code = 0;

    assert( bytes > 10 );

    // prfe, ob der Chunk ein Sequence-Header ist
    //if( memcmp( head, SEQUENCE_HEADER, 4 ) != 0 )
    //    return 0;

    // Prfe Gltigkeit des Headers
    if ((head[10] & 0x20) != 0x20)
        return 0;	/* missing marker_bit */

    // Lese Informationen von Videotrack
    // Bildbreite / -hhe
    hoehe = (head[4] << 16) | (head[5] << 8) | head[6];

    breite = ((hoehe >> 12) + 15) & ~15;
    hoehe = ((hoehe & 0xfff) + 15) & ~15;

    if ((breite > 768) || (hoehe > 576))
        return 0;	/* size restrictions for MP@ML or MPEG1 */

    /* this is not used by the decoder */
    aspect_ratio_code = head[7] >> 4;
    frame_rate_code = head[7] & 15;
    bitrate = (head[8]<<10)|(head[9]<<2)|(head[10]>>6);

    SetRate( decode_framerate( frame_rate_code ) );

    return 1;
}







/***** Framerate-Code in Framerate umwandeln *********************************/
double MpegVtr::decode_framerate( int code )
{
    for( int i=0; i<=code; i++ ) {
        if( mpeg_framerates[i] < 0 )
            return 0.0;
    }

    return mpeg_framerates[code];
}





/***** Timestamp des Frames berechnen ****************************************/
void MpegVtr::calc_timestamp( uint8_t *buf, size_t bytes, double &pts, double &dts )
{
    double frame_rate = Rate();
    int temporal_reference=0;
    int coding_type=0;
    uint32_t data;

    int pulldown_32=0, repeat_first_field=0;

    // Daten nach Bild-Startcode durchsuchen
    assert( bytes > 4 );
    data  = *(buf++) << 16;
    data |= *(buf++) << 8;
    data |= *(buf++);
    bytes -= 3;
        
    do {
        data = (data << 8) | *(buf++);
        bytes--;
    } while( data != (CHUNK_START_CODE | PICTURE_START_CODE) && bytes > 0 );

    if( bytes <= 0 ) {
        return;
    }

    
    // Bild gefunden - lese temporal_reference und Frame-Typ
    temporal_reference  = *(buf++) << 2;
    temporal_reference |= (*buf >> 6) & 3;
    coding_type = (*buf >> 3) & 7;
    //DEBUG_TXT( "Temporal Reference = %i", temporal_reference );
    
    // MPEG2: suche nach EXT_START_CODE


    // I-Frame
    //DEBUG_TXT( "Frame-Type = %i", coding_type );
    if( coding_type == I_TYPE ) {
        group_start_pic = decoding_order;
        group_start_field = fields_presented;
    }
    

    // Berechne Timestamps
    if( pulldown_32 )
    {
        int frames2field;
        int frames3field;
                
        dts = (fields_presented / 2) / frame_rate;
                
        if( repeat_first_field )
        {
            frames2field = (temporal_reference+1) / 2;
            frames3field = temporal_reference / 2;
            fields_presented += 3;
        }
        else
        {
            frames2field = (temporal_reference) / 2;
            frames3field = (temporal_reference+1) / 2;
            fields_presented += 2;
        }
                
        pts = (frames2field*2 + frames3field*3 + group_start_field+1) / 2;
                      
    } else {
        dts = decoding_order / frame_rate;
        pts = (temporal_reference + group_start_pic + 1) / frame_rate;
        fields_presented += 2;
    }

    //DEBUG_TXT( "pts = %.4f     dts = %.4f", pts, dts );
    decoding_order++;
}
