/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/



#include "System.h"
#include "debug.h"
#include "VideoFile/translate.h"
#include "MpegCodes.h"

#define END_OFFSET 50*1024
#define SCAN_PACKS 50



System::System( System *sys, FilePuffer *Puffer, off_t offset )
{
    puffer = Puffer;
    timestamp = -1.0;
    last_timestamp = -1.0;
    start_timestamp = sys->start_timestamp;
    end_time = sys->end_time;
    track_code = 0;
    file_offset = 0;
    start_offset = (offset == 0 ? sys->start_offset : offset);
    file_size = sys->file_size;
    have_timestamps = sys->have_timestamps;
    data_size = 0;
    mpeg_type = sys->mpeg_type;
    init = sys->init;
}






System::System( FilePuffer *Puffer, off_t offset )
{
    puffer = Puffer;
    timestamp = -1.0;
    last_timestamp = -1.0;
    start_timestamp = 0.0;
    end_time = -1.0;
    track_code = 0;
    file_offset = 0;
    start_offset = offset;
    file_size = -1;
    have_timestamps = false;
    data_size = 0;
    mpeg_type = 1;
    init = false;
}





System::~System()
{
}




/**Nchsten Pack-Header lesen und parsen
  */
int System::NextPack( void )
{
    int stat=1;

    // initialisieren
    if( !init ) {
        if( init_system() < 0 )
            return -1;
    }
        
    // Daten des letzten Pakets aus Puffer entfernen
    puffer->ClearData( data_size );
    data_size = 0;
        
    // Nchsten PES-Header suchen
    int i = 0;
    while( 1 ) {
        // Sind noch Daten da?
        if( (stat = puffer->NeedData(12)) < 1 )
            return stat;
        
        //Abbruchbedingung
        if( i++ > 1000 ) {
            ERROR_TXT( _("Mpeg-system stream contains no data!") );
            return -1;
        }
        
        // Auf nchsten Header synchronisieren
        if( (stat = Sync()) <= 0 )
            return stat;
            
        // Pack-Header
        if( Data(3) == PACK_HEADER_CODE ) {
            ParsePackHeader();
            continue;            
        }

        // Diese werden nicht weiter ausgewertet
        if( Data(3) == SYSTEM_HEADER_CODE || Data(3) == PRIVATE_HEADER_CODE || Data(3) == PADDING_HEADER_CODE ) {
            // berspringen
            ParseExtHeader();
            continue;
        }

        // PES-Header
        if( Data(3) == PES_EXT_CODE || ( (Data(3) & 0xF0) == TRACK_CODE_MP3 || (Data(3) & 0XF0) == TRACK_CODE_MPEG ) ) {
            break;
        }

        // berspringen
        puffer->ClearData( 4 );
    }

    // Offset dieses Datenblockes speichern
    file_offset = puffer->Offset();
    
    stat = ParsePesHeader();

    if( !have_timestamps && timestamp >= 0 ) {
        have_timestamps = true;
        start_timestamp = timestamp;
    }
    
    if( timestamp >= 0 ) {
        timestamp -= start_timestamp;
        last_timestamp = timestamp;
    }
    
    return stat;        
}




/**Den nchsten Block aus dem Puffer in das Datenfeld kopieren<\br>
  *Rckgabe: >0  erfolg<br>
  *           0  keine weiteren Daten<br>
  *          -1  Fehler
  */
int System::GetPackData( uint8_t *data, size_t &bytes )
{
    long ende = Bytes();

    if( ende <= 0 ) {
        bytes = 0;
        return ende;
    }

    // Daten kopieren
    if( bytes >= (size_t)ende )
        bytes = ende;

    puffer->GetData( 0, data, bytes );

    return 1;
}






/**Bytes im nchsten Paket
  */
long System::Bytes( void )
{
    return data_size;
}





/**Auf den Start eines Pakets synchronisieren<\br>
  *Rckgabe: >0 Erfolg<\br>
  *           0 keine weiteren Daten<\br>
  *          -1 Fehler
  */
int System::Sync( void )
{
    // suche nach Start des ersten Chunks im Puffer
    long begin = puffer->Finden( 0, CHUNK_START, 3 );

    if( begin < 0 ) {
        if( puffer->Eof() )
            return 0;
        return -1;
    }

    if( begin > 0 ) {
        //DEBUG_TXT( "Resync Stream %i Bytes", begin );
        puffer->ClearData( begin );
    }
    
    return 1;
}





/**Pack Header lesen und parsen<br>
 *Rckgabe: >0 Erfolg<br>
 *          -1 Fehler
 */
int System::ParsePackHeader( void )
{
    int header_size = 0;
    int stat;
        
    if( Data( 3 ) != PACK_HEADER_CODE ) {
        ERROR_TXT( _("No Mpeg1/2 pack-header, offset: 0x%lx"), puffer->Offset() );
        return -1;
    }

    if( (stat = puffer->NeedData(5)) < 1 )
        return stat;
        
    // Pack-Header decodieren
    if( (Data(4) & 0xc0) == 0x40 ) {
        // MPEG 2
        if( (stat = puffer->NeedData(14)) < 1 )
            return stat;

        mpeg_type = 2;
        header_size = 14 + (Data(13) & 7);
        
    } else if( (Data(4) & 0xf0) == 0x20 ) {
        // MPEG 1
        mpeg_type = 1;
        header_size = 12;
        
    } else {
        ERROR_TXT( _("No Mpeg1/2 pack-header, offset: 0x%lx"), puffer->Offset() );
        puffer->ClearData( 12 ); 
        return -1;
    }
    
    puffer->ClearData( header_size );
    return 1;
}





/**Sonstige Header lesen und parsen<br>
 *Rckgabe: >0 Erfolg<br>
 *          -1 Fehler
 */
int System::ParseExtHeader( void )
{
    int size;
    int stat;

    if( (stat = puffer->NeedData(6)) < 1 )
        return stat;

    // Gre der Daten auslesen
    size = 6 + (Data(4) << 8) + Data(5);

    puffer->ClearData( size );
    return 1;
}





/**PES Header lesen und parsen<br>
 *Rckgabe: >0 Erfolg<br>
 *          -1 Fehler
 */
int System::ParsePesHeader( void )
{
    double pts=-1.0, dts=-1.0;
    int header_size = 0;
    int pts_flags = 0;
    //uint16_t start_pos=0;
    int stat;
    
    css = false;

    if( (stat = puffer->NeedData( 7 )) < 1 )
        return stat;
        
    // PES-Header decodieren
    track_code = Data( 3 );
    if( track_code != 0xbd && (track_code < 0xc0 || track_code > 0xef) ) {
        ERROR_TXT( _("Found ilegal track code: no Mpeg1/2 system stream!") );
        return -1;
    }

    // Gre der Daten auslesen
    data_size = 6 + (Data(4) << 8) + Data(5);

    if( (stat = puffer->NeedData( data_size )) < 1 )
        return stat;

    if( (Data(6) & 0xc0) == 0x80 ) {
        // MPEG 2
        mpeg_type = 2;

        if( data_size < 9 )
            return -1;

        if( (Data(6)>>4)&3 )
            css=true;

        pts_flags = Data(7) >> 6;
        header_size = 9 + Data(8);
        
        if( header_size > data_size ) {
            ERROR_TXT( _("Error in data pack!") );
            return -1;
        }

        // Tiemstamp einlesen        
        if( pts_flags == 2 && header_size >= 13 ){
            pts = DecodeTimestamp( 9 );
        } else if ( pts_flags == 3 && header_size >= 18 ){
            pts = DecodeTimestamp( 9 );
            dts = DecodeTimestamp( 14 );
        }

        // erweiterten Tracktyp einlesen
        if( track_code == PES_EXT_CODE ) {
            track_code = Data( header_size );
            // 0x20..0x3F  subtitle
            // 0x80..0x9F  AC3 audio
            // 0xA0..0xBF  PCM audio

            header_size++;
            if( (track_code & 0xE0) == TRACK_CODE_AC3 || (track_code & 0xE0) == TRACK_CODE_PCM ) {
                // num_frames = Data( header_size );
                //start_pos = Data( header_size + 1 );
                //start_pos = (start_pos << 8) | Data( header_size + 2);             
                header_size += 3;
            }
        }
        
    } else {
        // MPEG 1
        mpeg_type = 1;

        if( data_size < 7 )
            return -1;

        header_size = 7;
        while( Data(header_size-1) == 0xff ) {
            header_size++;
            if( header_size == 23 ) {
                ERROR_TXT( _("Unable to decode Mpeg1 pack header!") );
                break;
            }
        }

        // STD Scale und Size berspringen
        if( (Data(header_size-1) & 0xc0) == 0x40 ) {
            header_size += 2;
        }

        // Mpeg1 Timestamp lesen:
        if( (Data(header_size-1) >> 4) == 2 ) {
            pts = DecodeTimestamp( header_size-1 );
            header_size += 4;
            
        } else if( (Data(header_size-1) >> 4) == 3 ) {
            pts = DecodeTimestamp( header_size-1 );

            if( (Data(header_size+4) >> 4) != 1 )
                pts=0;
                
            dts = DecodeTimestamp( header_size+4 );
            
            header_size += 9;
        }
    }

    // PES-Header berspringen
    puffer->ClearData( header_size );

    timestamp = pts;
    data_size -= header_size;
    if( data_size < 0 )
        return -1;

    if( (stat = puffer->NeedData(data_size)) < 1 )
        return stat;

    //DEBUG_TXT_3( "Trackcode: 0x%02x   Datengre: %li   Headergre: %li   Timestamp: %f   Offset: 0x%lx", track_code, data_size, header_size, timestamp, puffer->Offset() );
    //DEBUG_TXT_3( "Timestamp: %f", timestamp );
    
    return 1;
}






/**Timestamp lesen<br>
 *Rckgabe: >0 Timestamp in Sec.<br>
 *          <0 Fehler
 */
double System::DecodeTimestamp( size_t offset )
{
    uint8_t c;
    uint16_t d,e;
    uint32_t pts;

    c = Data(offset);
    d = ( Data(offset+1) << 8 ) + Data(offset+2);
    e = ( Data(offset+3) << 8 ) + Data(offset+4);

    if( (c&1)!=1 || (d&1)!=1 || (e&1)!=1 ) {
        // ungltiger Timestamp
        return -1;
    }
    
    pts = (((c>>1)&7) << 30) | ((d>>1)<<15) | (e>>1);
    
    return pts / 90000.0;
}




    
/**Zeichen an Offset in Chunk
  */
uint8_t System::Data( size_t offset )
{
    return puffer->Data( offset );
}




/**System Stream positionieren<br>
 *Rckgabe: >0  erfolg<br>
 *           0  eof<br>
 *          -1  Fehler
 */
int System::SeekTime( double sec )
{
    int stat = 1;
    double mult = 0.99;
    
    if( !have_timestamps || sec < 0 )
        return -1;

    DEBUG_TXT( "SeekTime %f", sec );
    
    // Datei positionieren
    do {
        if( end_time > 0 && mult > 0.76 && puffer->CanSeek() ) {
            off_t offset = (off_t)((double)(file_size - start_offset) * sec / end_time * mult);
            DEBUG_TXT( "Positioniere auf Offset %li", offset );
            SeekOffset( offset + start_offset );
        } else if( sec < last_timestamp && puffer->CanSeek() ) {
            // keine Endzeit vorhanden
            SeekOffset( start_offset );
        }

        // nchsten timestamp einlesen
        while( timestamp < 0 ) {
            if( (stat = NextPack()) <= 0 )
                return stat;
        }

        if( sec < timestamp )
            mult -= 0.05;
    } while( sec < timestamp && mult > 0.66 && puffer->CanSeek() );

    if( sec < timestamp )
        return -1;        
        
    if( sec > last_timestamp ) {
        while( sec > timestamp ) {
            //berspringe solange Blcke bis timestamp >= sec
            if( (stat = NextPack()) <= 0 )
                return stat;
        }
        return 1;
    }

    return 1;
}






/**System Stream positionieren<br>
  *Rckgabe: >0  erfolg<br>
  *           0  eof<br>
  *          -1  Fehler
  */
int System::SeekOffset( off_t offset )
{
    int stat = 1;

    if( !puffer->CanSeek() || offset < 0 )
        return -1;

    if( offset < start_offset )
        offset = start_offset;
        
    if( (stat = puffer->Seek(offset)) <= 0 )
        return stat;

    data_size = 0;
    if( (stat = NextPack()) <= 0 )
        return stat;
        
    return 1;
}





/**Endzeit und Dateigre einlesen
  */
void System::get_end_time( void )
{
    int stat;
    off_t offset;
    
    if( !puffer->CanSeek() || end_time > 0.0 ) {
        return;
    }

    // Dateigre
    file_size = puffer->FileSize();

    if( !have_timestamps ) {
        return;
    }

    // kurz vor das Ende der Datei springen
    offset = FileOffset();
    if( file_size > END_OFFSET ){
        stat = SeekOffset( file_size - END_OFFSET );
        if( stat <= 0 )
            return;
    }
    //DEBUG_TXT_3( "Dateigre: %li", file_size );
    
    // suche letztem Timestamp
    while( NextPack() > 0 ) {
        if( timestamp > 0.0 && timestamp > end_time ) {
            end_time = timestamp;
        }
    }

    // Datei neu positionieren
    SeekOffset( offset );
    
    INFO_TXT( _("Total time: %f    file size: %li"), end_time, file_size );
}






/**Datei initialisieren<br>
  *such nach Timestamps und ermittle Dateigre und End-Zeit
  */
int System::init_system( void )
{
    int stat;

    if( init )
        return 1;
    init = true;

    if( puffer->CanSeek() ) {
        // gehe zum Startoffset
        if( (stat = puffer->Seek(start_offset)) <= 0 )
            return stat;
    
        // nach Timestamps suchen
        for( int i=0; i<SCAN_PACKS; i++ ) {
            if( (stat = NextPack()) <= 0 )
                return stat;
            if( have_timestamps )
                break;
        }

        // Dateigre und Endzeit
        get_end_time();

        // gehe zum Startoffset
        if( (stat = puffer->Seek(start_offset)) <= 0 )
            return stat;
    }
    
    return 1;
}
