/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/



#include <string.h>
#include "System_Write.h"
#include "debug.h"
#include "VideoFile/translate.h"
#include "MpegCodes.h"

#define MAX_PACK_SIZE 2048
#define MAX_HEADER_SIZE 36



System_Write::System_Write( int fileno, int MpegType )
{
    stream = fileno;
    for( int i=0; i<256; i++ ) {
        last_timestamp[i] = -1;
    }
    mpeg_type = MpegType;
    pack_data = new uint8_t[MAX_PACK_SIZE];
    header_data = new uint8_t[MAX_HEADER_SIZE];
    bytes_written = 0;
    v_tracks = 0;
    a_tracks = 0;
    sys_header = false;
    VBR = true;
}




System_Write::~System_Write()
{
    delete [] pack_data;
    delete [] header_data;
}




/**Nchstes Paktet erstellen und schreiben
  */
int System_Write::WritePack( StreamPuffer *Puffer )
{
    int stat = 1;
    //uint data_size = MAX_DATA_SIZE;
    unsigned int data_size = MAX_PACK_SIZE;
    unsigned int bytes=0;
    bool ts_ok = false;

    int track_code = Puffer->Tracktype() | Puffer->Tracknumber();

    // Timestamp des ersten Packets im Puffer
    double timestamp = Puffer->Timestamp(0);

    // Timestamp wird nur geschrieben, wenn in diesem Paket ein neuer
    // Frame anfngt
    if( last_timestamp[track_code] == timestamp ) {
        timestamp = -1;
    } else {
        last_timestamp[track_code] = timestamp;
    }

    do {
        bytes = 0;
        
        // Pack-Header schreiben
        bytes += WritePackHeader( &pack_data[0], last_timestamp[track_code] );

        // System-Header schreiben
        if( !sys_header ) {
            bytes += WriteSysHeader( &pack_data[bytes], 4000000, VBR );
        }
        
        // PES-Header schreiben
        if( mpeg_type == 1 )
            bytes += WriteMpeg1PesHeader( &pack_data[bytes], track_code, data_size - bytes, timestamp, -1 );
        else
            bytes += WriteMpeg2PesHeader( &pack_data[bytes], track_code, data_size - bytes, timestamp, -1 );        
        
        // Ist der Timestamp fr das Pakte richtig?
        // Wenn der aktuelle Frame nicht genug Daten hat, um das
        // Paket zu fllen, fngt in diesem Paket ein neuer Frame an
        // -> Timestamp vom nchsten Frame bernehmen
        if( timestamp < 0  && Puffer->Packs() > 1 && data_size - bytes > Puffer->PackBytes(0) ) {
            // Nchstes Paket
            timestamp = Puffer->Timestamp(1);
        } else if( Puffer->Packs() > 1 && timestamp == Puffer->Timestamp(1) && data_size - bytes <= Puffer->PackBytes(0) ) {
            // Mit einem Timestamp ist nicht genug Platz fr das nchste Paket
            // -> aktueller Timstamp
            timestamp = Puffer->Timestamp(0);
        } else {
            ts_ok = true;
        }

    } while( !ts_ok );

    sys_header = true;
    
    // Daten in Puffer kopieren
    if( data_size - bytes >  Puffer->Bytes() ) {
        // Es werden mehr Daten bentigt, um das Paket zu fllen (Sektorgrenze)
        // Als im Puffer zur Verfgung stehen.
        // -> Rest mit unntzen Bytes auffllen
        
        //data_size = Puffer->Bytes();
        size_t buf_bytes = Puffer->Bytes();
        //size_t bytes = Puffer->PackBytes();
        Puffer->GetData( 0, pack_data + bytes, buf_bytes );
        Puffer->ClearData( buf_bytes );
        
        memset( pack_data + bytes + buf_bytes, PADDING_DATA, data_size - bytes - buf_bytes );
    } else {
        Puffer->GetData( 0, pack_data + bytes, data_size - bytes );
        Puffer->ClearData( data_size - bytes );
    }

    // Daten schreiben
    if( write( stream, (char*)pack_data, data_size ) < data_size )
        return 0;
        
    bytes_written += data_size;
    
    return stat;        
}








/**Pack Header schreiben<br>
 *Rckgabe: >0 geschriebene Bytes <br>
 *          0 Fehler
 */
int System_Write::WritePackHeader( uint8_t *buf, double timestamp )
{
    int header_size = 0;
    uint16_t rate = 0;

    if( timestamp > 0 ) {
        rate = (uint16_t)((bytes_written * 8) / timestamp);
    }
        
    // Header-Code schreiben
    memset( header_data, 0, MAX_HEADER_SIZE );
    memcpy( header_data, CHUNK_START, 3 );
    header_data[3] = PACK_HEADER_CODE;

        
    // Pack-Header schreiben
    if( mpeg_type == 2 ) {
        //Mpeg2 SCR-Marker
        header_data[4] = 0x40;

        // Timestamp schreiben
        WriteTimestampMpeg2( timestamp, header_data + 4 );
       
        header_data[10] = rate >> 14;
        header_data[11]  = 0xff & (rate >> 6);
        header_data[12]  = (0x03 | ((rate & 0x3f) << 2));
        header_data[13]  = 0xf8;
        
        header_size = 14;
        
        // (header_size - 14) schreiben
        //header_data[13] |= 0;
                  
    } else if( mpeg_type == 1 ) {
        // Mpeg1 SCR-Marker
        header_data[4] = 0x20;

        // Timestamp schreiben
        WriteTimestamp( timestamp, header_data + 4 );

        // Bitrate des Streams schreiben
        header_data[9]  = 0x80 | (rate >> 15);
        header_data[10] = 0xff & (rate >> 7);
        header_data[11] = 0x01 | ((rate & 0x7f) << 1);
        
        header_size = 12;
        
    } else {
        return 0;
    }

    // schreiben
    memcpy( buf, header_data, header_size );
    
    return header_size;
}






/**Mpeg1-PES Header schreiben<br>
  *Rckgabe: >0 geschriebene Bytes<br>
  *          0 Fehler
  */
int System_Write::WriteMpeg1PesHeader( uint8_t *buf, int track_code , int data_size, double pts, double dts )
{
    /**** MPEG 1 *****/
    uint16_t header_size = 4, header_data_size;

    if( track_code < 0xc0 || track_code > 0xef ) {
        ERROR_TXT( _("Track type 0x%02x not supported by fileformat MPEG %i."), track_code, mpeg_type );
        return -1;
    }

    // Header-Code schreiben
    memset( header_data, 0, MAX_HEADER_SIZE );
    memcpy( header_data, CHUNK_START, 3 );
    header_data[3] = track_code;    

    header_size = 6;
                
    // STD Buffer-Scale und Buffer-Size nicht setzen

    // Marker - keine Timestamps
    header_data[header_size] = 0x0f;
        
    // Mpeg1 Timestamp schreiben:
    if( dts >= 0 ) {
        header_data[header_size] = (3 << 4);    // Marker - PTS + DTS
        WriteTimestamp( pts, header_data + header_size - 1 );

        header_data[header_size+5] = (1 << 4);    // Marker - DTS
        WriteTimestamp( dts, header_data + header_size + 5 );

        header_size += 9;
    } else if( pts >= 0 ) {
        header_data[header_size] = (2<<4);      // Marker - PTS
        WriteTimestamp( pts, header_data + header_size);
        header_size += 4;
    }
    header_size++;

    // Gre der Daten berechnen
    //header_data_size = data_size + header_size - 6;
    header_data_size = data_size - 6;
    header_data[4] = header_data_size >> 8;
    header_data[5] = header_data_size & 0xFF;
    // DEBUG_TXT( "Track = 0x%02x,    header_data_size = %i,    header_size = %i,    timestamp = %f", track_code, header_data_size, header_size, pts );

    // schreiben
    memcpy( buf, header_data, header_size  );
        
    return header_size;
}






/**Mpeg2-PES Header schreiben<br>
  *Rckgabe: >0 geschriebene Bytes<br>
  *          0 Fehler
  */
int System_Write::WriteMpeg2PesHeader( uint8_t *buf, int track_code , int data_size, double pts, double dts )
{
    /***** MPEG 2 *****/
    uint16_t header_size = 4, header_data_size;
    int pts_flags = 0;
    int ext_code = 0;
    uint16_t start_pos=0;

    if( track_code < 0xc0 || track_code > 0xef ) {
        ext_code = track_code;
        track_code = PES_EXT_CODE;
    }

    // Header-Code schreiben
    memset( header_data, 0, MAX_HEADER_SIZE );
    memcpy( header_data, CHUNK_START, 3 );
    header_data[3] = track_code;

    header_size = 9;

    header_data[6] = 0x80;

    //if( (Data(6)>>4)&3 )
    //    css=true;

    if( dts >= 0 && pts >= 0 ) {
        pts_flags = 3;
    } else if( pts >= 0 ) {
        pts_flags = 2;
    }

    header_data[7] = pts_flags << 6;

    // Tiemstamp einlesen
    if( pts_flags == 2 ){
        WriteTimestamp( pts, &header_data[9] );
        header_size += 5;
    } else if ( pts_flags == 3 ){
        WriteTimestamp( pts, &header_data[9] );
        WriteTimestamp( dts, &header_data[14] );
        header_size += 10;
    }

    header_data[8] = header_size - 9;


    // erweiterten Tracktyp einlesen
    if( track_code == PES_EXT_CODE ) {
        header_data[header_size] = ext_code;
        // 0x20..0x3F  subtitle
        // 0x80..0x9F  AC3 audio
        // 0xA0..0xBF  PCM audio

        header_size++;
        if( (ext_code & 0xE0) == TRACK_CODE_AC3 || (ext_code & 0xE0) == TRACK_CODE_PCM ) {
            header_data[header_size] = 1;   // num_frames
            header_data[header_size + 1] = start_pos >> 8;
            header_data[header_size + 2] = start_pos & 0xFF;
            header_size += 3;
        }
    }

    // Gre der Daten berechnen
    //header_data_size = data_size + header_size - 6;
    header_data_size = data_size - 6;
    header_data[4] = header_data_size >> 8;
    header_data[5] = header_data_size & 0xFF;

    // schreiben
    memcpy( buf, header_data, header_size  );

    return header_size;
}







/**Sys-Headerschreiben
  *Rckgabe: geschriebene Bytes
  */
int System_Write::WriteSysHeader( uint8_t *buf, uint32_t mux_rate, bool vbr )
{
    int header_size = 0;
    int CSPS = 1;
    bool audio_lock = true, video_lock = true;
     
    // Header-Code schreiben
    memset( header_data, 0, MAX_HEADER_SIZE );
    memcpy( header_data, CHUNK_START, 3 );
    header_data[3] = SYSTEM_HEADER_CODE;
    
    header_size = 6;
    
    header_data[header_size++] = 0x80 | (mux_rate >>15);
    header_data[header_size++] = 0xff & (mux_rate >> 7);
    header_data[header_size++] = 0x01 | ((mux_rate & 0x7f)<<1);
    header_data[header_size++] = (a_tracks << 2)| (!vbr << 1) | CSPS;
    header_data[header_size++] = (audio_lock << 7) | (video_lock << 6) | 0x20 | v_tracks;

    header_data[header_size++] = 0xff;

    // Puffergren fr die einzelnen Tracks schreiben
    int buffer_size = 50;
    int buffer_scale = 1;            // 1: Puffergre = buffer_size * 1024
                                     // 0: Puffergre = buffer_size * 128
    if( v_tracks > 0 ) {
        header_data[header_size++] = TRACK_CODE_MPEG;
        header_data[header_size++] = 0xc0 | ( buffer_scale << 5) | ( buffer_size >> 8);
        header_data[header_size++] = buffer_size & 0xff;
    }

    if( a_tracks > 0 ) {
        header_data[header_size++] = TRACK_CODE_MP3;
        header_data[header_size++] = 0xc0 | ( buffer_scale << 5) | ( buffer_size >> 8);
        header_data[header_size++] = buffer_size & 0xff;
    }

    // Gre des Headers
    header_data[4] = ( header_size - 6 ) >> 8;
    header_data[5] = ( header_size - 6 ) & 0xff;

    // schreiben
    memcpy( buf, header_data, header_size );

    return header_size;
}







/**Timestamp schreiben<br>
 *Rckgabe: >0 Timestamp in Sec.<br>
 *          <0 Fehler
 */
void System_Write::WriteTimestamp( double Seconds, uint8_t *data )
{
    uint8_t c;
    uint16_t d,e;
    uint64_t pts;

    pts = (uint64_t)(Seconds * 90000.0);

    c = ((pts >> 30) & 0x07) << 1;
    d = ((pts >> 15) & 0x7FFF) << 1;
    e = (pts & 0x7FFF) << 1;

    c |= 1; d |= 1; e |= 1;

    data[0] |= c;
    data[1] = d >> 8;
    data[2] = d & 0xFF;
    data[3] = e >> 8;
    data[4] = e & 0xFF;
}



/**Mpeg2 SCR-Timestamp schreiben<br>
 *Rckgabe: >0 Timestamp in Sec.<br>
 *          <0 Fehler
 */
void System_Write::WriteTimestampMpeg2( double Seconds, uint8_t *data )
{

    uint64_t scr, scr_ext;
    uint8_t a;
    uint16_t b, c, d;

    scr     = (uint64_t)(Seconds * 90000);
    scr_ext = (uint64_t)(Seconds * 90000 * 300) % 300;

    a = ((scr >> 30) & 0x07) << 1;
    b = ((scr >> 15) & 0x7FFF) << 1;
    c = (scr & 0x7FFF) << 1;
    d = (scr_ext & 0x1ff) << 1;

    a |= 1; b |= 1; c|=1; d |= 1;
    
    data[0] |= (a << 2) | (b >> 14);
    data[1] = (b >> 6) & 0xff;
    data[2] = ((b & 0x3f) << 2 ) | (c >> 14);
    data[3] = (c >> 6) & 0xff;;
    data[4] = ((c & 0x3f) << 2 ) | (d >> 8);
    data[5] = d & 0xff;
}
