/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/



#include "chunk.h"
#include "debug.h"
#include "MpegCodes.h"



Chunk::Chunk( MpegIO *file, MpegTrack *tr)
{
    track = tr;
    puffer = tr->stream;
    mpeg_file = file;
}





Chunk::~Chunk()
{
}






/**Den nchsten Block aus dem Puffer in das Datenfeld kopieren<\br>
  *Rckgabe: >0  erfolg
  *           0  keine weiteren Daten
  *          -1  Fehler
  */
int Chunk::Lesen( uint8_t *data, size_t &bytes )
{
    long ende = Bytes();

    if( ende <= 0 ) {
        bytes = 0;
        return ende;
    }

    // Daten kopieren
    if( bytes >= (size_t)ende )
        bytes = ende;

    puffer->GetData( 0, data, bytes );

    // Daten aus Puffer entfernen
    puffer->ClearData( ende );

    return 1;
}






/**Liefert den Startcode des nchsten Blocks
  *Rckgabe: >=0  Startcode
  *           -1  Fehler / Dateiende
  */
int Chunk::Startcode( void )
{
    int stat = 0;
    uint8_t data;

    // gehe auf den Anfang des nchsten Blocks
    stat = Sync();
    if( stat <= 0 )
        return -1;

    data = puffer->Data( 3 );
    
    return data;
}





/**berspringt den nchsten Block
  */
int Chunk::Skip( void )
{
    long ende = Bytes();

    if( ende <= 0 )
        return ende;

    // Daten aus Puffer entfernen
    puffer->ClearData( ende );

    return 1;
}





/**Bytes im nchsten Block
  */
long Chunk::Bytes( void )
{
    int stat = 0;
    long ende;

    // gehe auf den Anfang des nchsten Blocks
    stat = Sync();
    if( stat <= 0 )
        return stat;

    // finde Ende des Blocks
    ende = puffer->Finden( 3, CHUNK_START, 3 );

    // Ende nicht im Puffer gefunden
    if( ende < 0 ) {
        stat = PufferFuellen();
        if( stat < 0 )
            return stat;

        ende = puffer->Finden( 3, CHUNK_START, 3 );
    }

    // immernoch nicht gefunden
    // Bei eof den Rest als einen Block behandeln
    if( ende < 0 && stat == 0 )
        return puffer->Bytes();
    else if( ende < 0 )
        return -1;

    return ende;
}





/**Auf den Start eines Blockes synchronisieren<\br>
  *Rckgabe: >0 Erfolg<\br>
  *           0 keine weiteren Daten<\br>
  *          -1 Fehler
  */
int Chunk::Sync( void )
{
    int stat = 0;

    // suche nach Start des ersten Chunks im Puffer
    long begin = puffer->Finden( 0, CHUNK_START, 3 );

    if( begin < 0 ) {
        // Puffer auffllen
        if( (stat = PufferFuellen()) <= 0 ){
            return stat;
        }
        // suche nochmal
        begin = puffer->Finden( 0, CHUNK_START, 3 );

        if( begin < 0 )
            return -1;
    }

    if( begin > 0 ) {
        //DEBUG_TXT( "Headercode = x%02x x%02x x%02x x%02x", puffer->Data(0), puffer->Data(1), puffer->Data(2), puffer->Data(3) );
        DEBUG_TXT( "Resync Stream %i Bytes", begin );
        puffer->ClearData( begin );
        //DEBUG_TXT( "Headercode = x%02x x%02x x%02x x%02x", puffer->Data(0), puffer->Data(1), puffer->Data(2), puffer->Data(3) );
    }

    return 1;
}





/**Zeichen an Offset in Chunk
  */
uint8_t Chunk::Data( size_t offset )
{
    return puffer->Data( offset );
}
