/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <stdio.h>
#include <assert.h>
#include <string.h>

#include "debug.h"

#include "OQuicktime.h"
#include "VideoFile/codecplugins.h"
#include "VideoFile/translate.h"



/***** Konstruktor / Destruktor *****/
OQuicktimeIO::OQuicktimeIO( const String Dateiname, IO_Mode mode )
:VideoFile( Dateiname, "Quicktime", mode )
{
    char *datei_name;

    // Variablen initialisieren
    Datei = NULL;

    datei_name = new char[ Dateiname.length() + 1 ];
    strcpy( datei_name, Dateiname.chars() );

    // Quicktime Datei ffnen
    if( iomode == read ) {
        // Datei zum lesen ffnen
        Datei = oqt_open_file(datei_name, OQT_MODE_READ);
        if( oqt_read_headers(Datei) ) {
            ErrorStatus() = Errors::FileType;
            throw( ErrorStatus() );
        }
        
        // Anzahl der Tracks
        SetVideoTracks( oqt_get_video_track_count( Datei ) );
        SetAudioTracks( oqt_get_audio_track_count( Datei ) );
    } else if( iomode == write ) {
        // Datei zum schreiben ffnen
        Datei = oqt_open_file(datei_name, OQT_MODE_WRITE);
        if( Datei == NULL ) {
            ErrorStatus() = Errors::FileType;
            throw( ErrorStatus() );
        }
        oqt_write_moviedata_init( Datei );
    } else {
        ErrorStatus() = Errors::IO_Mode;
        throw( ErrorStatus() );
    }

    delete [] datei_name;

    if( Datei == NULL ) {
        ErrorStatus() = Errors::FileType;
        throw( ErrorStatus() );
    }
}




OQuicktimeIO::~OQuicktimeIO()
{
    // Tracks schlieen
    deleteTracks();

    
    if( iomode == write ) {
        /* End of the movie data - and write 'moov' atom */
        oqt_write_moviedata_end( Datei );
    }
    
    // Datei schlieen
    if( Datei != NULL )
        oqt_close( Datei );
}







/***** Quicktime Datei-Descriptor ***********************************/
oqt_t *OQuicktimeIO::QuicktimeDatei( void )
{
    return Datei;
}




/***** Video-Funktionen *********************************************/
// Video codec initialisieren
// Rckgabewert: false - Fehler
//               true  - Erfolg
V_Track *OQuicktimeIO::openVideoTrack( double Rate, int x, int y, const char *Codec )
{
    if( iomode != write )
        return NULL;

    // Nummer des Tracks
    int track = oqt_get_video_track_count( Datei );
    
    return new OQuicktimeVtr( this, track, Rate, x, y, Codec );
}

// Track ffnen
V_Track *OQuicktimeIO::openVideoTrack( int track )
{
    if( iomode != read )
        return NULL;
        
    if( track >= VideoTracks() )
        return NULL;
        
    return new OQuicktimeVtr( this, track );
}





/***** Audiofunktionen **********************************************/

// Audio codec initialisieren
A_Track *OQuicktimeIO::openAudioTrack( int Sample_rate, int Channels, const char *Codec )
{
    if( iomode != write )
        return NULL;
        
    // Nummer des Tracks
    int track = oqt_get_audio_track_count( Datei );

    return new OQuicktimeAtr( this, track, Sample_rate, Channels, Codec );
}

// Track ffnen
A_Track *OQuicktimeIO::openAudioTrack( int track )
{
    if( iomode != read )
        return NULL;
        
    if( track >= AudioTracks() )
        return NULL;

    return new OQuicktimeAtr( this, track );
}
















/*****************************************************************************/
/***** Videotrack-Funktionen *************************************************/
/*****************************************************************************/
// Video Track initialisieren
// Rckgabewert: false - Fehler
//               true  - Erfolg
OQuicktimeVtr::OQuicktimeVtr( OQuicktimeIO *File, int track, double rate, int x, int y, const char *Codec_name )
:Track(File, track, Codec_name), V_Track( File, rate, x, y, Codec_name )
{
    file = File;

    if( Codec_name == NULL ) {
        ERROR_TXT( _("Video codec not implemented!") );
    }

    Datei = file->QuicktimeDatei();

    FrameBuf = NULL;
    BufGr = 0;

    // Codec erstellen
    // passenden Codec zuordnen und erstellen
    codec = AV_Codecs->CreateVCodec( this, "Quicktime", CodecName() );

    char codecid[5];
    strncpy( codecid, CodecID().chars(), 4 );
    oqt_add_video_track( Datei, Width(), Height(), Rate(), codecid );

    // Bildfrequenz mu nochmal explizit gesetzt werden
    // Fehler in Quicktime ???
    // oqt_set_video_framerate( Datei, Rate() );

    
    if( codec == NULL ) {
        // benutze eingebauten Quicktime-Codec
        if( oqt_supported_video( Datei, Number() ) ) {
            DEBUG_TXT( _("Create internal Quicktime codec for codec %s"), CodecName().chars() );
            codec = new OQuicktimeV( this );
        } else {
            ERROR_TXT( _("Video codec not implemented!") );
            codec = new V_Codec( this );
        }
    }

}





OQuicktimeVtr::OQuicktimeVtr( OQuicktimeIO *File, int track )
:Track(File, track), V_Track( File, track )
{
    char *Codec_name;

    file = File;
    Datei = file->QuicktimeDatei();

    FrameBuf = NULL;
    BufGr = 0;

    // Informationen aus Datei auslesen
    SetFrames( oqt_get_video_length( Datei, Number() ) );
    breite = oqt_get_video_width( Datei, Number() );
    hoehe = oqt_get_video_height( Datei, Number() );
    SetRate( oqt_get_video_framerate( Datei, Number() ) );

    Codec_name = oqt_get_video_compressor( Datei, Number() );
    char codecname[5];
    strncpy( codecname, Codec_name, 4 );
    codecname[4] = '\0';
    SetCodecName( codecname );


    // Codec erstellen
    // passenden Codec zuordnen und erstellen
    codec = AV_Codecs->CreateVCodec( this, "Quicktime", CodecName() );

    if( codec == NULL ) {
        // benutze eingebauten Quicktime-Codec
        if( oqt_supported_video( Datei, Number() ) ) {
            DEBUG_TXT( _("Create internal Quicktime codec for codec %s"), CodecName().chars() );
            codec = new OQuicktimeV( this );
        } else {
            ERROR_TXT( _("Video codec not implemented!") );
            codec = new V_Codec( this );
        }
    }
}



OQuicktimeVtr::~OQuicktimeVtr()
{
    delete codec;
    codec = NULL;
    delete [] FrameBuf;
}





// zu Bildposition springen
int OQuicktimeVtr::Seek( long bild )
{
    if( bild >= oqt_get_video_length( Datei, Number() ) )
        return 0;
    
    if( !oqt_set_video_position( Datei, Number(), bild ) ) {
        SetPosition( bild );
        return 1;
    }

    return 0;
}




// letzter Keyframe
long OQuicktimeVtr::GetKeyFrame( long bild )
{
    long i;

    if( IOMode() != VideoFile::read )
        return bild;


    i = oqt_get_video_keyframe_before( Datei, Number(), bild );
    if( i > 0 && i <= bild )
        return i;

    if( i == 0 && bild < 1000 )
        return 0;

    return bild;
}




int OQuicktimeVtr::readRaw( unsigned char *&data, long &bytes )
{
    long gr;

    // gehe zu aktuellem Bild
    if( oqt_set_video_position( Datei, Number(), Position() ) ) {
        data = NULL;
        return 0;
    }

    // Gre des nchsten Datensatzes
    gr = oqt_get_video_framesize( Datei, Number(), Position() );
    bytes = gr;

    // Speicherbereich anforden
    if( BufGr <= gr ) {
        delete [] FrameBuf;
        BufGr = gr + 1;
        FrameBuf = new unsigned char[BufGr];
    }

    if( oqt_read_data( Datei, (char*)FrameBuf, bytes ) >= 0 ) {
        data = FrameBuf;
        return 1;
    } else {
        data = NULL;
        return 0;
    }
}




int OQuicktimeVtr::writeRaw( unsigned char *data, long bytes, int keyframe )
{

    oqt_write_video_data( Datei, Number(), data, bytes, keyframe & IS_KEYFRAME ? 1 : 0 );
    
    /*
    if( keyframe & IS_KEYFRAME ) {
        quicktime_insert_keyframe( Datei, Position() - 1, Number() );
    }
    */

    return 1;
}













/*****************************************************************************/
/***** Videofunktionen *******************************************************/
/*****************************************************************************/


// Video codec initialisieren
// Rckgabewert: false - Fehler
//               true  - Erfolg
OQuicktimeV::OQuicktimeV( OQuicktimeVtr *Track )
:V_Codec( Track )
{
    //QuicktimeIO *QtFile = Track->file;

    Datei = Track->Datei;
    track = Track;

    // initialisiere row_pointers
    row_pointers = new unsigned char*[hoehe];
    assert( row_pointers != NULL );

    // Farbraum festlegen
    ColorModel( RGB24 );

}





OQuicktimeV::~OQuicktimeV()
{
    // Speicher freigeben
    delete [] row_pointers;
}






// Farbraum festlegen
bool OQuicktimeV::ColorModel( int cmodel )
{
    if( cmodel_qt( cmodel ) <= 0 ) {
        // verwende intern Standardfarbraum und benutze Konvertierung
        cmodel = qt_cmodel( oqt_preferred_colormodel( Datei, track->Number() ) );
        
        if( cmodel <= 0 ) {
            cmodel = RGB24;
        }
        
        V_Codec::ColorModel( cmodel );
        
        return false;
    }
    
    // wird Farbraum von Quicktimecodec untersttzt?
    if( (iomode == VideoFile::write && oqt_writes_colormodel(Datei, track->Number(), cmodel_qt(cmodel) )) ||
        (iomode == VideoFile::read && oqt_reads_colormodel(Datei, track->Number(), cmodel_qt(cmodel) )) ) {
        return V_Codec::ColorModel( cmodel );
    } else {
        // verwende intern Standardfarbraum und benutze Konvertierung
        cmodel = qt_cmodel( oqt_preferred_colormodel( Datei, track->Number() ) );
        
        if( cmodel <= 0 ) {
            cmodel = RGB24;
        }
                
        V_Codec::ColorModel( cmodel );
        
        return false;
    }
    
    return V_Codec::ColorModel( cmodel );
}






// Video-Bild schreiben
int OQuicktimeV::encode( VideoFrame *bild )
{
    uint8_t *bild_buf = NULL;
    
    if( bild->GetFrame( bild_buf ) <= 0 )
        return 0;
        
    /***** Rowpointer erstellen *****/
    erz_row_pointers( row_pointers, bild_buf );

    oqt_encode_video( Datei, track->Number(), cmodel_qt(color_model), row_pointers );

    return 1;
}






// Video-Bild lesen
int OQuicktimeV::decode( VideoFrame *bild )
{
    uint8_t *bild_buf = NULL;

    if( track->Position() >= oqt_get_video_length( Datei, track->Number() ) ) {
        return 0;
    }

    /***** Rowpointer erstellen *****/
    if( bild->FramePointer( bild_buf ) <= 0 )
        return 0;

    erz_row_pointers( row_pointers, bild_buf );

    //quicktime_decode_video( Datei, row_pointers, tr );


    oqt_decode_video( Datei, track->Number(), cmodel_qt(color_model), row_pointers );
    track->SetPosition( track->Position() + 1 );

    return 1;
}






/***** Hilfsfunktionen **********************************************/
int OQuicktimeV::qt_cmodel( int cmodel )
{
    int c_model = 0;

    switch( cmodel ) {
        case BC_YUV420P:
            c_model = YUV420P;
            break;
        case BC_YUV422P:
            c_model = YUV422P;
            break;
        case BC_YUV422:
            c_model = YUV422;
            break;
        case BC_RGB565:
            c_model = RGB16;
            break;
        case BC_BGR565:
            c_model = RGB16 | BGR;
            break;
        case BC_RGB888:
            c_model = RGB24;
            break;
        case BC_BGR888:
            c_model = RGB24 | BGR;
            break;
        case BC_RGBA8888:
            c_model = RGB32;
            break;
        case BC_BGR8888:
            c_model = RGB32 | BGR;
            break;
        default:
            //file->Fehler( "qt_c_model: Farbmodel %i nicht untersttzt\n", color_model_intern );
            c_model = -1;
    }

    return c_model;
}



int OQuicktimeV::cmodel_qt( int cmodel )
{
    int c_model = 0;

    switch( cmodel ) {
        case YUV420P:
            c_model = BC_YUV420P;
            break;

        case YUV422P:
            c_model = BC_YUV422P;
            break;
        case YUV422:
            c_model = BC_YUV422;
            break;
        case RGB16:
            c_model = BC_RGB565;
            break;
        case RGB16 | BGR:
            c_model = BC_BGR565;
            break;
        case RGB24:
            c_model = BC_RGB888;
            break;
        case RGB24 | BGR:
            c_model = BC_BGR888;
            break;
        case RGB32:
            c_model = BC_RGBA8888;
            break;
        case RGB32 | BGR:
            c_model = BC_BGR8888;
            break;
        case RGB15:
        case YUV420:
        default:
            //file->Fehler( "qt_c_model: Farbmodel %i nicht untersttzt\n", color_model_intern );
            c_model = -1;
    }

    return c_model;
}






void OQuicktimeV::erz_row_pointers( unsigned char **row_pointers, unsigned char *buf )
{
    switch( color_model ) {
    case YUV420P:
        row_pointers[0] = buf;                                           // y - Plane
        row_pointers[1] = row_pointers[0] + breite * hoehe;              // u - Plane
        row_pointers[2] = row_pointers[1] + breite/2 * ((hoehe + 1)/2) ; // v - Plane
        break;
    case YUV422P:
        row_pointers[0] = buf;                                           // y - Plane
        row_pointers[1] = row_pointers[0] + breite * hoehe;              // u - Plane
        row_pointers[2] = row_pointers[1] + breite/2 * hoehe;            // v - Plane
        break;
    case RGB24:
    case RGB24|BGR:
        for( int i=0; i<hoehe; i++ ) {
            row_pointers[i] = buf + i * 3 * breite;
        }
        break;
    default:
        ERROR_TXT( _("erz_row_pointers: Colormodel %i not implemented!"), color_model );
        return;
    }
}















/*****************************************************************************/
/***** Audiotrack - Funktionen ***********************************************/
/*****************************************************************************/
OQuicktimeAtr::OQuicktimeAtr( OQuicktimeIO *File, int track, int Sample_rate, int Channels, const char *Codec_name )
:Track(File, track, Codec_name), A_Track( File, Sample_rate, Channels, Codec_name )
{
    file = File;
    Datei = file->Datei;

    FrameBuf = NULL;
    BufGr = 0;
    BufSamples = 0;
    internal_codec = false;
    
    if( CodecName() == NULL ) {
        ERROR_TXT( _("Audio codec not implemented!") );
    }

    // Codec erstellen
    // passenden Codec zuordnen und erstellen
    codec = AV_Codecs->CreateACodec( this, "Quicktime", CodecName() );

    char codecid[5];
    strncpy( codecid, CodecID().chars(), 4 );
    oqt_add_audio_track( Datei, Channels, SampleRate(), 16, codecid );

    if( codec == NULL ) {
        // benutze eingebauten Quicktime-Codec
        if( oqt_supported_audio( Datei, Number() ) ) {
            DEBUG_TXT( _("Create internal Quicktime codec for codec %s"), CodecName().chars() );
            // Codec-Objekt erstellen
            codec = new OQuicktimeA( this );
            internal_codec = true;
        } else {
            ERROR_TXT( _("Audio codec not implemented!") );
            codec = new A_Codec( this );
        }
    }
}




OQuicktimeAtr::OQuicktimeAtr( OQuicktimeIO *File, int track )
:Track(File, track), A_Track( File, track )
{
    file = File;
    Datei = file->Datei;
    
    FrameBuf = NULL;
    BufGr = 0;
    BufSamples = 0;
    internal_codec = false;

    // Informationen au Datei auslesen
    SetSampleRate( oqt_get_audio_samplerate( Datei, Number() ) );
    total_samples = oqt_get_audio_length( Datei, Number() );
    SetChannels( oqt_get_audio_channels( Datei, Number() ) );
    SetFrames( oqt_get_audio_chunks( Datei, Number() ) );
    
    
    // zustzliche Informationen fr den Codec
    oqt_trak_t *atrak = Datei->atracks[track].track;
    
    //SoundDescription Version 1
    if( atrak->mdia.minf.stbl.stsd.table[0].version == 1 ) {
        SetFrameSamples( atrak->mdia.minf.stbl.stsd.table[0].samplesPerPacket );
        
        if( FrameSamples() > 0 )
            SetBitsPerSecond( (int)((atrak->mdia.minf.stbl.stsd.table[0].bytesPerFrame * SampleRate() * 8) / FrameSamples()) );
    }

    
    
    // berechne die Framerate der Audiochunks aus der Gre des ersten
    // chunks. Haben in Quicktime alle Audio-chunks die gleiche Gre???
    SetRate( (double)oqt_get_audio_chunk_samples( Datei, Number(), 0 ) / (double)SampleRate() );
    
    char *c_name;
    c_name = oqt_get_audio_compressor( Datei, Number() );
    char codecname[5];
    strncpy( codecname, c_name, 4 );
    codecname[4] = '\0';
    SetCodecName( codecname );

    
    // Codec erstellen
    // passenden Codec zuordnen und erstellen
    codec = AV_Codecs->CreateACodec( this, "Quicktime", CodecName() );

    if( codec == NULL ) {
        // benutze eingebauten Quicktime-Codec
        if( oqt_supported_audio( Datei, Number() ) ) {
            DEBUG_TXT( _("Create internal Quicktime codec for codec %s"), CodecName().chars() );
            codec = new OQuicktimeA( this );
            internal_codec = true;
        } else {
            ERROR_TXT( _("Audio codec not implemented!") );
            codec = new A_Codec( this );
        }
    }
}






OQuicktimeAtr::~OQuicktimeAtr()
{

    if( IOMode() == VideoFile::write && !internal_codec ) {
        // vom Codec gendertes Format bernehmen
        SetAFormat( *codec );
        
        oqt_trak_t *atrak = Datei->atracks[Number()].track;
        
        //SoundDescription Version 1
        if( FrameBytes() == 0 ) { // Variable Bitrate
            atrak->mdia.minf.stbl.stsd.table[0].compression_id = 65534;
        } else {
            atrak->mdia.minf.stbl.stsd.table[0].compression_id = 65535;
        }
        
        atrak->mdia.minf.stbl.stsd.table[0].version = 1;
        atrak->mdia.minf.stbl.stsd.table[0].samplesPerPacket = FrameSamples();
        atrak->mdia.minf.stbl.stsd.table[0].bytesPerPacket = FrameBytes() / Channels();
        atrak->mdia.minf.stbl.stsd.table[0].bytesPerFrame = FrameBytes();
        atrak->mdia.minf.stbl.stsd.table[0].bytesPerSample = Bits() / 8;
        
        //atrak->mdia.minf.stbl.stsz.sample_size = 0;
    }
    
        
        
    delete codec;
    codec = NULL;
    
    // schreibe restliche Daten in Datei
    if( IOMode() == VideoFile::write && BufSamples > 0 ) {
        oqt_write_audio_data( Datei, Number(), FrameBuf, BufBytes, BufSamples, 1, NULL );
        BufSamples = 0;
        BufBytes = 0;
    }
        
    delete [] FrameBuf;
}





// zu Position springen
int OQuicktimeAtr::Seek( long block )
{
    if( block >= oqt_get_audio_chunks( Datei, Number() ) )
        return 0;
    
    if( !oqt_set_audio_position_to_chunk( Datei, Number(), block ) ) {
        SetPosition( block );
        return 1;
    }

    return 0;
}


/** Aktuelle Position in Sekunden
  */
double OQuicktimeAtr::CurrentTime( void )
{
    if( internal_codec ) {
        return (double)oqt_get_audio_position( Datei, Number() ) / SampleRate();
    } else {
        return (double)Position() / Rate();
    }
}
    


int OQuicktimeAtr::readRaw( unsigned char *&data, long &bytes )
{
    long gr;

    // gehe zu aktuellem chunk
    if( oqt_set_audio_position_to_chunk( Datei, Number(), Position() ) ) {
        data = NULL;
        return 0;
    }

    // Gre des nchsten Datensatzes
    gr = oqt_get_audio_chunksize( Datei, Number(), Position() );
    bytes = gr;

    // Speicherbereich anforden
    if( BufGr <= gr ) {
        delete [] FrameBuf;
        BufGr = gr + 1;
        FrameBuf = new unsigned char[BufGr];
    }

    if( oqt_read_data( Datei, (char*)FrameBuf, bytes ) >= 0 ) {
        data = FrameBuf;
        return 1;
    } else {
        data = NULL;
        return 0;
    }
}




int OQuicktimeAtr::writeRaw( unsigned char *data, long bytes, int samp )
{

    // Puffergre: 1 Sekunde unkompremiert
    long gr = SampleRate() * Channels() * 2;
    
    // Speicherbereich anforden
    if( BufGr <= gr ) {
        delete [] FrameBuf;
        BufGr = gr + 1;
        FrameBuf = new unsigned char[BufGr];
    }

    // Audiodaten in Puffer einreihen
    if( BufBytes + bytes <= BufGr ) {
        memcpy( FrameBuf + BufBytes, data, bytes );
        BufBytes += bytes;
        BufSamples += samp;
    } else {
        oqt_write_audio_data( Datei, Number(), FrameBuf, BufBytes, BufSamples, 1, NULL );
        BufSamples = 0;
        BufBytes = 0;
        
        if( bytes <= BufGr ) {
            memcpy( FrameBuf, data, bytes );
            BufBytes = bytes;
            BufSamples = samp;
        } else {
            oqt_write_audio_data( Datei, Number(), data, bytes, samp, 1, NULL );;
            return 1;
        }
    }
    
    // schreibe in Datei, wenn chunk gro genug ist
    if( BufSamples >= SampleRate() * 0.25 ) {
        oqt_write_audio_data( Datei, Number(), FrameBuf, BufBytes, BufSamples, 1, NULL );
        BufSamples = 0;
        BufBytes = 0;
    }
    
    return 1;
}














/*****************************************************************************/
/***** Audiofunktionen *******************************************************/
/*****************************************************************************/



// Audio codec initialisieren
OQuicktimeA::OQuicktimeA( OQuicktimeAtr *Track )
:A_Codec( Track )
{
    Datei = Track->Datei;
    sample_pos = 0;
    sample_puffer = NULL;
    sample_puffer_gr = 0;
}




OQuicktimeA::~OQuicktimeA()
{
    delete [] sample_puffer;
    
    //if( track->IOMode() == VideoFile::write ) {
        // Audiopuffer leeren
        //oqt_encode_audio( Datei, track->Number(), NULL, 0 );
    //}
}




// Audiodaten schreiben
int OQuicktimeA::encode( float **daten, int samples )
{
    int16_t sample;
        
    if( sample_puffer_gr < samples * 2 * Channels() ) {
        delete [] sample_puffer;
        sample_puffer_gr = samples * 2 * Channels();
        sample_puffer = new uint8_t[sample_puffer_gr];
    }
    
    for( int i=0; i<samples; i++ ) {
        for( int j=0; j<Channels(); j++ ) {
            sample = (int16_t)( daten[j][i] * (1<<15) );
            sample_puffer[2*(i*Channels() + j)]     = (uint8_t)(sample & 0x00FF);
            sample_puffer[2*(i*Channels() + j) + 1] = (uint8_t)(sample >> 8);
        }
        
    }
    
    if( oqt_encode_audio( Datei, track->Number(), sample_puffer, samples ) < samples )
        return 0;

    return 1;
}




// Audiodaten lesen
int OQuicktimeA::decode( float **daten, int samples )
{
    int16_t sample;
    
    if( !SeekSample( sample_pos ) ){
            return 0;
    }
        
    if( sample_puffer_gr < samples * 2 * Channels() ) {
        delete [] sample_puffer;
        sample_puffer_gr = samples * 2 * Channels();
        sample_puffer = new uint8_t[sample_puffer_gr];
    }

    if( (samples = oqt_decode_audio( Datei, track->Number(), sample_puffer, samples * Channels() * 2 )) == 0 )
        return 0;

    // Kanle in einen Datenstrom multiplexen
    for( int i=0; i<samples; i++ ) {
        for( int j=0; j<Channels(); j++ ) {
            sample = ((int16_t)sample_puffer[2*(i*Channels()+j)+1] << 8) | (int16_t)sample_puffer[2*(i*Channels()+j)];
            daten[j][i] = (float)sample / (float)(1<<15);
        }
    }
        
    sample_pos += samples;

    return 1;
}


// zu Sample springen
int OQuicktimeA::SeekSample( long long sample )
{
    if( sample >= oqt_get_audio_length( Datei, track->Number() ) ) {
        return 0;
    }

    if( !oqt_set_audio_position( Datei, track->Number(), sample ) ) {
        sample_pos = sample;
        return 1;
    }

    return 0;
}
