/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <stdio.h>
#include <assert.h>
#include <string.h>

#include "debug.h"

#include "Quicktime.h"
#include "VideoFile/codecplugins.h"
#include "VideoFile/translate.h"



/***** Konstruktor / Destruktor *****/
QuicktimeIO::QuicktimeIO( const String Dateiname, IO_Mode mode )
:VideoFile( Dateiname, "Quicktime", mode )
{
    char *datei_name;

    // Variablen initialisieren
    Datei = NULL;

    datei_name = new char[ Dateiname.length() + 1 ];
    strcpy( datei_name, Dateiname.chars() );

    // Quicktime Datei ffnen
    if( iomode == read ) {
        // Datei zum lesen ffnen
        Datei = quicktime_open(datei_name, 1, 0);
    } else if( iomode == write ) {
        // Datei zum schreiben ffnen
        Datei = quicktime_open(datei_name, 0, 1);
    } else {
        ErrorStatus() = Errors::IO_Mode;
        throw( ErrorStatus() );
    }

    delete [] datei_name;

    if( Datei == NULL ) {
        ErrorStatus() = Errors::FileType;
        throw( ErrorStatus() );
    }
}




QuicktimeIO::~QuicktimeIO()
{
    // Tracks schlieen
    deleteTracks();

    // Datei schlieen
    if( Datei != NULL )
        quicktime_close( Datei );
}







/***** Quicktime Datei-Descriptor ***********************************/
quicktime_t *QuicktimeIO::QuicktimeDatei( void )
{
    return Datei;
}




/***** Video-Funktionen *********************************************/
// Video codec initialisieren
// Rckgabewert: false - Fehler
//               true  - Erfolg
V_Track *QuicktimeIO::openVideoTrack( double Rate, int x, int y, const char *Codec )
{
    if( iomode != write )
        return NULL;

    return new QuicktimeVtr( this, Rate, x, y, Codec );
}

// Track ffnen
V_Track *QuicktimeIO::openVideoTrack( int track )
{
    if( iomode != read )
        return NULL;
    return new QuicktimeVtr( this, track );
}





/***** Audiofunktionen **********************************************/

// Audio codec initialisieren
A_Track *QuicktimeIO::openAudioTrack( int Sample_rate, int Channels, const char *Codec )
{
    if( iomode != write )
        return NULL;

    return new QuicktimeAtr( this, Sample_rate, Channels, Codec );
}

// Track ffnen
A_Track *QuicktimeIO::openAudioTrack( int track )
{
    if( iomode != read )
        return NULL;

    return new QuicktimeAtr( this, track );
}









/*****************************************************************************/
/***** Videotrack-Funktionen *************************************************/
/*****************************************************************************/
// Video Track initialisieren
// Rckgabewert: false - Fehler
//               true  - Erfolg
QuicktimeVtr::QuicktimeVtr( QuicktimeIO *File, double rate, int x, int y, const char *Codec_name )
:Track(File, 0, Codec_name), V_Track( File, rate, x, y, Codec_name )
{
    file = File;

    if( Codec_name == NULL ) {
        ERROR_TXT( _("Video codec not implemented!") );
    }

    Datei = file->QuicktimeDatei();

    FrameBuf = NULL;
    BufGr = 0;

    // Codec erstellen
    // passenden Codec zuordnen und erstellen
    codec = AV_Codecs->CreateVCodec( this, "Quicktime", CodecName() );

    char codecid[5];
    strncpy( codecid, CodecID().chars(), 4 );
    quicktime_set_video( Datei, 1, Width(), Height(), Rate(), codecid );

    // Bildfrequenz mu nochmal explizit gesetzt werden
    // Fehler in Quicktime ???
    quicktime_set_framerate( Datei, Rate() );

    
    if( codec == NULL ) {
        // benutze eingebauten Quicktime-Codec
        if( quicktime_supported_video( Datei, 0 ) ) {
            DEBUG_TXT( _("Create internal Quicktime codec for codec %s"), CodecName().chars() );
            codec = new QuicktimeV( this );
        } else {
            ERROR_TXT( _("Video codec not implemented!") );
            codec = new V_Codec( this );
        }
    }

}





QuicktimeVtr::QuicktimeVtr( QuicktimeIO *File, int track )
:Track(File, track), V_Track( File, track )
{
    char *Codec_name;

    file = File;
    Datei = file->QuicktimeDatei();

    FrameBuf = NULL;
    BufGr = 0;

    // Informationen au Datei auslesen
    SetFrames( quicktime_video_length( Datei, Number() ) );
    breite = quicktime_video_width( Datei, Number() );
    hoehe = quicktime_video_height( Datei, Number() );
    SetRate( quicktime_frame_rate( Datei, Number() ) );

    Codec_name = quicktime_video_compressor( Datei, Number() );
    char codecname[5];
    strncpy( codecname, Codec_name, 4 );
    codecname[4] = '\0';
    SetCodecName( codecname );


    // Codec erstellen
    // passenden Codec zuordnen und erstellen
    codec = AV_Codecs->CreateVCodec( this, "Quicktime", CodecName() );

    if( codec == NULL ) {
        // benutze eingebauten Quicktime-Codec
        if( quicktime_supported_video( Datei, 0 ) ) {
            DEBUG_TXT( _("Create internal Quicktime codec for codec %s"), CodecName().chars() );
            codec = new QuicktimeV( this );
        } else {
            ERROR_TXT( _("Video codec not implemented!") );
            codec = new V_Codec( this );
        }
    }
}



QuicktimeVtr::~QuicktimeVtr()
{
    delete codec;
    codec = NULL;
}





// zu Bildposition springen
int QuicktimeVtr::Seek( long bild )
{
    if( bild >= quicktime_video_length( Datei, Number() ) )
        return 0;
    
    if( !quicktime_set_video_position( Datei, bild, Number() ) ) {
        SetPosition( bild );
        return 1;
    }

    return 0;
}




// letzter Keyframe
long QuicktimeVtr::GetKeyFrame( long bild )
{
    long i;

    if( IOMode() != VideoFile::read )
        return bild;

    if( !quicktime_has_keyframes( Datei, Number() ) )
        return bild;

    i = quicktime_get_keyframe_before( Datei, bild, Number() );
    if( i > 0 && i <= bild )
        return i;

    if( i == 0 && bild < 1000 )
        return 0;

    return bild;
}




int QuicktimeVtr::readRaw( unsigned char *&data, long &bytes )
{
    long gr;

    // gehe zu aktuellem Bild
    if( !Seek( Position() ) ) {
        data = NULL;
        return 0;
    }

    // Gre des nchsten Datensatzes
    gr = quicktime_frame_size( Datei, Position(), Number() );
    bytes = gr;

    // Speicherbereich anforden
    if( BufGr <= gr ) {
        delete [] FrameBuf;
        BufGr = gr + 1;
        FrameBuf = new unsigned char[BufGr];
    }

    if( quicktime_read_frame( Datei, FrameBuf, Number() ) >= 0 ) {
        data = FrameBuf;
        return 1;
    } else {
        data = NULL;
        return 0;
    }
}




int QuicktimeVtr::writeRaw( unsigned char *data, long bytes, int keyframe )
{

    quicktime_write_frame( Datei, data, bytes, Number() );
    if( keyframe & IS_KEYFRAME ) {
        quicktime_insert_keyframe( Datei, Position() - 1, Number() );
    }

    return 1;
}











/*****************************************************************************/
/***** Videofunktionen *******************************************************/
/*****************************************************************************/


// Video codec initialisieren
// Rckgabewert: false - Fehler
//               true  - Erfolg
QuicktimeV::QuicktimeV( QuicktimeVtr *Track )
:V_Codec( Track )
{
    //QuicktimeIO *QtFile = Track->file;

    Datei = Track->Datei;
    track = Track;

    // initialisiere row_pointers
    row_pointers = new unsigned char*[hoehe];
    assert( row_pointers != NULL );

    // Farbraum festlegen
    ColorModel( RGB24 );

}




QuicktimeV::~QuicktimeV()
{
    // Speicher freigeben
    delete [] row_pointers;
}







// Farbraum festlegen
bool QuicktimeV::ColorModel( int cmodel )
{
    if( cmodel != YUV420P || cmodel != YUV422P || cmodel != RGB24 || cmodel != (RGB24|BGR) ) {
        // verwende intern Standardfarbraum und benutze Konvertierung
        V_Codec::ColorModel( RGB24 );
        quicktime_set_cmodel( Datei, qt_c_model(RGB24) );
        return false;
    }
    
    // wird Farbraum von Quicktimecodec untersttzt?
    if( (iomode == VideoFile::write && quicktime_writes_cmodel(Datei, qt_c_model(cmodel), track->Number())) ||
        (iomode == VideoFile::read && quicktime_reads_cmodel(Datei, qt_c_model(cmodel), track->Number())) ) {
    
        quicktime_set_cmodel( Datei, qt_c_model(cmodel) );
    } else {
        // verwende intern Standardfarbraum und benutze Konvertierung
        V_Codec::ColorModel( RGB24 );
        quicktime_set_cmodel( Datei, qt_c_model(RGB24) );
        return false;
    }
    
    return V_Codec::ColorModel( cmodel );
}











// Video-Bild schreiben
int QuicktimeV::encode( VideoFrame *bild )
{
    uint8_t *bild_buf = NULL;
    
    if( bild->GetFrame( bild_buf ) <= 0 )
        return 0;
        
    /***** Rowpointer erstellen *****/
    erz_row_pointers( row_pointers, bild_buf );

    quicktime_encode_video( Datei, row_pointers, track->Number() );

    return 1;
}






// Video-Bild lesen
int QuicktimeV::decode( VideoFrame *bild )
{
    uint8_t *bild_buf = NULL;

    if( track->Position() >= quicktime_video_length( Datei, track->Number() ) ) {
        return 0;
    }

    /***** Rowpointer erstellen *****/
    if( bild->FramePointer( bild_buf ) <= 0 )
        return 0;

    erz_row_pointers( row_pointers, bild_buf );

    //quicktime_decode_video( Datei, row_pointers, tr );


    quicktime_decode_scaled( Datei,
                             0, 0, breite, hoehe,
                             breite, hoehe,
                             qt_c_model(color_model),
                             row_pointers, track->Number() );
    track->SetPosition( track->Position() + 1 );

    return 1;
}







// JPEG-Qualitt
bool QuicktimeV::jpegQuality( int q )
{
    const char *v_codec = track->CodecName();

    if( strcmp( v_codec, "jpeg" ) == 0 || strcmp( v_codec, "mjpa" ) == 0 ) {
        quicktime_set_jpeg( Datei, q, USE_FLOAT );
        return true;
    }

    return false;
}








/***** Hilfsfunktionen **********************************************/
int QuicktimeV::qt_c_model( int cmodel )
{
    int c_model = 0;

    switch( cmodel ) {
        case YUV420P:
            c_model = BC_YUV420P;
            break;

        case YUV422P:
            c_model = BC_YUV422P;
            break;
        case YUV422:
            c_model = BC_YUV422;
            break;
        case RGB16:
            c_model = BC_RGB565;
            break;
        case RGB16 | BGR:
            c_model = BC_BGR565;
            break;
        case RGB24:
            c_model = BC_RGB888;
            break;
        case RGB24 | BGR:
            c_model = BC_BGR888;
            break;
        case RGB32:
            c_model = BC_RGBA8888;
            break;
        case RGB32 | BGR:
            c_model = BC_BGR8888;
            break;
        case RGB15:
        case YUV420:
        default:
            //file->Fehler( "qt_c_model: Farbmodel %i nicht untersttzt\n", color_model_intern );
            c_model = -1;
    }

    return c_model;
}




void QuicktimeV::erz_row_pointers( unsigned char **row_pointers, unsigned char *buf )
{
    switch( color_model ) {
    case YUV420P:
        row_pointers[0] = buf;                                           // y - Plane
        row_pointers[1] = row_pointers[0] + breite * hoehe;              // u - Plane
        row_pointers[2] = row_pointers[1] + breite/2 * ((hoehe + 1)/2) ; // v - Plane
        break;
    case YUV422P:
        row_pointers[0] = buf;                                           // y - Plane
        row_pointers[1] = row_pointers[0] + breite * hoehe;              // u - Plane
        row_pointers[2] = row_pointers[1] + breite/2 * hoehe;            // v - Plane
        break;
    case RGB24:
    case RGB24|BGR:
        for( int i=0; i<hoehe; i++ ) {
            row_pointers[i] = buf + i * 3 * breite;
        }
        break;
    default:
        ERROR_TXT( _("erz_row_pointers: Colormodel %i not implemented!"), color_model );
        return;
    }
}











/*****************************************************************************/
/***** Audiotrack - Funktionen ***********************************************/
/*****************************************************************************/
QuicktimeAtr::QuicktimeAtr( QuicktimeIO *File, int Sample_rate, int Channels, const char *Codec_name )
:Track(File, 0, Codec_name), A_Track( File, Sample_rate, Channels, Codec_name )
{
    file = File;
    Datei = file->Datei;

    if( CodecName() == NULL ) {
        ERROR_TXT( _("Audio codec not implemented!") );
    }

    char codecname[5];
    strncpy( codecname, CodecName().chars(), 4 );
    quicktime_set_audio( Datei, Channels, SampleRate(), 16, codecname );

    if( !quicktime_supported_audio( Datei, 0 ) ) {
        ERROR_TXT( _("Audio codec not implemented!") );
    }

    // Codec-Objekt erstellen
    codec = new QuicktimeA( this );
}




QuicktimeAtr::QuicktimeAtr( QuicktimeIO *File, int track )
:Track(File, track), A_Track( File, track )
{
    file = File;
    Datei = file->Datei;

    // Informationen au Datei auslesen
    SetSampleRate( quicktime_sample_rate( Datei, Number() ) );
    total_samples = quicktime_audio_length( Datei, Number() );
    //bits = quicktime_audio_bits( Datei, nummer );
    SetChannels( quicktime_track_channels( Datei, Number() ) );

    char *c_name;
    c_name = quicktime_audio_compressor( Datei, Number() );
    char codecname[5];
    strncpy( codecname, c_name, 4 );
    codecname[4] = '\0';
    SetCodecName( codecname );

    // Codec erstellen
    if( !quicktime_supported_audio( Datei, 0 ) ) {
        ERROR_TXT( _("Audio codec not implemented!") );
    }

    // Codec-Objekt erstellen
    codec = new QuicktimeA( this );

}





QuicktimeAtr::~QuicktimeAtr()
{
    delete codec;
    codec = NULL;
}








/*****************************************************************************/
/***** Audiofunktionen *******************************************************/
/*****************************************************************************/



// Audio codec initialisieren
QuicktimeA::QuicktimeA( QuicktimeAtr *Track )
:A_Codec( Track )
{
    Datei = Track->Datei;
    sample_pos = 0;
    decode_puffer = NULL;
    decode_puffer_gr = 0;
}

QuicktimeA::~QuicktimeA()
{
    delete [] decode_puffer;
}




// Audiodaten schreiben
int QuicktimeA::encode( float **daten, int sampels )
{
    if( quicktime_encode_audio( Datei, NULL, daten, sampels ) )
        return 0;

    return 1;
}




// Audiodaten lesen
int QuicktimeA::decode( float **daten, int sampels )
{
    int i;
    
    if( decode_puffer_gr < sampels ) {
        delete [] decode_puffer;
        decode_puffer_gr = sampels;
        decode_puffer = new int16_t[decode_puffer_gr];
    }

    for( int j=0; j<Channels(); j++ ) {
       
        if( !SeekSample( sample_pos ) ){
            return 0;
        }

        // Quicktime kann offensichtlich Audiodaten nicht richtig als float lesen
        // => lese Audiodaten als int16 und konvertiere selber
        
        //if( quicktime_decode_audio( Datei, NULL, daten[j], sampels, j ) )
        //    return 0;
        if( quicktime_decode_audio( Datei, decode_puffer, NULL, sampels, j ) )
            return 0;
            
        for( i=0; i<sampels; i++ ) {
            daten[j][i] = (float)decode_puffer[i] / (float)(1<<15);
        }
        
    }
    sample_pos += sampels;

    return 1;
}


// zu Sample springen
int QuicktimeA::SeekSample( long long sample )
{
    if( sample >= quicktime_audio_length( Datei, track->Number() ) ) {
        return 0;
    }

    if( !quicktime_set_audio_position( Datei, sample, track->Number() ) ) {
        sample_pos = sample;
        return 1;
    }

    return 0;
}
