/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

using namespace std;

#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <iostream>

#include "WAV.h"
#include "debug.h"
#include "VideoFile/codecplugins.h"
#include "VideoFile/translate.h"

#define WAVE_FORMAT_PCM  0x0001
#define FORMAT_MULAW     0x0101
#define IBM_FORMAT_ALAW  0x0102
#define IBM_FORMAT_ADPCM 0x0103





/***** Konstruktor / Destruktor *****/
WavIO::WavIO( const String datei, IO_Mode mode )
:VideoFile( datei, "Wav", mode )
{
    // Variablen initialisieren
    wav_datei = NULL;
    wav_istream = NULL;
    wav_ostream = NULL;

    // Datei ffnen
    if( iomode == read ) {
        //Fehler( "yuvStream: IO-Modus \"lesen\" nicht implementiert\n" );
        if( datei != "" ) {
            wav_datei = new fstream( datei.chars(), ios::in );
            wav_istream = wav_datei;
        } else {
            wav_istream = &cin;
        }
        // lese Header und prfe Dateityp
        if( !read_riff_header() ) {
            ErrorStatus() = Errors::FileType;
            throw( ErrorStatus() );
        }
    } else if( iomode == write ) {
        if( datei != "" ) {
            wav_datei = new fstream( datei.chars(), ios::out );
            wav_ostream = wav_datei;
        } else {
            wav_ostream = &cout;
        }

        if( wav_datei == NULL ) {
            ErrorStatus() = Errors::Open;
            throw( ErrorStatus() );
        }
        
        // schreibe Header
        if( write_riff_header() < 1 ) {
            ErrorStatus() = Errors::Write;
            throw( ErrorStatus() );
        }
    } else {
        ErrorStatus() = Errors::IO_Mode;
        throw( ErrorStatus() );
    }

}




WavIO::~WavIO()
{
    // Tracks schlieen
    deleteTracks();

    delete wav_datei;
}








/***** Audio-Track *********************************************/
// Audio codec initialisieren
// Rckgabewert:  Zeiger auf Audiotrack
A_Track *WavIO::openAudioTrack( int Sample_rate, int Channels, const char *Codec )
{
    if( iomode == write )
        return new WavAtr( this, Sample_rate, Channels, Codec );
    return NULL;
}



// Track ffnen
A_Track *WavIO::openAudioTrack( int track )
{
    if( iomode != read )
        return NULL;

    if( track != 0 ) {
        ERROR_TXT( _("Track %i not found!"), track );
        return NULL;
    }

    return new WavAtr( this );
}



/***** Riff-Header lesen und schreiben **************************************************/
int WavIO::write_riff_header()
{
    struct riff_header riff;

    memset( &riff, 0, sizeof( struct riff_header ));

    strncpy( (char *)riff.id, "RIFF", 4 );
    strncpy( (char *)riff.wave_id, "WAVE", 4 );

    wav_ostream->write( (const char*)&riff, sizeof(struct riff_header) );
    if( !wav_ostream->good() )
        return 0;

    return 1;
}








int WavIO::read_riff_header()
{
    struct riff_header riff;

    memset( &riff, 0, sizeof(struct riff_header) );

    wav_istream->read( (char*)&riff, sizeof(struct riff_header) );
    // prfe, ob genug gelesen wurde
    if( !wav_istream->good() )
        return 0;

    // Integritt der gelesenen Daten prfen
    if( strncmp( (char *)riff.id, "RIFF", 4 ) != 0 )
        return 0;
    if( strncmp( (char *)riff.wave_id, "WAVE", 4 ) != 0 )
        return 0;

    return 1;
}















/*****************************************************************************/
/***** Audiotrack-Funktionen *************************************************/
/*****************************************************************************/

WavAtr::WavAtr( WavIO *File, int Sample_rate, int Channels, const char *Codec )
:Track(File, 0, Codec), A_Track( File, Sample_rate, Channels, Codec )
{
    file = File;
    wav_info = new struct wav_header;

    puffer_gr = 0;
    puffer = NULL;

    if( CodecName() == "" ) {
        SetCodecName( "pcm" );
    }
    
    codec = AV_Codecs->CreateACodec( this, "Wav", CodecName() );

    if( codec == NULL ) {
        ERROR_TXT( _("Codec \"%s\" not implemented!"), CodecName().chars() );
        codec = new A_Codec( this );
    }

    format = strtol( CodecID().chars(), NULL, 16 );
    
    // Header schreiben
    if( !write_wav_header() ) {
        ERROR_TXT( _("Write error!") );
        exit( 1 );
    }
}





WavAtr::WavAtr( WavIO *File )
:Track(File), A_Track( File )
{
    file = File;
    wav_info = new struct wav_header;

    // Header einlesen
    read_wav_header();

    // block_rate = 5;

    puffer_gr = 0;
    puffer = NULL;

    char codecname[5];
    sprintf( codecname, "%04x", format );
    SetCodecName( codecname );

    codec = AV_Codecs->CreateACodec( this, "Wav", CodecName() );

    if( codec == NULL ) {
        ERROR_TXT( _("Format \"%04x\" not implemented!"), format );
        codec = new A_Codec( this );
    }
}





WavAtr::~WavAtr()
{
    // Wav Header mit neuen Informationen ber Lnge und Bitrate schreiben
    // geht nur fr Dateien, nicht fr Streams
    if( IOMode() == VideoFile::write && file->wav_datei != NULL ) {
        file->wav_datei->seekg( sizeof(WavIO::riff_header), ios::beg );
        if( !file->wav_datei->good() ) {
            ERROR_TXT( _("Seek failed!") );
        } else if( !write_wav_header() ) {
            ERROR_TXT( _("Write error!") );
        }
    }

    delete codec;
    codec = NULL;

    delete [] puffer;

    delete wav_info;
}




int WavAtr::Seek( long block )
{
    if( file->wav_datei == NULL || codec == NULL ) {
        return 0;
    }

    // Positioniere Datei
    int64_t offset = (int64_t)( codec->BitsPerSecond() / 8.0 * (double)block / Rate() );
    DEBUG_TXT( _("Seek to block %li at offset %li."), block, offset );

    file->wav_datei->seekg( offset + sizeof(struct wav_header) + sizeof(struct WavIO::riff_header), ios::beg );
    if( !file->wav_datei->good() )
        return 0;

    SetPosition( block );

    return 1;
}





// Rohdaten lesen und schreiben
int WavAtr::readRaw( unsigned char *&data, long &bytes )
{
    // einen Block einlesen
    bytes = (long)( codec->BitsPerSecond() / ( Rate() * 8.0 ) );
    if( puffer_gr < bytes ) {
        delete [] puffer;
        puffer_gr = bytes;
        puffer = new unsigned char[ puffer_gr ];
    }

    file->wav_istream->read( (char*)puffer, bytes );
    data = puffer;
    if( file->wav_istream->good() ) {
        return 1;
    }

    return 0;
}



int WavAtr::writeRaw( unsigned char *data, long bytes, int samp )
{
    file->wav_ostream->write( (char*)data, bytes );
    if( file->wav_ostream->good() ) {
        return 1;
    }

    return 0;
}








/***** Wav-Header lesen und schreiben ***************************************************/
int WavAtr::write_wav_header()
{
    memset( wav_info, 0, sizeof( struct wav_header ));

    //strncpy(wave.riff.id, "RIFF", 4);
    //strncpy(wave.riff.wave_id, "WAVE",4);
    strncpy( (char *)wav_info->format.id, "fmt ",4 );
    wav_info->format.len = sizeof(struct common_struct);

    // Datenformat
    int bits = (int)( codec->BitsPerSecond() / ( SampleRate() * Channels() ) );
    wav_info->common.wFormatTag = format;
    wav_info->common.wChannels = Channels();
    wav_info->common.dwSamplesPerSec = SampleRate();
    wav_info->common.dwAvgBytesPerSec = (int32_t)(codec->BitsPerSecond() / 8);
    wav_info->common.wBlockAlign = Channels()*bits/8;
    wav_info->common.wBitsPerSample = bits;

    strncpy( (char *)wav_info->data.id, "data",4);
    // Datenlnge erstmal maximal setzen (2GB)
    //  ntig, damit Datenstream von Programmen verarbeitet wird, die die
    //  Datenlnge prfen
    if( BytesWritten() == 0 || BytesWritten() > LONG_MAX ) {
        wav_info->data.len = LONG_MAX;
    } else {
        wav_info->data.len = (unsigned long)BytesWritten();
    }

    file->wav_ostream->write( (const char*)wav_info, sizeof(struct wav_header) );
    if( !file->wav_ostream->good() )
        return 0;

    return 1;
}








int WavAtr::read_wav_header()
{
    memset( wav_info, 0, sizeof(wav_info) );

    file->wav_istream->read( (char*)wav_info, sizeof(struct wav_header) );
    // prfe, ob genug gelesen wurde
    assert( file->wav_istream->good() );

    // Integritt der gelesenen Daten prfen
    assert( strncmp( (char *)wav_info->format.id, "fmt ", 4 ) == 0 );
    assert( wav_info->format.len == sizeof(struct common_struct) );

    // Datenformat
    format = wav_info->common.wFormatTag;
    SetChannels( wav_info->common.wChannels );
    SetSampleRate( wav_info->common.dwSamplesPerSec );
    // wav_info->common.wBlockAlign = channels*bits/8;
    // bits = wav_info->common.wBitsPerSample;

    assert( strncmp( (char *)wav_info->data.id, "data", 4 ) == 0 );

    total_samples = (long)( (u_int64_t)wav_info->data.len * SampleRate() ) / wav_info->common.dwAvgBytesPerSec;
    if( total_samples == 0 ) {
        total_samples = -1;
    }

    return 1;
}
