/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

using namespace std;

#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <iostream>

#include "mp3.h"
#include "debug.h"
#include "VideoFile/codecplugins.h"
#include "VideoFile/translate.h"

#define MAX_OFFSET 1024




/***** Konstruktor / Destruktor *****/
mp3IO::mp3IO( const String datei, IO_Mode mode )
:VideoFile( datei, "mp3", mode )
{
    // Variablen initialisieren
    mp3_datei = NULL;
    mp3_istream = NULL;
    mp3_ostream = NULL;

    //memset( mp3_block_start, 0, 3 );
    mp3_block_start = 0;
    framesize = 1152;
    layer = 3;
    

    // Datei ffnen
    if( iomode == read ) {
        //Fehler( "yuvStream: IO-Modus \"lesen\" nicht implementiert\n" );
        if( datei != "" ) {
            mp3_datei = new fstream( datei.chars(), ios::in );
            mp3_istream = mp3_datei;
        } else {
            mp3_istream = &cin;
        }

        if( mp3_istream == NULL ) {
            ErrorStatus() = Errors::Open;
            throw( ErrorStatus() );
        }

        // prfe Dateityp
        if( !sync_block_start() ) {
            ErrorStatus() = Errors::FileType;
            throw( ErrorStatus() );
        }

    } else if( iomode == write ) {
        if( datei != "" ) {
            mp3_datei = new fstream( datei.chars(), ios::out );
            mp3_ostream = mp3_datei;
        } else {
            mp3_ostream = &cout;
        }

        if( mp3_ostream == NULL ) {
            ErrorStatus() = Errors::Open;
            throw( ErrorStatus() );
        }

    } else {
        ErrorStatus() = Errors::IO_Mode;
        throw( ErrorStatus() );
    }

}




mp3IO::~mp3IO()
{
    // Tracks schlieen
    deleteTracks();

    delete mp3_datei;
}








/***** Audio-Track *********************************************/
// Audio codec initialisieren
// Rckgabewert:  Zeiger auf Audiotrack
A_Track *mp3IO::openAudioTrack( int Sample_rate, int Channels, const char *Codec )
{
    if( iomode == write )
        return new mp3Atr( this, Sample_rate, Channels, Codec );
    return NULL;
}



// Track ffnen
A_Track *mp3IO::openAudioTrack( int track )
{
    if( iomode != read )
        return NULL;

    if( track != 0 ) {
        ERROR_TXT( _("Track %i not found!"), track );
        return NULL;
    }

    return new mp3Atr( this );
}







/***** Anfang des Streams suchen **********************************************/
bool mp3IO::sync_block_start( void )
{
    unsigned char daten[4]={0,0,0,0};
    unsigned long head = 0;
    int i = 0;
    int offset = 1;

    // Header einlesen
    mp3_istream->read( (char*)daten, 4 );
    if( !mp3_istream->good() )
        return false;

    if( memcmp( daten, "RIFF", 4 ) == 0 )
        offset = MAX_OFFSET;

    if( memcmp( daten, "ID3", 3 ) == 0 )
        offset = MAX_OFFSET;
    
    while( i < offset ) {

        head  = (unsigned long)daten[0] << 24;
        head |= (unsigned long)daten[1] << 16;
        head |= (unsigned long)daten[2] << 8;
        head |= (unsigned long)daten[3];
        
        // DEBUG_TXT( "Head = %8x", head );

        // Header auswerten
        if( decode_header( head ) ) {
            //strncpy( mp3_block_start, (char *)daten, 2 );
            //mp3_block_start[2] = 0;
            mp3_block_start = head;
            break;
        }

        // kein gltiger Header gefunden -> weiterlesen
        daten[0] = daten[1];
        daten[1] = daten[2];
        daten[2] = daten[3];
        
        // neues Zeichen lesen
        mp3_istream->read( (char*)&(daten[3]), 1 );
        if( !mp3_istream->good() )
            return false;

        i++;
    }
    
    if( i >= offset )
        return false;

    // Startzeichen wieder in Stream zurckschieben
    mp3_istream->putback( daten[3] );
    mp3_istream->putback( daten[2] );
    mp3_istream->putback( daten[1] );
    mp3_istream->putback( daten[0] );

    return true;
}






/***** Header dekodieren ******************************************************/
int mp3IO::decode_header( unsigned long head )
{
    int lsf;

    if( (head & 0xffe00000) != 0xffe00000 )
        return 0;
        
    if( head & (1<<20) ) {
        lsf = (head & (1<<19)) ? 0x0 : 0x1;
    } else {
        lsf = 1;
    }

    if( ((head>>10)&0x3) == 0x3) {
        return 0;
    }

    framesize = (lsf == 0) ? 1152 : 576;
    layer = 4-((head>>17)&3);
    
    return 1;
}










/*****************************************************************************/
/***** Audiotrack-Funktionen *************************************************/
/*****************************************************************************/

mp3Atr::mp3Atr( mp3IO *File, int Sample_rate, int Channels, const char *Codec_name )
:Track(File, 0, Codec_name), A_Track( File, Sample_rate, Channels, Codec_name )
{
    file = File;

    // Puffer festlegen - maximale Gre 1152 Samples/Block * 320 kbps bei 8 kHz Mono
    // mit Faktor 2 Sicherheit
    puffer_gr = 1152 / 8000 * 320000 * 2 ;
    puffer = new unsigned char[puffer_gr];
    puffer_block_len = 0;
    puffer_bytes = 0;

    if( CodecName() == "" ) {
        SetCodecName( "mp3" );
    }
    
    codec = AV_Codecs->CreateACodec( this, "mp3", CodecName() );

    if( codec == NULL ) {
        ERROR_TXT( _("Audio codec %s not implemented!"), CodecName().chars() );
        codec = new A_Codec( this );
    }
}





mp3Atr::mp3Atr( mp3IO *File )
:Track(File), A_Track( File )
{
    file = File;

    // Puffer festlegen - maximale Gre 1152 Samples/Block * 320 kbps bei 8 kHz Mono
    // mit Faktor 2 Sicherheit
    puffer_gr = (long)(1152.0 / 8000 * 320000.0 / 8 * 2);
    puffer = new unsigned char[puffer_gr];
    puffer_block_len = 0;
    puffer_bytes = 0;

    SetCodecName( "mp3" );

    codec = AV_Codecs->CreateACodec( this, "mp3", CodecName() );

    if( codec == NULL ) {
        ERROR_TXT( _("Audio codec %s not implemented!"), CodecName().chars() );
        codec = new A_Codec( this );
    }

    // vom Codec gendertes Format bernehmen
    (A_Format)*this = (A_Format)*codec;
    //block_rate = channels * (double)sample_rate / (double)file->framesize;
    SetRate( (double)SampleRate() / (double)file->framesize );
}





mp3Atr::~mp3Atr()
{

    delete codec;
    codec = NULL;

    delete [] puffer;
}




int mp3Atr::Seek( long block )
{
    long bytes=0;
    unsigned char *buf=NULL;

    DEBUG_TXT( _("Current position: %li"), Position() );

    if( Position() > block ) {
        DEBUG_TXT( _("Seek to start.") );

        if( file->mp3_datei == NULL || codec == NULL ) {
            return A_Track::Seek( block );
        }
        file->mp3_datei->seekg( 0, ios::beg );
        //Anfang neu suchen
        file->sync_block_start();

        SetPosition( 0 );
        puffer_bytes = 0;
    }

    DEBUG_TXT( _("Seek to block %li."), block );

    while( Position() < block ) {
        if( ! readData( buf, bytes ) ) {
            return 0;
        }
    }

    return 1;
}





// Rohdaten lesen und schreiben
int mp3Atr::readRaw( unsigned char *&data, long &bytes )
{
    long begin;
    uint32_t head;
    
    data = puffer;
    bytes = 0;

    // Puffer einen Block weiterschieben
    puffer_schieben( puffer_block_len );
    
    // einen Block einlesen
    if( !puffer_fuellen() )
        return 0;


    // Suche Start-Header - prfe, ob Header mit altem Header bereinstimmt
    begin = puffer_finden( 0, file->mp3_block_start & 0xfffffd00, 0xfffffd00, head );
    // begin = puffer_finden( 0, 0xffe00000, 0xffe00000, head );

    if( begin < 0 ) {
        //file->Fehler( "readData: Keine MP3 Datei\n" );
        return 0;
    }

    
    // Header gefunden!
    if( begin > 0 ) {
        DEBUG_TXT( _("Resync stream %i bytes"), begin );
        puffer_schieben( begin );

        // Puffer wieder auffllen
        puffer_fuellen();
    }
    

    long ende;                                                   
    // ende = puffer_finden( 4, 0xffe00000, 0xffe00000, head );
    ende = puffer_finden( 4, file->mp3_block_start & 0xfffffd00, 0xfffffd00, head );

    if( ende >= 0 ) {
        puffer_block_len = ende;
        bytes = puffer_block_len;
    } else {
        puffer_block_len = puffer_bytes;
        bytes = puffer_bytes;
    }
    
    return 1;
}






int mp3Atr::writeRaw( unsigned char *data, long bytes, int samp )
{
    file->mp3_ostream->write( (char*)data, bytes );
    if( file->mp3_ostream->good() ) {
        return 1;
    }

    return 0;
}





void mp3Atr::puffer_schieben( long bytes )
{
    if( bytes >= puffer_bytes ) {
        puffer_bytes = 0;
        return;
    }

    for( int i=0; i < puffer_bytes - bytes; i++ ) {
        puffer[i] = puffer[i + bytes];
    }

    puffer_bytes -= bytes;
}








/***** Puffer-Verwaltung *****************************************************/

long mp3Atr::puffer_finden( size_t start, uint32_t muster, uint32_t maske, uint32_t &gefunden )
{

    long begin;
    uint32_t head = 0;

    if( puffer_bytes < 4 )
        return -1;

    head = puffer[start]<<16 | puffer[start+1]<<8 | puffer[start+2];
    
    for( begin=start; begin < puffer_bytes - 3; begin++ ) {
        head = (head<<8) | puffer[begin + 3];

        if( (head & maske) == muster ) {
            gefunden = head;
            return begin;
        }
    }
    
    if( begin >= puffer_bytes - 3 ) {
        begin = -1;
    }

    return begin;
}




int mp3Atr::puffer_fuellen( void )
{
    long read_bytes;

    read_bytes = puffer_gr - puffer_bytes;
    file->mp3_istream->read( (char*)&(puffer[puffer_bytes]), read_bytes );
    read_bytes = file->mp3_istream->gcount();
    puffer_bytes += read_bytes;
    if( !file->mp3_istream->good() ) {
        if( read_bytes == 0 ) {
            return 0;
        }
    }

    return 1;
}

