/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>

#include "MpegVideo.h"
#include "debug.h"
#include "VideoFile/codecplugins.h"
#include "VideoFile/translate.h"


#define MAX_OFFSET 100

// Puffer festlegen - maximale Gre: ein umkompremiertes Bild mit 800*600 Punkten YUV420
#define PUFFER_GR 800*900

/* picture coding type */
#define I_TYPE 1
#define P_TYPE 2
#define B_TYPE 3
#define D_TYPE 4


const double MpegVideoIO::mpeg_framerates[] = {
    00.000, 23.976,
    24.000,                      /* 24fps movie */
    25.000,                      /* PAL */
    29.970,                      /* NTSC */
    30.000, 50.000,
    59.940, 60.000,
    -1
};



/***** Konstruktor / Destruktor *****/
MpegVideoIO::MpegVideoIO( const String datei, IO_Mode mode )
:VideoFile( datei, "MpegVideo", mode )
{
    // Variablen initialisieren
    mpeg_datei = NULL;
    mpeg_istream = NULL;
    mpeg_ostream = NULL;
    puffer = NULL;
    chunk = NULL;
    sequence_header = NULL;
    sequence_header_gr = 0;

    framerate = 25.0;
    version = 1;

    // Datei ffnen
    if( iomode == read ) {
        if( datei != "" ) {
            mpeg_datei = new fstream( datei.chars(), ios::in );
            mpeg_istream = mpeg_datei;
        } else {
            mpeg_istream = &cin;
        }

        if( mpeg_istream == NULL ) {
            ErrorStatus() = Errors::Open;
            throw( ErrorStatus() );
        }

        puffer = new FifoPuffer( mpeg_istream, PUFFER_GR );
        chunk = new Chunk( puffer );
    
        // prfe Dateityp
        if( !sync_block_start() ) {
            ErrorStatus() = Errors::FileType;
            throw( ErrorStatus() );
        }

    } else if( iomode == write ) {
        if( datei != "" ) {
            mpeg_datei = new fstream( datei.chars(), ios::out );
            mpeg_ostream = mpeg_datei;
        } else {
            mpeg_ostream = &cout;
        }

        if( mpeg_ostream == NULL ) {
            ErrorStatus() = Errors::Open;
            throw( ErrorStatus() );
        }

    } else {
        ErrorStatus() = Errors::IO_Mode;
        throw( ErrorStatus() );
    }

}




MpegVideoIO::~MpegVideoIO()
{
    // Tracks schlieen
    deleteTracks();

    delete mpeg_datei;

    delete puffer;

    delete [] sequence_header;
}








/***** Video-Track *********************************************/
// Video codec initialisieren
// Rckgabewert:  Zeiger auf Videotrack
V_Track *MpegVideoIO::openVideoTrack( double Rate, int x, int y, const char *Codec )
{
    if( iomode == write ) {
        framerate = Rate;
        width = x;
        height = y;
        return new MpegVideoVtr( this, Rate, x, y, Codec );
    }
    return NULL;
}



// Track ffnen
V_Track *MpegVideoIO::openVideoTrack( int track )
{
    if( iomode != read )
        return NULL;

    if( track != 0 ) {
        ERROR_TXT( _("Track %i not found!"), track );
        return NULL;
    }

    return new MpegVideoVtr( this );
}







/***** Anfang des Streams suchen **********************************************/
bool MpegVideoIO::sync_block_start( void )
{
    int i = 0;
    int code;

    while( i < MAX_OFFSET ) {

        code = chunk->Startcode();

        if( code < 0 || code >= SYSTEM_HEADER_CODE )
            return false;

        if( code == SEQUENCE_HEADER_CODE )
            break;
                    
        i++;
    }

    if( i >= MAX_OFFSET )
        return false;

    // Header auswerten
    if( sequence_header_gr == 0 ) {
        sequence_header_gr = chunk->Bytes();
        sequence_header = new uint8_t[ sequence_header_gr ];
        chunk->Lesen( sequence_header, sequence_header_gr );
    
        if( !decode_header( sequence_header, sequence_header_gr ) ) {
            return false;
        }
    }
    
    return true;
}






/***** Header dekodieren ******************************************************/
int MpegVideoIO::decode_header( unsigned char *head, int bytes )
{
    int frame_rate_code = 0;
    
    assert( bytes > 10 );
    
    // prfe, ob der Chunk ein Sequence-Header ist
    //if( memcmp( head, SEQUENCE_HEADER, 4 ) != 0 )
    //    return 0;

    // Prfe Gltigkeit des Headers
    if ((head[10] & 0x20) != 0x20)
        return 0;	/* missing marker_bit */

    // Lese Informationen von Videotrack
    // Bildbreite / -hhe
    height = (head[4] << 16) | (head[5] << 8) | head[6];

    width = ((height >> 12) + 15) & ~15;
    height = ((height & 0xfff) + 15) & ~15;

    if ((width > 768) || (height > 576))
        return 0;	/* size restrictions for MP@ML or MPEG1 */

    /* this is not used by the decoder */
    aspect_ratio_code = head[7] >> 4;
    frame_rate_code = head[7] & 15;
    bitrate = (head[8]<<10)|(head[9]<<2)|(head[10]>>6);

    framerate = decode_framerate( frame_rate_code );

    return 1;
}







/***** Framerate-Code in Framerate umwandeln *********************************/
double MpegVideoIO::decode_framerate( int code )
{
    for( int i=0; i<=code; i++ ) {
        if( mpeg_framerates[i] < 0 )
            return 0.0;
    }
    
    return mpeg_framerates[code];
}
















/*****************************************************************************/
/***** Videotrack-Funktionen *************************************************/
/*****************************************************************************/

MpegVideoVtr::MpegVideoVtr( MpegVideoIO *File, double Rate, int x, int y, const char *Codec_name )
:Track(File, 0, Codec_name), V_Track( File, Rate, x, y, Codec_name )
{
    file = File;
    data_puffer_gr = 0;
    data_puffer = NULL;
    max_pos = 0;
    sequence_header_gelesen = false;
    
    if( CodecName() == "" ) {
        SetCodecName( "mpeg1" );
    }
    
    codec = AV_Codecs->CreateVCodec( this, "MpegVideo", CodecName() );

    if( codec == NULL ) {
        ERROR_TXT( _("Video codec \"%s\" not implemented!"), CodecName().chars() );
        codec = new V_Codec( this );
    }
}





MpegVideoVtr::MpegVideoVtr( MpegVideoIO *File )
:Track(File), V_Track( File )
{
    file = File;
    data_puffer_gr = PUFFER_GR;
    data_puffer = new uint8_t[data_puffer_gr];
    max_pos = 0;
    sequence_header_gelesen = false;
    
    SetCodecName( "mpeg" );

    breite = file->width;
    hoehe = file->height;
    SetRate( file->framerate );

    codec = AV_Codecs->CreateVCodec( this, "MpegVideo", CodecName() );

    if( codec == NULL ) {
        ERROR_TXT( _("Video codec \"%s\" not implemented!"), CodecName().chars() );
        codec = new V_Codec( this );
    }

    // vom Codec gendertes Format bernehmen
    (V_Format)*this = (V_Format)*codec;
}





MpegVideoVtr::~MpegVideoVtr()
{

    delete codec;
    codec = NULL;
}




int MpegVideoVtr::Seek( long block )
{
    //long bytes=0;
    //unsigned char *buf=NULL;
    //bool header_gelesen = sequence_header_gelesen;
    DEBUG_TXT( _("Seek to block %i"), block );
    
    if( Position() > block ) {
        if( file->mpeg_datei == NULL || codec == NULL ) {
            return V_Track::Seek( block );
        }

        //suche nach einem Eintrag in der Keytable
        long long offset = 0;
        long key = key_table.GetKeyFrame( block, offset );
        
        file->mpeg_datei->seekg( offset, ios::beg );

        file->puffer->Reset();
        
        if( offset == 0 ) {
            //Anfang neu suchen
            file->sync_block_start();
        }

        SetPosition( key );
    }

    while( Position() < block ) {
        if( !skip_frame() ) {
            return 0;
        }
    }

    //sequence_header_gelesen = header_gelesen;
    sequence_header_gelesen = false;

    return 1;
}






/**liefert den letzten Keyframe zur angegebenen Position
  */
long MpegVideoVtr::GetKeyFrame( long bild )
{
    long long offset = 0;
    long key = 0;

    if( max_pos >= bild ) {
        key = key_table.GetKeyFrame( bild, offset );

        //DEBUG_TXT( "Keyframe %i fr Bild %i", key, bild );
        return key;
    }

    return bild;
}






// Rohdaten lesen und schreiben
int MpegVideoVtr::readRaw( unsigned char *&data, long &bytes )
{
    size_t offset = 0;
    size_t n = 0;
    int code = 0;
    
    // einen Block einlesen
    
    // Ist schonmal ein Sequenceheader gelesen worden?
    if( !sequence_header_gelesen ) {
        DEBUG_TXT( _("Sequence header not read.") );
        // enthlt der aktuelle Block schon einen Srequence Header?
        if( file->chunk->Startcode() != SEQUENCE_HEADER_CODE ) {
            DEBUG_TXT( _("Copy sequence header.") );
            
            memcpy( data_puffer, file->sequence_header, file->sequence_header_gr );
            offset += file->sequence_header_gr;
        }
        sequence_header_gelesen = true;
    }

    // kopiere alles bis zum ersten Bild
    while( (code = file->chunk->Startcode()) != PICTURE_START_CODE && code >= 0 ) {
        n = data_puffer_gr - offset;
        if( file->chunk->Lesen( data_puffer + offset, n ) < 1 )
            return 0;
        offset += n;
    }

    if( code < 0 )
        return 0;

    n = data_puffer_gr - offset;
    if( file->chunk->Lesen( data_puffer + offset, n ) < 1 )
        return 0;
    offset += n;

    
    // kopiere alles bis zum Ende des Bildes
    while( (code = file->chunk->Startcode()) != PICTURE_START_CODE && code >= 0 ) {
        n = data_puffer_gr - offset;
        if( file->chunk->Lesen( data_puffer + offset, n ) < 1 ) {
            bytes = offset;
            data = data_puffer;
            return 1;
        }
        offset += n;
    }

    bytes = offset;
    data = data_puffer;
    
    if( code < 0 )
        return 1;

                                
    // prfe, ob nchstes Bild ein I-Frame ist
    if( test_keyframe() ) {
        long long file_offset = 0;
        if( file->mpeg_datei != NULL ) {
            file_offset = file->mpeg_datei->tellg();
        }

        // Fge I-Frame in Key-Table ein
        key_table.AddKeyFrame( Position()+1, file_offset - file->puffer->Bytes() );
    }

    // merke die maximale Position, fr die schon Key-Frame informationen existieren
    max_pos = (Position()+1 > max_pos)? Position()+1 : max_pos;

    return 1;
}






int MpegVideoVtr::writeRaw( unsigned char *data, long bytes, int keyframe )
{
    file->mpeg_ostream->write( (char*)data, bytes );
    if( file->mpeg_ostream->good() ) {
        return 1;
    }

    return 0;
}







/***** Prfe, ob Frame ein I-Frame ist ***************************************/
bool MpegVideoVtr::test_keyframe(void)
{
    if( file->chunk->Startcode() != PICTURE_START_CODE )
        return false;

    uint8_t data = file->chunk->Data(5);       
    int coding_type = (data >> 3) & 7;

    return (coding_type == I_TYPE);
}





/***** Ein Bild berspringen *************************************************/
int MpegVideoVtr::skip_frame( void )
{
    int code = 0;

    // einen Block einlesen
    
    // berspringe alles bis zum ersten Bild
    while( (code = file->chunk->Startcode()) != PICTURE_START_CODE && code >= 0 ) {
        if( file->chunk->Skip() < 1)
            return 0;
    }

    if( code < 0 )
        return 0;

    if( file->chunk->Skip() < 1 )
        return 0;

    // berspringe alles bis zum Ende des Bildes
    while( (code = file->chunk->Startcode()) != PICTURE_START_CODE && code >= 0 ) {
        if( file->chunk->Skip() < 1 )
            return 1;
    }

    if( code < 0 )
        return 1;

    // prfe, ob nchstes Bild ein I-Frame ist
    if( test_keyframe() ) {
        long long offset = 0;
        if( file->mpeg_datei != NULL ) {
            offset = file->mpeg_datei->tellg();
        }

        // Fge I-Frame in Key-Table ein
        key_table.AddKeyFrame( Position()+1, offset - file->puffer->Bytes() );
    }

    // merke die maximale Position, fr die schon Key-Frame informationen existieren
    max_pos = (Position()+1 > max_pos)? Position()+1 : max_pos;

    SetPosition( Position() + 1 );

    return 1;
}
