/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <fcntl.h>

#include "yuvStream.h"
#include "debug.h"
#include "VideoFile/translate.h"




/***** Konstruktor / Destruktor *****/
yuvStream::yuvStream( const String file, IO_Mode mode )
:VideoFile( file, "yuv4mpeg", mode )
{
    // Variablen initialisieren
    fileno = -1;
    
    // Datei ffnen
    if( iomode == read ) {
        if( file != "" ) {
            fileno = open( file.chars(), O_RDONLY );
            if( fileno < 0 ) {
                ErrorStatus() = Errors::Open;
                throw( ErrorStatus() );
            }
        } else {
            fileno = STDIN_FILENO;
        }

        // Header einlesen
        if( yuv_read_header ( fileno, &breite, &hoehe, &rate_code) != 0 ) {
            ErrorStatus() = Errors::FileType;
            throw( ErrorStatus() );
        }
        

    } else if( iomode == write ) {
        if( file != "" ) {
            fileno = open( file.chars(), O_WRONLY | O_CREAT | O_TRUNC );
            if( fileno < 0 ) {
                ErrorStatus() = Errors::Open;
                throw( ErrorStatus() );
            }
        } else {
            fileno = STDOUT_FILENO;
        }
    } else {
        ErrorStatus() = Errors::IO_Mode;
        throw( ErrorStatus() );
    }
}




yuvStream::~yuvStream()
{
    // Tracks schlieen
    deleteTracks();

    if( fileno > STDIN_FILENO && fileno > STDOUT_FILENO ) {
        close( fileno );
    }
}








/***** Video-Funktionen *********************************************/
// Video codec initialisieren
// Rckgabewert:  Zeiger auf Videotrack
V_Track *yuvStream::openVideoTrack( double Rate, int x, int y, const char *Codec )
{
    if( iomode != write )
        return NULL;

    return new yuvStreamVtr( this, Rate, x, y );
}

// Track ffnen
V_Track *yuvStream::openVideoTrack( int track )
{
    if( iomode != read )
        return NULL;

    if( track != 0 ) {
        ERROR_TXT( _("Track %i not found!"), track );
        return NULL;
    }

    return new yuvStreamVtr( this );
}






/*****************************************************************************/
/***** Videotrack-Funktionen *************************************************/
/*****************************************************************************/

yuvStreamVtr::yuvStreamVtr( yuvStream *File, float rate, int x, int y )
:Track(File), V_Track( File, rate, x, y )
{
    file = File;

    puffer = new unsigned char[ breite * hoehe * 2 ];
    // Header schreiben
    yuv_write_header( file->fileno, breite, hoehe, yuv_fps2mpegcode( rate ) );

    codec = new yuvStreamV( this );
}




yuvStreamVtr::yuvStreamVtr( yuvStream *File )
:Track(File), V_Track( File )
{
    int rate_code;

    file = File;

    breite = file->breite;
    hoehe = file->hoehe;
    rate_code = file->rate_code;
    
    SetRate( (float)yuv_mpegcode2fps( (unsigned int)rate_code ) );

    puffer = new unsigned char[ breite * hoehe * 2 ];

    codec = new yuvStreamV( this );
}




yuvStreamVtr::~yuvStreamVtr()
{
    delete [] puffer;
}





// Bild-Position setzen
// nur in Vorwrtsrichtung mglich
int yuvStreamVtr::Seek( long block )
{
    long pos=Position(), bytes=0;
    unsigned char *buf=NULL;

    if( pos > block ) {
        return V_Track::Seek( block );
    }

    while( Position() < block ) {
        if( ! readData( buf, bytes ) ) {
            return 0;
        }
    }

    return 1;
}





// Rohdaten lesen und schreiben
int yuvStreamVtr::readRaw( unsigned char *&data, long &bytes )
{
    unsigned char *yuv[3];

    bytes = 0;

    /***** Zeiger auf Planes erstellen *****/
    yuv[0] = puffer;                               // y - Plane
    yuv[1] = yuv[0] + breite * hoehe;              // u - Plane
    yuv[2] = yuv[1] + breite/2 * ((hoehe + 1)/2) ; // v - Plane

    if( !yuv_read_frame( file->fileno, yuv, breite, hoehe ) ) {
        return 0;
    }
    bytes = breite * hoehe + (breite * hoehe + 1) / 2;

    data = puffer;
    return 1;
}



int yuvStreamVtr::writeRaw( unsigned char *data, long bytes, int keyframe )
{
    unsigned char *yuv[3];

    /***** Zeiger auf Planes erstellen *****/
    yuv[0] = data;                                 // y - Plane
    yuv[1] = yuv[0] + breite * hoehe;              // u - Plane
    yuv[2] = yuv[1] + breite/2 * ((hoehe + 1)/2) ; // v - Plane

    yuv_write_frame( file->fileno, yuv, breite, hoehe );

    return 1;
}








/*****************************************************************************/
/***** Videofunktionen *******************************************************/
/*****************************************************************************/

yuvStreamV::yuvStreamV( yuvStreamVtr *Track )
:V_Codec( Track )
{
    ColorModel( YUV420P );
}



yuvStreamV::~yuvStreamV()
{
}




// Farbraum festlegen
bool yuvStreamV::ColorModel( int cmodel )
{
    // verwende intern Standardfarbraum und benutze Konvertierung
    if( cmodel != YUV420P ) {
        V_Codec::ColorModel( YUV420P );
        return false;
    }

    return V_Codec::ColorModel( cmodel );
}






// Video-Bild schreiben
int yuvStreamV::encode( VideoFrame *bild )
{
    uint8_t *bild_buf = NULL;

    if( bild->GetFrame( bild_buf ) <= 0 )
        return 0;

    track->writeData( bild_buf, breite*hoehe + breite * ((hoehe + 1)/2) );
    return 1;
}



// Bild lesen
int yuvStreamV::decode( VideoFrame *bild )
{
    uint8_t *puffer=NULL;
    long bytes;
    int stat;
    //uint8_t *bild_buf = NULL;

    /*
    if( bild->FramePointer( bild_buf ) <= 0 )
        return 0;
    */
    
    stat = track->readData( puffer, bytes );

    if( !stat ) {
        return stat;
    }

    // memcpy( bild_buf, puffer, bytes );
    bild->SetFrame( puffer );

    return stat;
}
