/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <stdio.h>
#include <assert.h>


#include "ffmpegAudio.h"
#include "debug.h"
#include "VideoFile/translate.h"
#include "plugins/RIFFio/Wave_Format.h"





/***** Konstruktor / Destruktor *****/
ffmpegAudio::ffmpegAudio( A_Track *Track )
:A_Codec( Track )
{
    // Puffer initialisieren
    puffer = NULL;
    puffer_gr = 0;
    sample_puffer = NULL;
    frame_size = 0;
    sample_rest = 0;
    sample_pos = 0;
    data = NULL;
    data_pos = 0;
    data_rest = 0;
    format_info = NULL;

    // Codec-Context initialisieren
    memset( &context, 0, sizeof(AVCodecContext) );

    init = false;

    enum CodecID cid = CODEC_ID_NONE;

    // Optionen
    bitrate = 128;

    // Compressor / Dekompressor initialisieren
    if( CodecName() == "mp2" ) {
        cid = CODEC_ID_MP2;
    } else if ( CodecName() == "ac3" ) {
        cid = CODEC_ID_AC3;
    } else if ( CodecName() == "alaw" ) {
        cid = CODEC_ID_PCM_ALAW;
    } else if ( CodecName() == "MULAW" ) {
        cid = CODEC_ID_PCM_MULAW;
    } else if ( CodecName() == "ADPCM-MS" ) {
        cid = CODEC_ID_ADPCM_MS;
    } else if ( CodecName() == "ADPCM-IMA" ) {
        cid = CODEC_ID_ADPCM_IMA_WAV;
    } else if ( CodecName() == "WMA1" ) {
        cid = CODEC_ID_WMAV1;
    } else if ( CodecName() == "WMA2" ) {
        cid = CODEC_ID_WMAV2;
    }  else {
        ERROR_TXT( _("Codec %s not implemented!"), CodecName().chars() );
        cid = CODEC_ID_NONE;
    }



    if( iomode == VideoFile::read )
        avcodec = avcodec_find_decoder( cid );
    else
        avcodec = avcodec_find_encoder( cid );

    assert( avcodec != NULL );
    
    if( CodecName() == "mp2" || CodecName() == "ac3" ) {
        AddOption( "Bitrate", _("average bitrate in kbps") ) = bitrate;
    }

    // Ersten Block einlesen, um Daten vom Codec zu bekommen
    if( iomode==VideoFile::read ) {
        decode( NULL, 0 );
    }
}




ffmpegAudio::~ffmpegAudio()
{
    if( iomode == VideoFile::write && sample_rest > 0 ) {
        // Restliche Samples im Puffer kompremieren
        for( int i=sample_rest * Channels(); i < frame_size * Channels(); i++ ) {
            sample_puffer[i] = 0;
        }

        long bytes = avcodec_encode_audio( &context, puffer, puffer_gr, sample_puffer );
        track->writeData( puffer, bytes, sample_rest );
    }

    if( init )
        avcodec_close( &context );

    delete [] puffer;
    delete [] sample_puffer;
    delete [] (uint8_t*)context.extradata;
    delete [] format_info;
}






// Codec ffen und Optionen setzen
void ffmpegAudio::init_codec( void )
{
    avcodec_get_context_defaults( &context );
    
    
    
    // context.sub_id;    /* some codecs needs additionnal format info. It is
    //                       stored there */

    // audio only
    context.sample_rate = SampleRate();

    // ffmpeg kann nicht mit mehr als 2 Kanlen umgehen
    if( Channels() > 2 )
        SetChannels( 2 );
        
    context.channels = Channels();

    
    
    // Codecparameter vom Track bernehmen
    if( IOMode() == VideoFile::read ) {
        context.block_align = FrameBytes();
        context.bit_rate = BitsPerSecond();
    } else {
        context.bit_rate = bitrate * 1000;
    }

    
    // Zustzliche Daten vom IO-Modul
    if( IOMode() == VideoFile::read && file->ModulName() == "Avi" ) {
        uint8_t *data = NULL;
        size_t size = 0;
        WaveFormat w_format;
        
        if( track->CodecExtraInfo( data, size ) > 0 && w_format.Read(data, size) ) {
            //context.block_align = w_format.BlockAlign;
            //context.bit_rate = w_format.AvgBytesPerSec * 8;
            
            if( w_format.DataSize > 0 ) {
                context.extradata_size = w_format.DataSize;
                context.extradata = new uint8_t[context.extradata_size];
                memcpy( context.extradata, w_format.Data, context.extradata_size );
            }
        }
    }


    if( avcodec_open( &context, avcodec ) < 0 ) {
        ERROR_TXT( _("Error opening the audio-codec!") );
        return;
    }

    frame_size = context.frame_size;

    // frame_size gibt nicht zuverlssig an, wieviele samples vom Codec geliefert werden
    // auch AVCODEC_MAX_AUDIO_FRAME_SIZE ist keine sichere Begrenzung, aber der Wert
    // sollte gro genug sein, damit die Puffergre ausreicht
    sample_puffer = new short[ AVCODEC_MAX_AUDIO_FRAME_SIZE ];
    
    // Puffer anfordern
    // Gre fr ein Datenpaket (unkopremierte Daten)
    delete [] puffer;
    puffer_gr = AVCODEC_MAX_AUDIO_FRAME_SIZE * Channels() * 2;
    puffer = new uint8_t[puffer_gr];

    
    
    // Codecoptionen in Audio-Formatbeschreibung bernehmen
    
    if( CodecName() == "mp2" || CodecName() == "ac3" ) {
        SetBitsPerSecond( context.bit_rate );
        SetFrameSamples( context.frame_size );
    } else {
        SetBitsPerSecond( (double)(context.block_align * context.sample_rate * 8.0) / context.frame_size  );
        SetFrameSamples( context.frame_size );
    }
    
    init = true;
}






// Audiodaten schreiben
int ffmpegAudio::encode( float **daten, int samples )
{
    if( !init )
        init_codec();

    if( !init )
        return 0;
            
    int i,j, samples_write=0, samples_read=0, stat = 0;
    long bytes = 0;
    int16_t sample;


    // Audio-Frames verarbeiten
    bytes = 0;
    while( samples + sample_rest >= frame_size ) {
        // Einen Frame kopieren
        for( i=0; i < frame_size - sample_rest; i++ ) {
            for( j=0; j<Channels(); j++ ) {           
                sample = (int16_t)( daten[j][samples_read+i] * (1<<15) * 0.8 );
                sample_puffer[ (sample_rest + i)*Channels() +j ] = sample;
            }
        }

        // Frame kompremieren
        bytes = avcodec_encode_audio( &context, puffer, puffer_gr, sample_puffer );

        sample_rest = 0;
        samples -= i;
        samples_read += i;
        samples_write += frame_size;
        
        assert( bytes > 0);
            
        // Puffer in Datei schreiben
        if( !(stat = track->writeData( puffer, bytes, frame_size )) )
            return stat;
    }

    // restliche Samples speichern
    assert( samples + sample_rest < frame_size );
    if( samples > 0 ) {
        for( i=0; i < samples; i++ ) {
            for( j=0; j<Channels(); j++ ) {
                sample = (int16_t)( daten[j][samples_read+i] * (1<<15) );
                sample_puffer[ (sample_rest + i) * Channels() + j ] = sample;
             }
        }
        sample_rest += i;
    }

}






// Audiodaten lesen
int ffmpegAudio::decode( float **daten, int samples )
{
    if( !init )
        init_codec();

    if( !init )
        return 0;
        
    int i=0, j, stat=0;
    int frame_size_ptr = frame_size;

    do {
        // noch alte Daten im Puffer?
        if( sample_rest == 0 ) {
            // einen Frame lesen und dekodieren


            // Puffer aus Datei lesen
            if( context.codec_id == CODEC_ID_WMAV1 || context.codec_id == CODEC_ID_WMAV2 ) {
                if( data_rest != 0 ) {
                    ERROR_TXT( _("WMA-Codec read no data") );
                    //return 0;
                }
                
                // der WMA-Codec ignoriert offensichtlich vllig die angegebene
                // Puffergre und liest IMMER "block_align" Bytes
                // das entspricht so nicht dem "dokumentierten" Verhlaten und wird
                // hoffentlich in einer spteren Version von ffmpeg behoben
                
                if( puffer_gr < context.block_align ) {
                    delete [] puffer;
                    puffer_gr = context.block_align;
                    puffer = new uint8_t[puffer_gr];
                }

                data_rest = context.block_align;
                data_pos = 0;
                data = puffer;
                
                if( !track->readBytes( puffer, data_rest ) )
                    return 0;
                
            } else {
                
                if( data_rest <= 0 ) {           
                    if( !track->readData( data, data_rest ) )
                        return 0;
                    data_pos = 0;
                }
            }

            
            // Frame decodieren
            stat = avcodec_decode_audio( &context, sample_puffer, &frame_size_ptr,
                                         data + data_pos, data_rest );

            if( stat < 0 ) {
                return 0;
            }

            data_rest -= stat;
            data_pos += stat;

            SetChannels( context.channels );
            SetSampleRate( context.sample_rate );

            sample_rest = frame_size_ptr / (2 * Channels());
            sample_pos = 0;
        }


        // Kanle in Ausgabe kopieren
        while( i < samples && sample_rest > 0 && daten != NULL ) {
            for( j=0; j<Channels(); j++ ) {
                daten[j][i] = (float)sample_puffer[ sample_pos * Channels() + j ] / (float)(1<<15);
            }
            sample_rest--;
            sample_pos++;
            i++;
        }
    } while( i < samples );

    return 1;

}











int ffmpegAudio::SeekSample( int64_t sample )
{
    if( iomode != VideoFile::read )
        return 0;
    
    sample_rest = 0;
    sample_pos = 0;
    data_rest = 0;
    data_pos = 0;

    return A_Codec::SeekSample( sample );
}





// Optionen setzen
int ffmpegAudio::SetOption( OptionEntry *opt )
{

    if( opt->Name() == "Bitrate" ) {
        if( (int)*opt < 2 )
            *opt = 2;
        else
            bitrate = (int)*opt;

    } else {
        return 0;
    }

    return 1;
}




int ffmpegAudio::GetExtraInfo( uint8_t *&info, unsigned int &size )
{
    if( file->ModulName() == "Avi" ) {
        size_t n = size;
        WaveFormat w_format;

        w_format.Read( info, n );
        
        //w_format.BlockAlign = context.block_align;
        
        if( context.extradata_size > 0 ) {
            w_format.DataSize = context.extradata_size;
            
            if( w_format.DataSize > 0 )
                delete [] w_format.Data;

            w_format.Data = new uint8_t[context.extradata_size];
            memcpy( w_format.Data, context.extradata, context.extradata_size );
        }
        
        if( size < w_format.Size() ) {
            format_info = new uint8_t[w_format.Size()];
            info = format_info;
        }
        
        n = w_format.Size();
        w_format.Write( info, n );
        size = w_format.Size();

        return 1;
    }
    
    return 0;
}
