/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>
#include <assert.h>

#include "JpegCodec.h"
#include "debug.h"
#include "VideoFile/translate.h"



const bool JpegCodec::jpeg_baseline = false;
const J_DCT_METHOD JpegCodec::dct_method = JDCT_IFAST;
//const J_DCT_METHOD JpegCodec::dct_method = JDCT_FLOAT;

/*****************************************************************************/
/***** Videofunktionen *******************************************************/
/*****************************************************************************/

JpegCodec::JpegCodec( V_Track *Track )
:V_Codec( Track )
{
    int i;

    // initialisiere Daten-Puffer (maximale Gre bei RGB32 - unkompremiert)
    out_bytes = breite * hoehe * 4;
    out_data = new unsigned char[ out_bytes ];
    assert( out_data != NULL );
    in_data = NULL;
    in_bytes = 0;

    // initialisiere row_pointers und Komponenten-Array
    row_pointers = new (unsigned char *)[hoehe];
    assert( row_pointers != NULL );

    Komponenten = new JSAMPARRAY[3];
    assert( Komponenten != NULL );


    // Jpeg - Compressor initialisieren
    cinfo.err = jpeg_std_error( &jcerr );
    jpeg_create_compress( &cinfo );

    // Werte setzen
    cinfo.image_width = breite;
    cinfo.image_height = hoehe;
    cinfo.input_components = 3;
    cinfo.in_color_space = JCS_YCbCr;           // schreibe Farbraum YUV
    jpeg_set_defaults( &cinfo );
    jpeg_set_colorspace( &cinfo, JCS_YCbCr );

    // Rohdaten verarbeiten, samp_factor setzen (fr YUV420 downsampling)
    cinfo.raw_data_in = true;
    cinfo.comp_info[0].v_samp_factor = 2;
    cinfo.comp_info[0].h_samp_factor = 2;
    cinfo.comp_info[1].v_samp_factor = 1;
    cinfo.comp_info[1].h_samp_factor = 1;
    cinfo.comp_info[2].v_samp_factor = 1;
    cinfo.comp_info[2].h_samp_factor = 1;

    // Erstelle Zeiger auf Bildreihen der Komponenten
    // Anzahl = v_samp_factor * DCTSIZE
    // es werden von jpeglib immer Vielfache von v_samp_factor * DCTSIZE Zeilen verarbeitet
    for( i=0; i<3; i++ ) {
        Komponenten[i] = new JSAMPROW[ cinfo.comp_info[i].v_samp_factor * DCTSIZE ];
        assert( Komponenten[i] != NULL );
    }


    // Zeiger auf Codec-Daten
    cinfo.client_data = (void *)this;

    // DCT - Einstellungen
    cinfo.dct_method = dct_method;
    //cinfo.data_precision = 8;

    jpeg_suppress_tables( &cinfo, true );

    // Methoden zum schreiben der kompremierten Daten
    dest.init_destination = &JpegCodec::init_destination;
    dest.empty_output_buffer = &JpegCodec::empty_output_buffer;
    dest.term_destination = &JpegCodec::term_destination;
    cinfo.dest = &dest;


    // Decoder initialisieren
    dinfo.err = jpeg_std_error( &jderr );
    jpeg_create_decompress( &dinfo );

    // Methoden zum einlesen der komprimierten Quelldaten
    source.init_source = &JpegCodec::init_source;
    source.fill_input_buffer = &JpegCodec::fill_input_buffer;
    source.skip_input_data = &JpegCodec::skip_input_data;
    source.resync_to_restart = &jpeg_resync_to_restart;
    source.term_source = &JpegCodec::term_source;
    dinfo.src = &source;

    // Zeiger auf Codec-Daten
    dinfo.client_data = (void *)this;

    AddOption( "JpegQuality", _("quality factor for JPEG compression 0..100") ) = 80;

    // Color Model einstellen - Standard YUV420P
    ColorModel( YUV420P );
}





JpegCodec::~JpegCodec()
{
    jpeg_destroy_compress( &cinfo );
    jpeg_destroy_decompress( &dinfo );

    delete [] out_data;
    delete [] row_pointers;
}




// Farbraum festlegen
bool JpegCodec::ColorModel( int cmodel )
{
    // verwende intern Standardfarbraum und benutze Konvertierung
    if( iomode == VideoFile::write ) {
        if( cmodel != YUV420P ) {
            V_Codec::ColorModel( YUV420P );
            return false;
        }
    } else {
        if( cmodel != RGB24 ) {
            V_Codec::ColorModel( RGB24 );
            return false;
        }
    }

    return V_Codec::ColorModel( cmodel );
}






// JPEG-Qualitt
bool JpegCodec::jpegQuality( int q )
{
    jpeg_set_quality( &cinfo, q, jpeg_baseline );
    return true;
}





int JpegCodec::SetOption( OptionEntry *opt )
{
    if( opt->Name() == "JpegQuality" ) {
        jpegQuality( (int)*opt );

    } else {
        return 0;
    }

    return 1;
}




// Bild schreiben
int JpegCodec::encode( VideoFrame *bild )
{
    int i, z=0, Zeilen;

    jpeg_start_compress( &cinfo, true );

    // erstelle Zeiger auf YUV-Komponenten
    if( bild->GetFrame( row_pointers[0], row_pointers[1], row_pointers[2] ) <= 0 )
        return 0;
   
    // i = jpeg_write_scanlines( &cinfo, row_pointers, hoehe);
    // Daten schreiben
    // Zeilenanzahl fr write_raw mu vielfaches von 2*DCTSIZE sein
    Zeilen = (int)(( cinfo.comp_info[0].height_in_blocks + 1 ) / 2) * 2 * DCTSIZE;
    do {
        // Komponenten-Felder
        //   Y-Plane
        for( i=0; i < 2 * DCTSIZE; i++ ) {
            (Komponenten[0])[i] = row_pointers[0] + (i+z) * breite;
        }
        //  U- und V-Plane
        for( i=0; i < DCTSIZE; i++ ) {
            (Komponenten[1])[i] = row_pointers[1] + (i+z/2) * breite/2;
            (Komponenten[2])[i] = row_pointers[2] + (i+z/2) * breite/2;
        }

        z += jpeg_write_raw_data( &cinfo, Komponenten, 2 * DCTSIZE );
    } while( z < Zeilen );
    assert( z == Zeilen );

    jpeg_finish_compress( &cinfo );

    return 1;
}







// Video-Bild lesen
int JpegCodec::decode( VideoFrame *bild )
{
    int i=0, z=0, stat;
    uint8_t *bild_buf = NULL;

    if( bild->FramePointer( bild_buf ) <=0 )
        return 0;
        
    erz_row_pointers( row_pointers, bild_buf );
    
    if( !(stat = track->readData( in_data, in_bytes )) )
        return stat;

    i = jpeg_read_header( &dinfo, true );
    assert( i == JPEG_HEADER_OK );

    // Farbraum fr Decodierung einstellen ( lese RGB-Daten )
    dinfo.out_color_space = JCS_RGB;

    // DCT - Einstellungen
    cinfo.dct_method = dct_method;
    
    jpeg_start_decompress( &dinfo );

    // Bild decodieren
    for( z=0; z<hoehe; ) {
        i = jpeg_read_scanlines( &dinfo, &(row_pointers[z]), hoehe - z );
        if( i==0 ) break;
        z += i;
    }
    assert( z == hoehe );

    jpeg_finish_decompress( &dinfo );

    return 1;
}





int JpegCodec::CopyFrame( void )
{
    if( iomode != VideoFile::write )
        return 0;

    return track->writeData( out_data, out_bytes - (cinfo.dest->free_in_buffer) );
}




/****************************************************************************************/
/***** Hilfsfunktionen                                                              *****/
/****************************************************************************************/
inline void JpegCodec::erz_row_pointers( unsigned char **row_pointers, unsigned char *buf )
{
    switch( color_model ) {
    case RGB24:
    case RGB24|BGR:
        for( int i=0; i<hoehe; i++ ) {
            row_pointers[i] = buf + i * 3 * breite;
        }
        break;
    default:
        ERROR_TXT( _("erz_row_pointers: color model %i not implemented!"), color_model_intern );
        return;
    }
}








/****************************************************************************************/
/***** Interne Funktionen fr Ein-/Ausgabe von libjpeg                              *****/
/****************************************************************************************/

// Compressor
void JpegCodec::init_destination( j_compress_ptr cinfo ) {
    JpegCodec *Codec;
    Codec = (JpegCodec *)cinfo->client_data;

    cinfo->dest->next_output_byte = Codec->out_data;
    cinfo->dest->free_in_buffer = Codec->out_bytes;
}


boolean JpegCodec::empty_output_buffer( j_compress_ptr cinfo ) {
    assert( false );
    return false;
}


void JpegCodec::term_destination( j_compress_ptr cinfo ) {
    JpegCodec *Codec;
    Codec = (JpegCodec *)cinfo->client_data;

    Codec->track->writeData( Codec->out_data, Codec->out_bytes - (cinfo->dest->free_in_buffer) );
}





// Decompressor
void JpegCodec::init_source( j_decompress_ptr dinfo )
{
    JpegCodec *Codec;
    Codec = (JpegCodec *)dinfo->client_data;

    dinfo->src->next_input_byte = Codec->in_data;
    dinfo->src->bytes_in_buffer = Codec->in_bytes;
}



void JpegCodec::skip_input_data ( j_decompress_ptr dinfo, long num_bytes )
{
    if( num_bytes <= 0 )
        return;

    assert( dinfo->src->bytes_in_buffer >= (unsigned long)num_bytes );

    dinfo->src->next_input_byte += num_bytes;
    dinfo->src->bytes_in_buffer -= num_bytes;
}



boolean JpegCodec::fill_input_buffer( j_decompress_ptr dinfo )
{
    assert( false );
    return false;
}


void JpegCodec::term_source( j_decompress_ptr dinfo )
{
    return;
}
