/* 
 *  video_out_null.c
 *
 *	Copyright (C) Rainer Johanni, Andrew Stevens Jan 2001
 *
 *  This file is part of mpeg2dec, a free MPEG-2 video stream decoder.
 *	
 *  mpeg2dec is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  mpeg2dec is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <inttypes.h>

#include "video_out.h"
#include "video_out_internal.h"

typedef struct yuv_instance_s {
    vo_instance_t vo;
    int prediction_index;
    vo_frame_t * frame_ptr[3];
    vo_frame_t frame[3];
    int width;
    int height;
	int frame_rate_code;
	int owidth;
	int oheight;
	int omargin;
	int8_t *image_data;
    int framenum;
	FILE *out_strm;
} yuv_instance_t;

extern vo_instance_t yuv_vo_instance;
extern vo_instance_t yuvh_vo_instance;
extern vo_instance_t yuvs_vo_instance;

static int yuv_core_setup (yuv_instance_t *this,
						   int width, int height, 
						   int frame_rate_code,
						   void (* draw_frame) (vo_frame_t *))
{

	this->width = width;
	this->height = height;
	this->frame_rate_code = frame_rate_code;
	this->framenum = -2;
	this->vo.close = libvo_common_free_frames;
    this->vo.get_frame = libvo_common_get_frame;

	/* RJ: Framerate specification should really be based on
	   something more reliable than height */
	this->out_strm = stdout;

    return libvo_common_alloc_frames ((vo_instance_t *)this, width, height,
									  sizeof (vo_frame_t), NULL, NULL, draw_frame);
}

static void yuv_common_free_frames (vo_instance_t * _this)
{
	yuv_instance_t *this = (yuv_instance_t *) _this;
	if( this->out_strm != stdout )
		fclose( this->out_strm );
    libvo_common_free_frames (_this);
}

static void yuv_draw_frame (vo_frame_t * frame)
{
	yuv_instance_t *this = (yuv_instance_t *) frame->instance;
	size_t n = 0;

	if (++(this->framenum) < 0 )
		return;

  	fwrite("FRAME\n",6,1,this->out_strm);
  	n += fwrite(frame->base[0],
				1, this->width * this->height,  this->out_strm);
  	n += fwrite(frame->base[1],
				1, this->width * this->height/4,  this->out_strm);
  	n += fwrite(frame->base[2],
				1, this->width * this->height/4,  this->out_strm);

 	if(n!=this->width * this->height * 3/2)
  	{
    	fprintf(stderr,"Write of YUV output failed %d not written\n", this->width * this->height *3/2 - n);
   		exit(1);
  	}

}

static int vo_yuv_setup( vo_instance_t *_this, int width, int height, int frame_rate_code )
{
	char str[256];
	yuv_instance_t *this = (yuv_instance_t *)_this;
	int res = yuv_core_setup( this, width, height, frame_rate_code, yuv_draw_frame );

	if( ! res )
	{
		sprintf(str,"YUV4MPEG %d %d %d\n",width,height,frame_rate_code);
		fwrite(str,strlen(str),1, this->out_strm);
	}
	return res;
}

vo_instance_t * vo_yuv_open(void)
{
	yuv_instance_t *instance = malloc(sizeof(yuv_instance_t));

	if( instance == NULL )
		return NULL;

	instance->vo.setup = vo_yuv_setup;
	return (vo_instance_t *) instance;
}


static void yuvh_draw_frame (vo_frame_t * frame)
{
	uint8_t *dst;
	uint8_t **src = frame->base;
	yuv_instance_t *this = (yuv_instance_t *) frame->instance;	
	int iw = this->owidth;
	int ih = this->oheight;
	int is = this->omargin;
	int i, j;
	size_t n;

	if (++(this->framenum) < 0 )
		return;


	dst = this->image_data;
	for(j=0;j<ih;j++)
		for(i=is;i<iw+is;i++)
			*(dst++) = (src[0][2*j*this->width+2*i]+src[0][2*j*this->width+2*i+1])>>1;

	dst = this->image_data + iw * ih;
	for(j=0;j<ih/2;j++)
		for(i=is/2;i<(iw+is)/2;i++)
			*(dst++) = (src[1][2*j*this->width/2+2*i]+src[1][2*j*this->width/2+2*i+1])>>1;

	dst = this->image_data + iw * ih * 5 / 4;
	for(j=0;j<ih/2;j++)
		for(i=is/2;i<(iw+is)/2;i++)
			*(dst++) = (src[2][2*j*this->width/2+2*i]+src[2][2*j*this->width/2+2*i+1])>>1;


	fwrite("FRAME\n",6,1,this->out_strm);
	n = fwrite(this->image_data,1,iw * ih * 3/2,this->out_strm);
	if(n!=iw * ih * 3/2)
	{
		fprintf(stderr,"Write of output failed\n");
		exit(1);
	}
 
}

static int  vo_yuvh_setup( vo_instance_t *_this, int width, int height, int frame_rate_code )
{
	yuv_instance_t *this = (yuv_instance_t *)_this;
	int res = yuv_core_setup( this, width, height, frame_rate_code, yuvh_draw_frame );
	char str[255];

	this->oheight = height/2;
	this->owidth = (width/32)*16;
	this->omargin = (width - 2*this->owidth)/4;

	this->image_data = (uint8_t *)malloc(this->owidth*this->oheight*3/2);
	if( this->image_data == NULL )
		return 1;

	sprintf(str,"YUV4MPEG %d %d %d\n",this->owidth,this->oheight,frame_rate_code);
	fwrite(str,strlen(str),1, this->out_strm);

	return res;
}


vo_instance_t * vo_yuvh_open(void)
{
	yuv_instance_t *instance = malloc(sizeof(yuv_instance_t));

	if( instance == NULL )
		return NULL;

	instance->vo.setup = vo_yuvh_setup;
	return (vo_instance_t *) instance;
}


static void yuvs_draw_frame (vo_frame_t * frame)
{
	uint8_t *dst;
	uint8_t **src = frame->base;
	uint8_t *srcp;
	yuv_instance_t *this = (yuv_instance_t *) frame->instance;	
	int iw = this->owidth;
	int ih = this->oheight;
	int is = this->omargin;
	int i, j;
	size_t n;

	if (++(this->framenum) < 0 )
		return;

	dst = this->image_data;
	for(j=0;j<ih;j++)
		for(i=is;i<iw+is;i+=2)
		{
			srcp = &src[0][j*this->width+i*3/2];
			dst[0] = (2*srcp[0]+srcp[1])/3;
			dst[1] = (2*srcp[2]+srcp[1])/3;
			dst += 2;
		}

	dst = this->image_data + iw * ih;
	for(j=0;j<ih/2;j++)
		for(i=is/2;i<(iw+is)/2;i+=2)
		{
			srcp = &src[1][j*this->width/2+i*3/2];
			dst[0] = (2*srcp[0]+srcp[1])/3;
			dst[1] = (2*srcp[2]+srcp[1])/3;
			dst += 2;
		}
	dst = this->image_data + iw * ih * 5 / 4;
	for(j=0;j<ih/2;j++)
		for(i=is/2;i<(iw+is)/2;i+=2)
		{
			srcp = &src[2][j*this->width/2+i*3/2];
			dst[0] = (2*srcp[0]+srcp[1])/3;
			dst[1] = (2*srcp[2]+srcp[1])/3;
			dst += 2;
		}


	fwrite("FRAME\n",6,1,this->out_strm);
	n = fwrite(this->image_data,1,iw * ih * 3/2,this->out_strm);
	if(n!=iw * ih * 3/2)
	{
		fprintf(stderr,"Write of output failed\n");
		exit(1);
	}
 
}

static int vo_yuvs_setup( vo_instance_t *_this, int width, int height, int frame_rate_code )
{
	yuv_instance_t *this = (yuv_instance_t *)_this;
	int res = yuv_core_setup( this, width, height, frame_rate_code, yuvs_draw_frame );
	char str[255];

	this->oheight = height;
	this->owidth = (width*2/(3*16))*16;
	this->omargin = (width*2/3 - this->owidth)/2;
	if( this->owidth != 480 )
	{
		fprintf( stderr, "Warning: Output width is %d which is not suitable for SVCD!\n", this->owidth);
	}

	this->image_data = (uint8_t *)malloc(this->owidth*this->oheight*3/2);
	if( this->image_data == NULL )
		return 1;

	sprintf(str,"YUV4MPEG %d %d %d\n",this->owidth,this->oheight,frame_rate_code);
	fwrite(str,strlen(str),1, this->out_strm);

	return res;
}


vo_instance_t * vo_yuvs_open(void)
{
	yuv_instance_t *instance = malloc(sizeof(yuv_instance_t));

	if( instance == NULL )
		return NULL;

	instance->vo.setup = vo_yuvs_setup;
	return (vo_instance_t *) instance;
}

