/*****************************************************************************\
                          RTjpegCodec.C  -  description
                             -------------------
    begin                : Sun Nov 25 2001
    copyright            : (C) 2001 by Christian Tusche
    email                : ctusche@gmx.net
\*****************************************************************************/

/*****************************************************************************\
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <stdio.h>
#include <assert.h>
#include "debug.h"
#include "RTjpegCodec.h"
#include "VideoFile/translate.h"

RTjpegCodec::RTjpegCodec( V_Track *Track )
:V_Codec( Track )
{
    int w = breite;
    int h = hoehe;

    // Optionen
    quality = 100;
    key_interval = 25;
    motion_level = 1;


    // Compressor / Dekompressor initialisieren
    compress_struct = RTjpeg_init();
    int i = RTJ_YUV420;
    RTjpeg_set_format( compress_struct, &i );
    RTjpeg_set_size( compress_struct, &w, &h );

    // DCT-Datenkompression
    RTjpeg_set_compress_method( compress_struct, RTJ_COMPRESS_ZLIB, 1 );

    decompress_struct = RTjpeg_init();
    i = RTJ_YUV420;
    RTjpeg_set_format( decompress_struct, &i );

    // Datenpuffer fr kompremierte rjpg - Daten
    // nehme maximale gre fr unkomopremierte RGB32 Daten an
    // das sollte au jeden Fall reichen
    data_buf_bytes = w * h * 4;
    data_buf = new uint8_t[ data_buf_bytes ];
    data_bytes = 0;

    AddOption( "JpegQuality", _("quality factor for JPEG compression 0..100") ) = quality;
    AddOption( "KeyInterval", _("distance between key-frames") ) = key_interval;
    AddOption( "MotionLevel", _("motion level for intraframe compression") ) = motion_level;

    letztes_bild = 0;

    ColorModel( YUV420P );
}






RTjpegCodec::~RTjpegCodec()
{
    RTjpeg_close( compress_struct );
    RTjpeg_close( decompress_struct );

    delete [] data_buf;
}






// Farbraum festlegen
bool RTjpegCodec::ColorModel( int cmodel )
{
    DEBUG_TXT( "RTjpegCodec::ColorModel: setze Colormodel %i", cmodel );

    // verwende intern Standardfarbraum und benutze Konvertierung
    if( cmodel != YUV420P ) {
        V_Codec::ColorModel( YUV420P );
        return false;
    }

    return V_Codec::ColorModel( cmodel );
}






// JPEG-Qualitt
bool RTjpegCodec::jpegQuality( int q )
{
    DEBUG_TXT_3( "setze Qualitt %i", q );

    if( q < 0 )
        q = 0;
    else if( q > 100 ) 
        q = 100;

    quality = q;

    int i = (int)((float)quality / 100.0 * 160.0) + 32;
    RTjpeg_set_quality( compress_struct, &i );

    return true;
}





int RTjpegCodec::SetOption( OptionEntry *opt )
{
    if( opt->Name() == "JpegQuality" ) {
        jpegQuality( (int)*opt );

    } else if( opt->Name() == "KeyInterval" ) {
        key_interval = (int)*opt;
        if( key_interval < 0 )
            key_interval = 0;

        // Intra-Frame-Kompression
        RTjpeg_set_intra( compress_struct, &key_interval, &motion_level, &motion_level );

    } else if( opt->Name() == "MotionLevel" ) {
        motion_level = (int)*opt;
        if( motion_level < 0 )
            motion_level = 0;

        // Intra-Frame-Kompression
        RTjpeg_set_intra( compress_struct, &key_interval, &motion_level, &motion_level );

    } else {
        return 0;
    }

    return 1;
}







// Bild schreiben
int RTjpegCodec::encode( VideoFrame *bild )
{
    uint8_t *yuv[3];

    // DEBUG_TXT( "RTjpegCodec::encode" );

    // Zeiger auf Planes erstellen
    if( bild->GetFrame( yuv[0], yuv[1], yuv[2] ) <= 0 )
        return 0;
    
    data_bytes = RTjpeg_compress( compress_struct, (int8_t *)data_buf, yuv );

    // Daten schreiben
    return track->writeData( data_buf, data_bytes );
}







// Video-Bild lesen
int RTjpegCodec::decode( VideoFrame *bild )
{
    int i=0;
    uint8_t *data, *yuv[3];
    long bytes;
    int key;

    // Daten einlesen
    if( (i = track->readData( data, bytes )) < 1 ) {
        return i;
    }

    // Zeiger auf Planes erstellen
    if( bild->FramePointer( yuv[0], yuv[1], yuv[2] ) <= 0 )
        return 0;

    // Bilder seit letztem Keyframe
    key =  RTjpeg_frame_key( (int8_t *)data );
    if( key != 0 && track->Position() != letztes_bild + 1 ) {
        DEBUG_TXT( _("reposition for frame %li"), track->Position() - 1 );
        // gehe zu letztem key-Frame zurck
        if( track->Seek( track->Position() - key - 1 ) ) {
            for( i=0; i<key+1; i++ ) {
                track->readData( data, bytes );
                RTjpeg_decompress( decompress_struct, (int8_t *)data, yuv );
            }
        }
    } else {
        RTjpeg_decompress( decompress_struct, (int8_t *)data, yuv );
    }

    letztes_bild = track->Position();

    return 1;
}




int RTjpegCodec::CopyFrame( void )
{
    if( iomode != VideoFile::write )
        return 0;

    return track->writeData( data_buf, data_bytes );
}
