/*****************************************************************************\
**  Author  : Christian Tusche                                               **
**  E-Mail  : ctusche@gmx.net                        **
**                                                                           **
** This program is free software; you can redistribute it and/or modify      **
** it under the terms of the GNU General Public License as published by      **
** the Free Software Foundation; either version 2 of the License, or         **
** (at your option) any later version.                                       **
**                                                                           **
** This program is distributed in the hope that it will be useful,           **
** but WITHOUT ANY WARRANTY; without even the implied warranty of            **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
** GNU General Public License for more details.                              **
**                                                                           **
** You should have received a copy of the GNU General Public License         **
** along with this program; if not, write to the Free Software               **
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 **
\*****************************************************************************/

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>
#include <assert.h>
#include <malloc.h>

#include "XvidCodec.h"
#include "debug.h"
#include "VideoFile/translate.h"



/*****************************************************************************/
/***** Videofunktionen *******************************************************/
/*****************************************************************************/

XvidCodec::XvidCodec( V_Track *Track )
:V_Codec( Track )
{
    // Ausgabe erzeugen
    render = 1;

    decode_setup = false;
    encode_setup = false;

    // Optionen registrieren
    AddOption( "Bitrate", _("average bitrate in kbps") ) = 980;
    AddOption( "RcPeriod", _("averageing period for rate control (reaction time)") ) = 500;
    AddOption( "MinQuant", _("lowest quantisation (1: high quality - 32: low quality)") ) = 4;
    AddOption( "MaxQuant", _("highest quantisation (1: high quality - 32: low quality)") ) = 8;
    AddOption( "KeyInterval", _("distance between I-frames") ) = 200;
    AddOption( "Quality", _("quality of the encoder (1: fast - 7: slow)") ) = 7;
    AddOption( "QuantType", _("quantisation method - 0: h.263 / 1: MPEG4") ) = 0;
    //AddOption( "Lumimasking", _("use Lumimasking 0: off / 1: on") ) = 1;

    motion_opt = 0;
    general_opt = 0;
    postproces = 0;     // Ausgabefilter

    //AddOption( "Postproces" ) = postproces;

    // Bildrate fr encoder
    rate = track->Rate();

    // Ausgabepuffer fr encode
    puffer = new unsigned char[ breite * hoehe * 3];

    // Puffer fr ein leeres Bild
    null_puffer = NULL;
    null_puffer_gr = 0;

    // Color Model einstellen - Standard RGB24
    colorspace = XVID_CSP_I420;

    ColorModel( YUV420P );
}



XvidCodec::~XvidCodec()
{
    // lsche decoder
    if( decode_setup ) {
        xvid_decore( decode_handle, XVID_DEC_DESTROY, NULL, NULL );
    }

    // lsche Encoder
    if( encode_setup ) {
        xvid_encore( encode_handle, XVID_ENC_DESTROY, NULL, NULL );
    }

    delete [] puffer;
}






// Farbraum festlegen
bool XvidCodec::ColorModel( int cmodel )
{
    // verwende intern Standardfarbraum und benutze Konvertierung
    if( cmodel != YUV420P ) {
        V_Codec::ColorModel( YUV420P );
        return false;
    }

    return V_Codec::ColorModel( cmodel );
}






// positionieren
int XvidCodec::SeekFrame( long bild )
{
    long keyframe = track->GetKeyFrame( bild );

    // gehe zu letztem Keyframe
    DEBUG_TXT( _("seek to keyframe %li for frame %li"), keyframe, bild );
    // auf jeden Fall erstes Bild decodieren, um Puffer zu fllen
    if( track->Position() == 0 ) {
        render = 0;
        decode( NULL );
        render = 1;
    }

    if( !track->Seek( keyframe ) ) {
        return 0;
    }

    // decodiere Frames bis zur aktuellen Position
    render = 0;
    while( track->Position() < bild ) {
        decode( NULL );
    }
    render = 1;

    return 1;
}






// Video-Bild lesen
int XvidCodec::decode( VideoFrame *bild )
{
    uint8_t *data=NULL, *bild_buf = NULL;
    long bytes=0;
    XVID_DEC_FRAME dec_frame;
    int stat;

    if( !decode_setup )
        init_decode();
    if( !decode_setup )
        return 0;

    if( bild == NULL ) {
        bild_buf = puffer;
    } else {
        if( bild->FramePointer( bild_buf ) <= 0 )
            return 0;
    }
        
    if( !track->readData( data, bytes ) )
        return 0;

    // Bild decodieren
    dec_frame.bitstream = data;
    dec_frame.image = bild_buf;
    dec_frame.length = bytes;
    dec_frame.stride = breite;
    if( render )
        dec_frame.colorspace = colorspace;
    else
        dec_frame.colorspace = XVID_CSP_NULL;

    stat = xvid_decore( decode_handle, XVID_DEC_DECODE, &dec_frame, NULL );

    if( stat != XVID_ERR_OK ) {
        DEBUG_TXT( _("decore status = %i"), stat );
        return 0;
    }

    return 1;
}







// Video-Bild schreiben
int XvidCodec::encode( VideoFrame *bild )
{
    XVID_ENC_FRAME frame;
    XVID_ENC_STATS result;
    uint8_t *bild_buf = NULL;

    if( bild->GetFrame( bild_buf ) <= 0 )
        return 0;
        
    if( !encode_setup )
        init_encode();
    if( !encode_setup )
        return 0;
    

    memset( &frame, 0 , sizeof(XVID_ENC_FRAME) );
    frame.general = general_opt;
    frame.motion = motion_opt;

    frame.image = bild_buf;             // the image frame to be encoded
    frame.bitstream = puffer;           // the buffer for encoded bitstream
    frame.length = 0;                   // the length of the encoded bitstream
    frame.colorspace = colorspace;      // the format of image frame
    frame.quant = 0;                    // quantizer for this frame; only used in VBR modes

    if( motion_opt == 0 )
        frame.intra = 1;
    else
        frame.intra = -1;                   // force this frame to be intra/inter; only used in VBR 2-pass

    xvid_encore( encode_handle, XVID_ENC_ENCODE, &frame, &result );

    int keyframe = 0;
    if( frame.intra ) {
        keyframe = V_Track::IS_KEYFRAME;
    }

    // Bild zum Ausfllen von ausgelassenen Bildern speichern
    if( null_puffer_gr == 0 && keyframe == 0 && frame.length < 250 ) {
        null_puffer_gr = frame.length;
        null_puffer = new unsigned char[ null_puffer_gr ];
        memcpy( null_puffer, puffer, frame.length );
    }

    //DEBUG_TXT( "schreibe %li Bytes, Keyframe=%i", frame.length, keyframe );
    track->writeData( puffer, frame.length, keyframe );

    return 1;
}







/***** letztes Bild kopieren *****/
int XvidCodec::CopyFrame( void )
{
    if( iomode != VideoFile::write )
        return 0;

    if( null_puffer_gr > 0 ) {
        return track->writeData( null_puffer, null_puffer_gr, 0 );
    } else {
        return V_Codec::CopyFrame();
    }
}





int XvidCodec::SetOption( OptionEntry *opt )
{
    DEBUG_TXT_3( "setze Option %s = %s", (const char*)opt->Name(), (const char*)*opt );
    /*
    if( opt->Name() == "Postproces" ) {
        postproces = (int)*opt;
        if( decode_setup ) {
            DEC_SET dec_set;
            memset( &dec_set, 0, sizeof( DEC_SET ) );
            dec_set.postproc_level = postproces;
            decore( decode_handle, DEC_OPT_SETPP, &dec_set, NULL);
        }
    }
    */

    if( opt->Name() == "Bitrate" ) {
        if( (int)*opt < 10 )
            *opt = 10;
    } else if( opt->Name() == "RcPeriod" ) {
        if( (int)*opt < 1 )
            *opt = 1;
    } else if( opt->Name() == "MinQuant" ) {
        if( (int)*opt < 1 )
            *opt = 1;
        else if( (int)*opt > 32 )
            *opt = 32;
    } else if( opt->Name() == "MaxQuant" ) {
        if( (int)*opt < 1 )
            *opt = 1;
        else if( (int)*opt > 32 )
            *opt = 32;
    } else if( opt->Name() == "KeyInterval" ) {
        if( (int)*opt < 1 )
            *opt = 1;
    } else {
        return 0;
    }

    return 1;
}








/***** Decoder initialisieren *****/
void XvidCodec::init_decode( void )
{
    // Xvid kann keine kleineren Bilder vrearbeiten!
    if( breite < 128 || hoehe < 128 ) {
        ERROR_TXT( _("Frames smaller than 128x128 pixesl cannot be processed!") );
        return;
    }
        
    if( decode_setup )
        return;

    XVID_INIT_PARAM xinit;

    xinit.cpu_flags = 0;
    xvid_init(NULL, 0, &xinit, NULL);

    dec_param.width = breite;
    dec_param.height = hoehe;
    xvid_decore(NULL, XVID_DEC_CREATE, &dec_param, NULL);
    decode_handle = dec_param.handle;

    decode_setup = true;
}








/***** Encoder initialisieren *****/
void XvidCodec::init_encode( void )
{
    // Xvid kann keine kleineren Bilder vrearbeiten!
    if( breite < 128 || hoehe < 128 ) {
        ERROR_TXT( _("Frames smaller than 128x128 pixesl cannot be processed!") );
        return;
    }
    
    if( encode_setup )
        return;

    XVID_INIT_PARAM xinit;
    int bitrate;        // Mittlere Bittrate
    int rc_period;      // Mittelungsperiode fr Ratecontrol
    int min_quant;      // Quantisierung - Untergrenze
    int max_quant;      // Quantisierung - Obergrenze
    int key_interval;   // Abstand zwischen I-Frames

    // Parameter einlesen
    bitrate = (int)Option( "Bitrate" );
    rc_period = (int)Option( "RcPeriod" );
    min_quant = (int)Option( "MinQuant" );
    max_quant = (int)Option( "MaxQuant" );
    key_interval = (int)Option( "KeyInterval" );
   
    // Encoder initialisieren
    xinit.cpu_flags = 0;
    xvid_init(NULL, 0, &xinit, NULL);

    memset( &enc_param, 0, sizeof(XVID_ENC_PARAM) );
    enc_param.width = breite;                            // the x dimension of the frames to be encoded
    enc_param.height = hoehe;                            // the y dimension of the frames to be encoded
    enc_param.fincr = 1000;                              // frame increment, fbase. each frame = "fincr/fbase" seconds
    enc_param.fbase = (int)(rate * 1000);
    enc_param.rc_bitrate = bitrate * 1000;               // the bitrate of the target encoded stream, in bits/s
    enc_param.rc_averaging_period = rc_period;           // Tunes how fast the rate control reacts - lower values are faster.
    enc_param.rc_reaction_delay_factor = 1 + rc_period / 25;  // Tunes how fast the rate control reacts - lower values are faster.
    enc_param.rc_buffer = bitrate * 5;                   // the rate control buffersize / max. allowed deviation
    enc_param.max_quantizer = max_quant;                 // the upper limit of the quantizer
    enc_param.min_quantizer = min_quant;                 // the lower limit of the quantizer
    enc_param.max_key_interval = key_interval;           // the maximum interval between key frames

    // Optionen fr Motion Compensation und Quantisierung
    if( (int)Option( "QuantType" ) == 0 ) {
        general_opt = XVID_H263QUANT;
    } else {
        general_opt = XVID_MPEGQUANT;
    }

    /*
    if( (int)Option( "Lumimasking" ) == 0 ) {
        general_opt |= XVID_LUMIMASKING;
    }
    */
    
    general_opt |= XVID_HALFPEL;

    switch( (int)Option( "Quality" ) ) {
        case 1:
            general_opt |= XVID_ME_PMVFAST;
            motion_opt = PMV_ADVANCEDDIAMOND16;
            break;
        case 2:
            general_opt |= XVID_ME_PMVFAST;
            motion_opt = PMV_QUICKSTOP16 |PMV_ADVANCEDDIAMOND16 ;
            break;
        case 3:
            general_opt |= XVID_ME_PMVFAST;
            motion_opt = PMV_EARLYSTOP16 | PMV_ADVANCEDDIAMOND16;
            break;
        case 4:
            general_opt |= XVID_HALFPEL | XVID_ME_PMVFAST;
            motion_opt = PMV_EARLYSTOP16 | PMV_HALFPELREFINE16;
            break;
        case 5:
            general_opt |= XVID_HALFPEL | XVID_INTER4V | XVID_ME_PMVFAST;
            motion_opt = PMV_EARLYSTOP16 | PMV_HALFPELREFINE16 | PMV_EARLYSTOP8 | PMV_HALFPELREFINE8;
            break;
        case 6:
            general_opt |= XVID_HALFPEL | XVID_INTER4V | XVID_ME_EPZS;
            motion_opt = PMV_EARLYSTOP16 | PMV_HALFPELREFINE16 | PMV_HALFPELDIAMOND16 |
                         PMV_EARLYSTOP8  | PMV_HALFPELREFINE8 | PMV_HALFPELDIAMOND8;
            break;
        case 7:
        default:
            general_opt |= XVID_HALFPEL | XVID_INTER4V | XVID_ADAPTIVEQUANT | XVID_ME_EPZS;
            motion_opt = PMV_EARLYSTOP16 | PMV_HALFPELREFINE16 | PMV_EXTSEARCH16 | PMV_HALFPELDIAMOND16 |
                         PMV_EARLYSTOP8  | PMV_HALFPELREFINE8  | PMV_EXTSEARCH8  | PMV_HALFPELDIAMOND8;
            break;
    }

    xvid_encore(NULL, XVID_ENC_CREATE, &enc_param, NULL);
    encode_handle = enc_param.handle;

    encode_setup = true;
}
